#!/usr/bin/env python

# written by:    Ka Ming Nip
# updated on:    September 9, 2013
# compatibility: Trans-ABySS 1.4.8 (public)

import os
import sys
import glob
import random
from utilities import cfg
from utilities import job_runner
from utilities import prepare_reads
from utilities.job_runner import MyJob

try:
    PACKAGE_DIR = os.environ['TRANSABYSS_PATH']
except KeyError:
    message = "Cannot find TRANSABYSS_PATH in your environment!"
    print_message("ERROR", message, True)

JOBNAME = 'JOBNAME'
QUEUE = 'QUEUE'
THREADS = 'THREADS'
MEM = 'MEM'
TMPMEM = 'TMPMEM'


def print_message(message_type, message, quit=False):
    """
    Prints a message, ie.
    message_type: message
    
    Args:
        message_type (str)  : heading of the message
        message      (str)  : body of the message
        quit         (bool) : quit this program with a non-zero exit status
    """
    sys.stderr.write(str(message_type) + ": " + str(message) + os.linesep)
    if quit:
        sys.exit(1)


def create_sge_req_string(default='-now n -R y', resources=None):
    """
    Returns the SGE resource request string. This is the value for 'SGE_RREQ' in qmake.
    
    Args:
         resources ({str:str,...}) : a dictionary of resources to request
         
    Returns:
         (str) : the SGE resource request string
    """

    items = []
    
    if default and len(default) > 0:
        items.append(default)
    
    if resources and len(resources) > 0:
        if JOBNAME in resources:
            var = str(resources[JOBNAME])
            if var and len(var) > 0:
                items.append('-N %s' % var)

        if QUEUE in resources:
            var = str(resources[QUEUE])
            if var and len(var) > 0:
                items.append('-q %s' % var)

        if THREADS in resources:
            var = str(resources[THREADS])
            if var and len(var) > 0:
                items.append('-pe ncpus %s' % var)

        if MEM in resources:
            var = str(resources[MEM])
            if var and len(var) > 0:
                items.append('-l mem_token=%s,mem_free=%s,h_vmem=%s' % (var, var, var))
        
        if TMPMEM in resources:
            var = str(resources[TMPMEM])
            if var and len(var) > 0:
                items.append('-l tmpfree=%s,tmp_token=%s' % (var, var))    
    
    return ' '.join(items)


def create_resources(jobname=None, config_key=None, settings=None):
    """
    Returns a dictionary of SGE resource requests
    
    Args:
        jobname    (str)           : the job name
        config_key (str)           : the configuration key
        settings   ({str:str:str}) : the Trans-ABySS configurations dictionary. {section : key : value}
    
    Returns:
        ({str:str}) : the dictionary of SGE resource requests
    """

    resources = {}
    
    resources[JOBNAME] = jobname
    
    mem_threads_str = cfg.get_value(settings, 'memory', config_key)
    if mem_threads_str:
        items = mem_threads_str.split(',')
        num_items = len(items)

        if num_items > 0:
            memory_str_stripped = items[0].strip()
            if len(memory_str_stripped) > 0:
                resources[MEM] = memory_str_stripped

        if num_items > 1:
            threads_str_stripped = items[1].strip()
            if len(threads_str_stripped) > 0:
                resources[THREADS] = int(threads_str_stripped)
    
    tmpmem = cfg.get_value(settings, 'tmpmem', config_key)
    if tmpmem:
        tmpmem_stripped = tmpmem.strip()
        if len(tmpmem_stripped) > 0:
            resources[TMPMEM] = tmpmem_stripped
    
    return resources


def create_patsubst_cmd(source, dest):
    """
    Return a patsubst-wildcard command for a source and a destination.
    
    Args:
        source (str) : path of the source file. It must contain '%'.
        dest   (str) : path of the destination file. It must contain '%'.
        
    Returns:
        (str) : the command
    """

    #patsubst assembly/k%/, filter/k%/HS0842-contigs.fa, $(wildcard assembly/k*/)
    wildcardstr = source.replace('%', '*')
    return 'patsubst %s, %s, $(wildcard %s)' % (source, dest, wildcardstr)


class Rule:
    """
    Class to represent the contents of a rule in a makefile.
    ie.
    targets: prerequisites
        preSgeReqCmds && \
        SGE_RREQ=sgereq && \
        { \
        postSgeReqCmds \
        ; } &> logfile    
    """

    def __init__(self, targets, prerequisites=None, sgereq=None, preSgeReqCmds=None, postSgeReqCmds=None, logfile=None):
        """
        Args:
            targets        ([str]) : list of targets to build
            prerequisites  ([str]) : list of prerequisites
            sgereq         (str)   : the SGE resource request string
            preSgeReqCmds  ([str]) : list of commands to run locally
            postSgeReqCmds ([str]) : list of commands to run on the cluster
            logfile        (str)   : path of the log file
        """
        self.targets = targets # list of string
        self.prerequisites = prerequisites # list of string
        self.sgereq = sgereq # string
        self.preSgeReqCmds = preSgeReqCmds # list of string
        self.postSgeReqCmds = postSgeReqCmds # list of string
        self.logfile = logfile # string
    
    def assemble(self):
        """
        Assembles the object into a rule string
        
        Returns:
            (str) : the rule string
        """

        #targets: prerequisites 
        rule = ' '.join(self.targets)
        rule += ':'    
        
        if self.prerequisites and len(self.prerequisites) > 0:
            rule += ' '
            rule += ' '.join(self.prerequisites)
        
        rule += os.linesep

        cmdseparator = ' && \\' + os.linesep + '	'

        #	preSgeReqCmds && \ 
        if self.preSgeReqCmds and len(self.preSgeReqCmds) > 0:
            rule += '	' + cmdseparator.join(self.preSgeReqCmds)
            
            if self.sgereq or self.postSgeReqCmds:
                rule += ' && \\'
            
            rule += os.linesep

        #	SGE_RREQ=sgereq && \ 
        if self.sgereq and len(self.sgereq) > 0:
            rule += '	SGE_RREQ="%s" && \\' % self.sgereq
            rule += os.linesep
        
        writeLog = self.logfile and len(self.logfile) > 0

        #	{ \ 
        if writeLog:
            rule += '	{ \\'
            rule += os.linesep
        
        #	postSgeReqCmds \ 
        if self.postSgeReqCmds and len(self.postSgeReqCmds) > 0:
            cmds = ['hostname']
            cmds.append('echo "START: `date`"')
            cmds.extend(self.postSgeReqCmds)
            cmds.append('echo "END: `date`"')
            rule += '	%s || exit 1 \\' % cmdseparator.join(cmds)
            rule += os.linesep
        
        #	; } &> logfile 
        if writeLog:
            rule += '	; } &> %s' % self.logfile
            
            #if it is a cluster job, add a job id postfix
            if self.sgereq and len(self.sgereq) > 0:
                rule += '$$JOB_ID'
            rule += os.linesep
        
        return rule


class Macro:
    """
    Class to represent the contents of a macro in a makefile.
    ie.
    name=definition
    """

    def __init__(self, name, definition, immediate=False):
        """
        Args:
            name       (str)  : name of the macro
            definition (str)  : definition of the macro
            immediate  (bool) : use immediate variable assignment instead of deferred variable assignment
        """
        self.name = name
        self.definition = definition
        self.immediate = immediate
    
    def assemble(self):
        """
        Assembles the object into a macro string.

        Returns:
            (str) : the macro string
        """
        operator = '='
        if self.immediate:
            operator = ':='
        
        return self.name + operator + self.definition + os.linesep


def write_items_to_file(items, makeFile):
    """
    Write rules and macros to a makefile.
    
    Args:
        items    ([qmake.Rule|qmake.Macro]) : a list of Rule objects or Macro objects
        makeFile (str)                      : path of makefile
    """
        
    fh = open(makeFile, 'w')
    
    for d in create_default_rules_and_macros():
        fh.write(d.assemble())        
        fh.write(os.linesep) # an empty line        
    
    for i in items:
        fh.write(i.assemble())        
        fh.write(os.linesep) # an empty line
    
    fh.close()


def create_default_rules_and_macros():
    """
    Returns the list of default macros and/or rules.
    
    Returns:
        ([qmake.Macro|qmake.Rule]) : the list of macros and/or rules
    """
    items = []
    
    items.append(Macro(name='SHELL', definition='/bin/bash -eu -o pipefail'))
    
    for target in ['.DELETE_ON_ERROR', '.SECONDARY', '.ONESHELL']:
        items.append(Rule(targets=[target]))
        
    return items


def run_pipe(pipe, options, max_parallel_make_tasks=100):
    """
    Run the Trans-ABySS pipe.

    Args:
        pipe    (trans-abyss.Pipe) : the Trans-ABySS Pipe object
        options (optparse.Values)  : the Trans-ABySS optparse options object
        max_parallel_make_tasks (int)   : the maximum number of make tasks to run in parallel
    """
    if isinstance( max_parallel_make_tasks, ( int, long ) ):
        if max_parallel_make_tasks < 1:
            print_message('ERROR', 'The maximum number of make tasks cannot be smaller than 1.', quit=True)
    else:
        print_message('ERROR', 'The maximum number of make tasks must be an integer greater than or equal to 1', quit=True)

    qmakedir = os.path.join(pipe.working_dir, 'qmake')
    if not os.path.isdir(qmakedir):
        os.makedirs(qmakedir)

    cur_path = os.getcwd()
    os.chdir(pipe.working_dir)
    
    mainMakeFile = None
    lib_name = pipe.lib
    
    use_ref_genome = True
    use_r2g_bam = True
    
    no_specific_targets = not options.fem and \
                          not options.prepare_reads and \
                          not options.prepare_contigs and \
                          not options.align_reads and \
                          not options.align_contigs and \
                          not options.fusion and \
                          not options.indel and \
                          not options.splicing and \
                          not options.track
    if not no_specific_targets:
        use_ref_genome = options.prepare_contigs or \
                         options.align_contigs or \
                         options.fusion or \
                         options.indel or \
                         options.splicing or \
                         options.track
    
    if pipe.assembly_method == 'transcriptome':
        if not no_specific_targets:
            use_r2g_bam = options.fusion
            
        mainMakeFile = runTranscriptome(pipe, use_ref_genome=use_ref_genome, use_r2g_bam=use_r2g_bam)
    elif pipe.assembly_method == 'stranded':
        if not no_specific_targets:
            use_r2g_bam = options.fusion
            
        lib_name = pipe.meta_lib
        mainMakeFile = runStrandedTranscriptome(pipe, use_ref_genome=use_ref_genome, use_r2g_bam=use_r2g_bam)
    elif pipe.assembly_method == 'genome' or pipe.assembly_method == 'targeted_genome':
        if not no_specific_targets:
            use_r2g_bam = options.fusion or options.indel

        mainMakeFile = runGenome(pipe, use_ref_genome=use_ref_genome, use_r2g_bam=use_r2g_bam)
    else:
        print_message('ERROR', 'unknown pipe', quit=True)
    
    if mainMakeFile:
        targets = 'all'
        if not no_specific_targets:
            targets_list = []
            
            if options.track:
                targets_list.append('tracks/%s.merged.best.unique.m90.gmap.psl.gz' % lib_name)
            if options.indel:
                targets_list.append('indels/INDELS.COMPLETE')
            if options.fusion:
                targets_list.append('fusions/FUSIONS.COMPLETE')
            if options.splicing:
                targets_list.append('splicing/SPLICING.COMPLETE')
            if options.align_contigs:
                targets_list.append('contigs_to_genome/C2G.COMPLETE')
            if options.align_reads:
                targets_list.append('reads_to_contigs/R2C.COMPLETE')
            if options.prepare_reads:
                if pipe.assembly_method == 'stranded':
                    targets_list.extend(['reads_to_contigs/plus_strand/%s.in' % pipe.lib_plus, \
                                         'reads_to_contigs/minus_strand/%s.in' % pipe.lib_minus, \
                                         'reads_to_contigs/unknown/%s.in' % pipe.meta_lib]) 
                else:
                    targets_list.append('reads_to_contigs/%s.in' % lib_name)
            if options.prepare_contigs:
                targets_list.appned('contigs_to_genome/%s-contigs/input/SPLIT_FASTA.COMPLETE' % lib_name)
            if options.fem:
                targets_list.append('merge/%s-contigs.fa' % lib_name)
            
            targets = ' '.join(targets_list)
    
        jobname = '%s-qmake' % lib_name        
        
        job_content = []
        qmake_cmd = 'qmake -cwd -v PATH -inherit -- -C %s --silent -j %d -f %s %s' % (pipe.working_dir, max_parallel_make_tasks, mainMakeFile, targets)
        if pipe.cluster:
            qmake_cmd = 'ssh %s "%s"' % (pipe.cluster, qmake_cmd)        
        job_content.append(qmake_cmd)
        job = MyJob(content=job_content, name=jobname, queue=pipe.queue)                            
        job_runner.run_locally([job], jobname, logdir=qmakedir, debug=pipe.debug, workingdir=pipe.working_dir)
    
    os.chdir(cur_path)


def get_ref_genome_path(pipe):
    return pipe.settings['genomes'][get_ref_genome_name(pipe)]


def get_ref_genome_name(pipe):
    """
    Return the name of the reference genome in the configuration file for this pipe.
    
    Args:
        pipe (trans-abyss.Pipe) : the Trans-ABySS Pipe object
        
    Returns:
        (str) : name of the reference genome
    """
    if not pipe.genome:
        try:
            # get the name of reference genome from config file
            pipe.genome = pipe.settings[pipe.project]['reference']
        except KeyError:
            print_message('ERROR', 'Cannot find the name of the reference genome for project "%s" in configuration file "%s"' % (pipe.project, pipe.cfg_file), True)
    
    if pipe.genome and not pipe.settings['genomes'].has_key(pipe.genome):
        print_message('ERROR', 'Reference genome "%s" has not been set up in the [genomes] section of configuration file "%s"' % (pipe.genome, pipe.cfg_file), True)
                
    if pipe.genome:
        reference_fasta_path = pipe.settings['genomes'][pipe.genome]
        if len(reference_fasta_path) > 0:
            return pipe.genome
        else:
            print_message('ERROR', 'The path to the FASTA file for reference genome "%s" is not specified in the [genomes] section of configuration file "%s"' % (pipe.genome, pipe.cfg_file), True)
    
    return None


def get_r2g_bam(pipe):
    """
    Return the path of the reads-to-genome BAM file for this pipe.
    
    Args:
        pipe (trans-abyss.Pipe) : the Trans-ABySS Pipe object
        
    Returns:
        (str) : the path of the BAM
    """
    r2gdir = os.path.join(pipe.working_dir, 'reads_to_genome')
    if os.path.isdir(r2gdir):
        genome_bam = pipe.get_bam_file(r2gdir)
        if not genome_bam:
            message = 'Cannot find BAM file (*.bam) in directory: %s' % r2gdir
            print_message("ERROR", message, True)
        if not os.path.isfile(genome_bam + '.bai'):
            message = 'Cannot find BAM index file (*.bam.bai) in directory: %s' % r2gdir
            print_message("ERROR", message, True)            
        return genome_bam
    else:
        message = 'Directory does not exists: %s' % r2gdir
        print_message("ERROR", message, True)
    return None


def make_prepare_contigs_rule(pipe=None, prereqs=None, library_macro=None):
    """
    Return the rule for preparing contigs fasta files for contigs-to-genome alignments.
    
    Args:
        pipe          (trans-abyss.Pipe) : the Trans-ABySS Pipe object
        prereqs       ([str])            : the list of prerequisites (ie. file paths or other rule targets)
        library_macro (qmake.Marco)      : the Macro object for the library name of this pipe
        
    Returns:
        (qmake.Rule) : the Rule object
    """

    task = 'split_fasta.py'
    args = {}
    args['FASTA'] = 'merge/$(%s)-contigs.fa' % library_macro.name
    args['OUTPUT_DIR'] = 'contigs_to_genome/$(%s)-contigs/input' % library_macro.name
    script = os.path.join(PACKAGE_DIR, 'utilities', 'split_fasta.py')
    cmds = [cfg.get_setup_cmd()]
    cmds.append('time python %s %s' % (script, cfg.create_args(pipe.settings, 'commands', task, args)) )    
    
    resources = create_resources(jobname='$(%s)-split_fasta' % library_macro.name, config_key=task, settings=pipe.settings)
    sgereq = create_sge_req_string(resources=resources) + ' ' + cfg.get_value(pipe.settings, 'commands', 'qsub')
    
    prepare_contigs = Rule(targets=['contigs_to_genome/$(%s)-contigs/input/SPLIT_FASTA.COMPLETE' % library_macro.name], \
                           prerequisites=prereqs, \
                           sgereq=sgereq.strip(), \
                           preSgeReqCmds=['mkdir -p contigs_to_genome/$(%s)-contigs/{input,cluster}' % library_macro.name], \
                           postSgeReqCmds=cmds, \
                           logfile='contigs_to_genome/$(%s)-contigs/cluster/splitfasta.log' % library_macro.name)
    return prepare_contigs


def make_calculate_assembly_stats_rule(pipe=None, prereqs=None, library_macro=None, fastas=None, threshold=200):
    """
    Return the rule for calculating assembly statistics.
    
    Args:
        pipe          (trans-abyss.Pipe) : the Trans-ABySS Pipe object
        prereqs       ([str])            : the list of prerequisites (ie. names of files, targets of other rules) to this rule
        library_macro (qmake.Macro)      : the Macro object for the library name of this pipe
        fastas        ([str])            : the list of paths of fasta files
        threshold     (int)              : the length of the shortest contig to consider for the statistics
        
    Returns:
        (qmake.Rule) : the Rule object
    """
    cmds = [cfg.get_setup_cmd()]
    cmds.append('abyss-fac -j -t %s %s > merge/stats.txt.incomplete' % (str(threshold), ' '.join(fastas)))
    cmds.append('mv merge/stats.txt.incomplete merge/stats.txt')
    
    resources = create_resources(jobname='$(%s)-stats' % library_macro.name, config_key='abyss-fac', settings=pipe.settings)
    sgereq = create_sge_req_string(resources=resources) + ' ' + cfg.get_value(pipe.settings, 'commands', 'qsub')
    
    calc_stats = Rule(targets=['merge/stats.txt'], \
                                prerequisites=prereqs, \
                                preSgeReqCmds=['mkdir -p merge/cluster'], \
                                sgereq=sgereq.strip(), \
                                postSgeReqCmds=cmds, \
                                logfile='merge/cluster/stats.log')
    return calc_stats


def make_align_contigs_rule(pipe=None, prereqs=None, library_macro=None):
    """
    Return the rule for contigs-to-genome alignments.
    A sub-make file would be created.
 
    Args:
        pipe          (trans-abyss.Pipe) : the Trans-ABySS pipe object
        prereqs       ([str])            : the list of prerequisites (ie. names of files, targets of other rules) to this rule
        library_macro (qmake.Macro)      : the Macro object for the library name of this pipe
 
    Returns:
        (qmake.Rule) : the Rule object
    """

    c2g_makefile = os.path.join('qmake', 'c2g.makefile')

    align_contigs = Rule(targets=['contigs_to_genome/C2G.COMPLETE'], \
                         prerequisites=prereqs, \
                         preSgeReqCmds=['mkdir -p contigs_to_genome/$(%s)-contigs/{output,cluster}' % library_macro.name, \
                                        '$(MAKE) -inherit -- -f ' + c2g_makefile, \
                                        'ls contigs_to_genome/C2G.COMPLETE'])
    
    # sub-make file for C2G
    c2g_makefile_components = [library_macro]
    sams = Macro(name='SAMS', \
                 definition='$(%s)' % create_patsubst_cmd('contigs_to_genome/$(' + library_macro.name + ')-contigs/input/seq.%.fa', 'contigs_to_genome/$(' + library_macro.name + ')-contigs/output/seq.%.sam'))
    c2g_makefile_components.append(sams)
    
    fastas_wildcard = 'contigs_to_genome/$(' + library_macro.name + ')-contigs/input/seq.*.fa'
    sams_wildcard = 'contigs_to_genome/$(' + library_macro.name + ')-contigs/output/seq.*.sam'
    cmds = ['if [ `ls %s |wc -l` -eq `ls %s |wc -l` ]; then touch %s; else exit 1; fi' % (fastas_wildcard, sams_wildcard, align_contigs.targets[0])]
    
    touch_c2g_complete = Rule(targets=align_contigs.targets, \
                              prerequisites=['$(%s)' % sams.name], \
                              preSgeReqCmds=cmds)
    c2g_makefile_components.append(touch_c2g_complete)
    
    script = None
    task = None
    
    args = {}    
    
    jobname = None
    if pipe.assembly_method == 'genome' or pipe.assembly_method == 'targeted_genome':
        script = 'bwa'
        task = 'c2g_bwa'
        args['TARGET'] = get_ref_genome_path(pipe)
        jobname = '$(%s)-bwamem-$*' % library_macro.name
    elif pipe.assembly_method == 'transcriptome' or pipe.assembly_method == 'stranded':
        script = 'gmap'
        task = 'gmap'        
        args['GENOME'] = get_ref_genome_name(pipe)
        jobname = '$(%s)-gmap-$*' % library_macro.name
    else:
        print_message('ERROR', 'Cannot determine assembly method.', quit=True)
    
    args['QUERY'] = '$^'
    args['OUTPUT'] = '$@.incomplete'
    cmds=[cfg.get_setup_cmd()]    
    cmds.append('time %s %s' % (script, cfg.create_args(pipe.settings, 'commands', task, args) ) )
    cmds.append('mv $@.incomplete $@')
    
    resources = create_resources(jobname=jobname, config_key=task, settings=pipe.settings)
    sgereq = create_sge_req_string(resources=resources) + ' ' + cfg.get_value(pipe.settings, 'commands', 'qsub')
        
    fasta_to_sam = Rule(targets=['contigs_to_genome/$(' + library_macro.name + ')-contigs/output/seq.%.sam'], \
                            prerequisites=['contigs_to_genome/$(' + library_macro.name + ')-contigs/input/seq.%.fa'], \
                            sgereq=sgereq.strip(), \
                            preSgeReqCmds=['mkdir -p contigs_to_genome/$(%s)-contigs/{output,cluster}' % library_macro.name], \
                            postSgeReqCmds=cmds, \
                            logfile='contigs_to_genome/$(%s)-contigs/cluster/c2g.$*.log' % library_macro.name)
    c2g_makefile_components.append(fasta_to_sam)
    
    write_items_to_file(c2g_makefile_components, c2g_makefile)
 
    return align_contigs


def make_tracks_rule(pipe=None, prereqs=None, library_macro=None):
    """
    Return the rule for UCSC custom tracks.
    
    Args:
       pipe         (trans-abyss.Pipe) : the Trans-ABySS Pipe object
       prereqs      ([str]             : the list of prerequisites (ie. names of files, targets of other rules) to this rule
       library_macro (qmake.Macro)      : the Macro object for the library name of this pipe
    """

    task = 'align_parser.py'
    script = os.path.join(PACKAGE_DIR, "utilities", task)    
    args = {}
    args['PATH'] = '.'
    args['LIB'] = '$(%s)' % library_macro.name
    args['GENOME'] = get_ref_genome_name(pipe)
    cmds=[cfg.get_setup_cmd()]
    cmds.append('time python %s %s' % (script, cfg.create_args(pipe.settings, 'commands', task, args) ))

    resources = create_resources(jobname='$(%s)-tracks' % library_macro.name, config_key=task, settings=pipe.settings)
    sgereq = create_sge_req_string(resources=resources) + ' ' + cfg.get_value(pipe.settings, 'commands', 'qsub')
    
    tracks = Rule(targets=['tracks/$(%s).merged.best.unique.m90.gmap.psl.gz' % library_macro.name], \
                                prerequisites=prereqs, \
                                sgereq=sgereq.strip(), \
                                preSgeReqCmds=['mkdir -p tracks/cluster'], \
                                postSgeReqCmds=cmds, \
                                logfile='tracks/cluster/tracks.log')
    return tracks

 
def make_splicing_rule(pipe=None, prereqs=None, library_macro=None):
    """
    Create the rule for calling splicing events.
    A sub-make file would be created.
    
    Args:
        pipe          (trans-abyss.Pipe) : the Trans-ABySS Pipe object
        prereqs       ([str])            : the list of prerequisites (ie. names of files, targets of other rules) to this rule
        library_macro (qmake.Macro)      : the Macro object for the library name of this pipe
    
    Returns:
        (qmake.Rule) : the Rule object
    """

    splicing_makefile = os.path.join('qmake', 'splicing.makefile')

    splicing = Rule(targets=['splicing/SPLICING.COMPLETE'], \
                            prerequisites=prereqs, \
                            preSgeReqCmds=['mkdir -p splicing/cluster', \
                                           '$(MAKE) -inherit -- -f ' + splicing_makefile, \
                                           'ls splicing/SPLICING.COMPLETE'])

    # sub-make for splicing
    splicing_makefile_components = [library_macro]
    splicing_tsvs = Macro(name='SPLICING_TSVS', \
                          definition='$(%s)' % create_patsubst_cmd('contigs_to_genome/$(' + library_macro.name + ')-contigs/output/seq.%.sam', 'splicing/cluster/%/events.tsv'))
    splicing_makefile_components.append(splicing_tsvs)
    
    #filter splicing events
    task = 'model_matcher.py-filter'
    script = os.path.join(PACKAGE_DIR, 'analysis', 'model_matcher.py')
    args = {}
    args['PATH'] = '.'
    args['GENOME'] = get_ref_genome_name(pipe)    
    cmds=[cfg.get_setup_cmd()]
    cmds.append('time python %s %s' % (script, cfg.create_args(pipe.settings, 'commands', task, args) ))
    cmds.append('touch %s' % splicing.targets[0])
    
    resources = create_resources(jobname='$(%s)-splicing-filter' % library_macro.name, config_key=task, settings=pipe.settings)
    sgereq = create_sge_req_string(resources=resources) + ' ' + cfg.get_value(pipe.settings, 'commands', 'qsub')
    
    filter_splicing = Rule(targets=splicing.targets, \
                                 prerequisites=['$(%s)' % splicing_tsvs.name], \
                                 sgereq=sgereq.strip(), \
                                 postSgeReqCmds=cmds, \
                                 logfile='splicing/cluster/splicing.filter.log')
    splicing_makefile_components.append(filter_splicing)
    
    #sam to splicing tsv
    task = 'model_matcher.py'
    script = os.path.join(PACKAGE_DIR, 'analysis', task)
    args = {}
    args['PATH'] = '.'
    args['LIB'] = '$(%s)' % library_macro.name
    args['GENOME'] = get_ref_genome_name(pipe)
    args['ALN'] = 'contigs_to_genome/$(%s)-contigs/output/seq.$*.sam' % library_macro.name
    args['CTG'] = 'contigs_to_genome/$(%s)-contigs/input/seq.$*.fa' % library_macro.name
    args['OUTPUT_DIR'] = 'splicing/cluster/$*'
    cmds=[cfg.get_setup_cmd()]
    cmds.append('time python %s %s' % (script, cfg.create_args(pipe.settings, 'commands', task, args) ))

    resources = create_resources(jobname='$(%s)-splicing-$*' % library_macro.name, config_key=task, settings=pipe.settings)
    sgereq = create_sge_req_string(resources=resources) + ' ' + cfg.get_value(pipe.settings, 'commands', 'qsub')
    
    sam_to_splicing_tsv = Rule(targets=['splicing/cluster/%/events.tsv'], \
                              prerequisites=['contigs_to_genome/$('+ library_macro.name +')-contigs/output/seq.%.sam'], \
                              sgereq=sgereq.strip(), \
                              preSgeReqCmds=['mkdir -p splicing/cluster/$*'], \
                              postSgeReqCmds=cmds, \
                              logfile='splicing/cluster/splicing.$*.log')
    splicing_makefile_components.append(sam_to_splicing_tsv)
    
    #write sub-make for splicing
    write_items_to_file(splicing_makefile_components, splicing_makefile)

    return splicing


def make_fusions_fule(pipe=None, prereqs=None, library_macro=None, r2gbam=None):
    """
    Create the rule for calling fusions events
    A sub-make file would be created.
 
    Args:
        pipe          (trans-abyss.Pipe) : the Trans-ABySS Pipe object
        prereqs       ([str])            : the list of prerequisites (ie. names of files, targets of other rules) to this rule
        library_macro (qmake.Macro)      : the Macro object for the library name of this pipe
        r2gbam        str                : path of reads-to-genome bam

    Returns:
        (qmake.Rule) : the Rule object
    """

    fusions_makefile = os.path.join('qmake', 'fusions.makefile')

    fusions = Rule(targets=['fusions/FUSIONS.COMPLETE'], \
                   prerequisites=prereqs, \
                   preSgeReqCmds=['mkdir -p fusions/cluster', \
                                  '$(MAKE) -inherit -- -f ' + fusions_makefile, \
                                  'ls fusions/FUSIONS.COMPLETE'])
    
    # sub-make for fusions
    fusions_makefile_components = [library_macro]
    fusions_tsvs = Macro(name='FUSIONS_TSVS', \
                         definition='$(%s)' % create_patsubst_cmd('contigs_to_genome/$(' + library_macro.name + ')-contigs/output/seq.%.sam', 'fusions/cluster/%/fusions.tsv'))
    fusions_makefile_components.append(fusions_tsvs)
    
    task = 'fusion.py-filter'
    script = os.path.join(PACKAGE_DIR, 'analysis', 'fusion.py')
    args = {}
    args['PATH'] = '.'
    args['LIB'] = '$(%s)' % library_macro.name
    args['GENOME'] = get_ref_genome_name(pipe)
    cmds=[cfg.get_setup_cmd()]
    cmds.append('time python %s %s' % (script, cfg.create_args(pipe.settings, 'commands', task, args) ))
    cmds.append('touch %s' % fusions.targets[0])
    
    resources = create_resources(jobname='$(%s)-fusions-filter' % library_macro.name, config_key=task, settings=pipe.settings)
    sgereq = create_sge_req_string(resources=resources) + ' ' + cfg.get_value(pipe.settings, 'commands', 'qsub')
    
    filter_fusions = Rule(targets=fusions.targets, \
                                 prerequisites=['$(%s)' % fusions_tsvs.name], \
                                 sgereq=sgereq.strip(), \
                                 postSgeReqCmds=cmds, \
                                 logfile='fusions/cluster/fusions.filter.log')
    fusions_makefile_components.append(filter_fusions)

    task = 'fusion.py'
    script = os.path.join(PACKAGE_DIR, 'analysis', task)    
    args = {}
    args['PATH'] = '.'
    args['LIB'] = '$(%s)' % library_macro.name
    args['GENOME'] = get_ref_genome_name(pipe)
    args['GENOME_BAM'] = r2gbam
    args['INPUT'] = 'contigs_to_genome/$(%s)-contigs/output/seq.$*.sam' % library_macro.name
    args['CTG'] = 'contigs_to_genome/$(%s)-contigs/input/seq.$*.fa' % library_macro.name
    args['OUTPUT_DIR'] = 'fusions/cluster/$*'
    cmds=[cfg.get_setup_cmd()]
    cmds.append('time python %s %s' % (script, cfg.create_args(pipe.settings, 'commands', task, args) ))
                                     
    resources = create_resources(jobname='$(%s)-fusions-$*' % library_macro.name, config_key=task, settings=pipe.settings)
    sgereq = create_sge_req_string(resources=resources) + ' ' + cfg.get_value(pipe.settings, 'commands', 'qsub')
    
    sam_to_fusions_tsv = Rule(targets=['fusions/cluster/%/fusions.tsv'], \
                              prerequisites=['contigs_to_genome/$(' + library_macro.name + ')-contigs/output/seq.%.sam'], \
                              sgereq=sgereq.strip(), \
                              preSgeReqCmds=['mkdir -p fusions/cluster/$*'], \
                              postSgeReqCmds=cmds, \
                              logfile='fusions/cluster/fusions.$*.log')
    fusions_makefile_components.append(sam_to_fusions_tsv)
    write_items_to_file(fusions_makefile_components, fusions_makefile)
    
    return fusions


def make_indels_rule(pipe=None, prereqs=None, library_macro=None, r2gbam=None):
    """
    Create the rule for calling indels events.
    A sub-make file would be created.
    
    Args:
        pipe          (trans-abyss.Pipe) : the Trans-ABySS pipe
        prereqs       [str]              : a list of prerequisites (ie. names of files, targets of other rules) to this rule
        library_macro (qmake.Macro)      : a Macro object for the library name of this pipe
        r2gbam        str                : path of reads-to-genome bam
    
    Returns:
        (qmake.Rule) : the Rule object
    """
    
    indels_makefile = os.path.join('qmake', 'indels.makefile')

    indels = Rule(targets=['indels/INDELS.COMPLETE'], \
                  prerequisites=prereqs, \
                  preSgeReqCmds=['mkdir -p indels/cluster', \
                                 '$(MAKE) -inherit -- -f ' + indels_makefile, \
                                 'ls indels/INDELS.COMPLETE'])

    # sub-make for indels
    indels_makefile_components = [library_macro]
    indels_tsvs = Macro(name='INDELS_TSVS', \
                        definition='$(%s)' % create_patsubst_cmd('contigs_to_genome/$(' + library_macro.name + ')-contigs/output/seq.%.sam', 'indels/cluster/%/events.tsv'))
    indels_makefile_components.append(indels_tsvs)
    
    # filter indels
    task = 'snv_caller.py-filter'
    script = os.path.join(PACKAGE_DIR, 'analysis', 'snv_caller.py')
    args = {}
    args['PATH'] = '.'
    args['GENOME'] = get_ref_genome_name(pipe)    
    cmds=[cfg.get_setup_cmd()]
    cmds.append('time python %s %s' % (script, cfg.create_args(pipe.settings, 'commands', task, args)))
    cmds.append('touch %s' % indels.targets[0])
    
    resources = create_resources(jobname='$(%s)-indels-filter' % library_macro.name, config_key=task, settings=pipe.settings)
    sgereq = create_sge_req_string(resources=resources) + ' ' + cfg.get_value(pipe.settings, 'commands', 'qsub')
        
    filter_indels = Rule(targets=indels.targets, \
                                 prerequisites=['$(%s)' % indels_tsvs.name], \
                                 sgereq=sgereq.strip(), \
                                 postSgeReqCmds=cmds, \
                                 logfile='indels/cluster/indels.filter.log')
    indels_makefile_components.append(filter_indels)
    
    # sam to indels tsv
    task = 'snv_caller.py'
    script = os.path.join(PACKAGE_DIR, 'analysis', task)    
    args = {}
    args['PATH'] = '.'
    args['LIB'] = '$(%s)' % library_macro.name
    args['GENOME'] = get_ref_genome_name(pipe)
    
    if pipe.assembly_method == 'genome' or pipe.assembly_method == 'targeted_genome':
        args['GENOME_BAM'] = r2gbam
    
    args['ALN'] = 'contigs_to_genome/$(%s)-contigs/output/seq.$*.sam' % library_macro.name
    args['CTG'] = 'contigs_to_genome/$(%s)-contigs/input/seq.$*.fa' % library_macro.name
    args['OUTPUT_DIR'] = 'indels/cluster/$*'
    cmds = [cfg.get_setup_cmd()]
    cmds.append('time python %s %s' % (script, cfg.create_args(pipe.settings, 'commands', task, args) ))
    
    resources = create_resources(jobname='$(%s)-indels-$*' % library_macro.name, config_key=task, settings=pipe.settings)
    sgereq = create_sge_req_string(resources=resources) + ' ' + cfg.get_value(pipe.settings, 'commands', 'qsub')
    
    sam_to_indels_tsv = Rule(targets=['indels/cluster/%/events.tsv'], \
                              prerequisites=['contigs_to_genome/$('+ library_macro.name +')-contigs/output/seq.%.sam'], \
                              sgereq=sgereq.strip(), \
                              preSgeReqCmds=['mkdir -p indels/cluster/$*'], \
                              postSgeReqCmds=cmds, \
                              logfile='indels/cluster/indels.$*.log')
    indels_makefile_components.append(sam_to_indels_tsv)
    write_items_to_file(indels_makefile_components, indels_makefile)
    
    return indels


def make_prepare_reads_rules(pipe=None):
    """
    Create the rules for converting read files to FASTQ format.
    
    Args:
        pipe (trans-abyss.Pipe) : the Trans-ABySS pipe
    Returns:
        ([Rule]) : a list of Rule objects
    """
    components = []
    args = {}
    args['INPUT_READSLIST'] = 'assembly/in' 
    args['OUTPUT_READSLIST'] = 'reads_to_contigs/%s.in' % pipe.lib
    args['LIB'] = pipe.lib
    args['OUTPUTDIR'] = 'reads_to_contigs/reads'
    args['JOBSCRIPTDIR'] = 'reads_to_contigs/cluster'
    args['PROJECT'] = pipe.project    
    args_str = ' --double-dollar-sign '
    if pipe.debug:
        args_str += ' --debug '    
    if pipe.local:
        args_str += ' --local '
    args_str += cfg.create_args(pipe.settings, 'commands', 'prepare_reads.py', args)
    
    [prep_reads_jobs, cleanup_reads_job] = prepare_reads.parseArgsFromCommand(args_str)
    for j in prep_reads_jobs:
        prerequisites = []
        if j.predecessors:
            for p in j.predecessors:
                if p.outputfiles:
                    prerequisites.append(p.outputfiles[-1])
        
        resources = create_resources(jobname=j.name, config_key=j.config_key, settings=pipe.settings)
        sgereq = create_sge_req_string(resources=resources) + ' ' + cfg.get_value(pipe.settings, 'commands', 'qsub')
        
        cmds = [cfg.get_setup_cmd()]
        cmds.extend(j.content)
        
        components.append( Rule(targets=[j.outputfiles[-1]], \
                                prerequisites=prerequisites, \
                                sgereq=sgereq.strip(), \
                                postSgeReqCmds=cmds, \
                                logfile='reads_to_contigs/cluster/%s.log' % j.name) )
    return components


def make_r2c_rule(pipe=None, prereqs=None, library_macro=None):
    """
    Create the rule for reads-to-contigs alignments.
    A sub-make file would be created WHEN THIS RULE IS RUN.
    
    Args:
        pipe          (trans-abyss.Pipe) : the Trans-ABySS pipe
        prereqs       [str]              : a list of prerequisites (ie. names of files, targets of other rules) to this rule
        library_macro (qmake.Macro)      : a Macro object for the library name of this pipe
    Returns:
        (qmake.Rule) : the Rule object
    """

    r2c_makefile = os.path.join('qmake', 'r2c.makefile')

    task = 'r2c.py'
    script = os.path.join(PACKAGE_DIR, "utilities", task)

    args = {}
    args['CONTIGS'] = 'merge/$(%s)-contigs.fa' % library_macro.name
    args['READSLIST'] = 'reads_to_contigs/$(%s).in' % library_macro.name
    args['LIB'] = '$(%s)' % library_macro.name
    args['OUTPUTDIR'] = 'reads_to_contigs'
    args['JOBSCRIPTDIR'] = 'reads_to_contigs/cluster'
    args['PROJECT'] = pipe.project
    args['MINK'] = min(pipe.ks.keys())
    args_str = cfg.create_args(pipe.settings, 'commands', task, args)

    create_r2c_makefile_cmd = 'python %s --qmake %s %s' % (script, r2c_makefile, args_str)
    align_reads = Rule(targets=['reads_to_contigs/R2C.COMPLETE'], \
                       prerequisites=prereqs, \
                       preSgeReqCmds=['mkdir -p reads_to_contigs/cluster', \
                                      cfg.get_setup_cmd(), \
                                      create_r2c_makefile_cmd, \
                                      '$(MAKE) -inherit -- -f ' + r2c_makefile, \
                                      'ls reads_to_contigs/R2C.COMPLETE'])
    return align_reads
    

def runGenome(pipe, use_ref_genome=True, use_r2g_bam=True):
    """
    Create the makefiles for running the Genome pipeline.
    
    Args:
        pipe           (trans-abyss.PipeGenome) : the Trans-ABySS genome pipe object
        use_ref_genome (bool)                   : the pipeline will use the reference genome
        use_r2g_bam    (bool)                   : the pipeline will use the reads-to-genome BAM file
    Returns:
        (str) : path of the driver makefile
    """

    qmakedir = os.path.join(pipe.working_dir, 'qmake')
    main_makefile_components = []
    makeAllPrereqs = []
    
    library = Macro(name='LIBRARY', \
                    definition=pipe.lib)
    
    if not pipe.ks:
        pipe.ks = pipe.get_ks(os.path.join(pipe.working_dir, "assembly"))
        
    if pipe.assembly_method == 'targeted_genome':
        # -0    merge contigs and indels
        # 1. get the kmers
        kmers = Macro(name='KMERS', \
                      definition='$(%s)' % create_patsubst_cmd('assembly/k%/', 'filter/k%/$(' + library.name + ')-contigs.fa'), \
                      immediate=True)
        main_makefile_components.append(kmers)
        
        # 2. cat indels and contigs
        cmds = []
        cmds.append('cat $^/$(%s)-contigs.fa $^/$(%s)-indel.fa > $@' % (library.name, library.name) )
        
        resources={}
        resources[JOBNAME] = '$(%s)-k$*-cat' % library.name
        
        sgereq = create_sge_req_string(resources=resources) + ' ' + cfg.get_value(pipe.settings, 'commands', 'qsub')
        
        cat_indels_contigs = Rule(targets=['filter/k%/$(' + library.name + ')-contigs.fa'], \
                                    prerequisites=['assembly/k%/'], \
                                    sgereq=sgereq.strip(), \
                                    preSgeReqCmds=['mkdir -p filter/{k$*,cluster}'], \
                                    postSgeReqCmds=cmds, \
                                    logfile='filter/cluster/cat.k$*.log')
        main_makefile_components.append(cat_indels_contigs)        

        # 3. merge
        task = 'abyss-rmdups-iterative'
        script = os.path.join(PACKAGE_DIR, "wrappers", task)
        args = {}
        args['LIB'] = '$(%s)' % library.name
        args['READLENGTH'] = pipe.readlength
        args['INPUT_DIR'] = 'filter'
        args['OUTPUT_DIR'] = 'merge'
        args['K'] = max(pipe.ks.keys())
        cmds = [cfg.get_setup_cmd()]
        cmds.append('time %s %s' % (script, cfg.create_args(pipe.settings, 'commands', task, args) ) )

        resources = create_resources(jobname='$(%s)-merge' % library.name, config_key=task, settings=pipe.settings)
        sgereq = create_sge_req_string(resources=resources) + ' ' + cfg.get_value(pipe.settings, 'commands', 'qsub')
        
        merge_contigs = Rule(targets=['merge/$(%s)-contigs.fa' % library.name], \
                             prerequisites=['$(%s)' % kmers.name], \
                             sgereq=sgereq.strip(), \
                             preSgeReqCmds=['mkdir -p merge/cluster'], \
                             postSgeReqCmds=cmds, \
                             logfile='merge/cluster/merge.log')
        main_makefile_components.append(merge_contigs)
        makeAllPrereqs.extend(merge_contigs.targets)
    else:
        # -0    extend indels, merge overlapping indels, merge contigs and indels
        # 1. get the kmers
        kmers = Macro(name='KMERS', \
                      definition='$(%s)' % create_patsubst_cmd('assembly/k%/', 'merge/indel_bubbles/k%/$(' + library.name + ')-b.fa'), \
                      immediate=True)
        main_makefile_components.append(kmers)

        # 1. extend the indels for each k
        extension_rules = []
        
        task = 'abyss-ta-filter'
        script = os.path.join(PACKAGE_DIR, 'wrappers', task)
                
        args = {}
        args['LIB'] = '$(%s)' % library.name
        args['READLENGTH'] = pipe.readlength
        args['INPUT_DIR'] = '$^'
        args['OUTPUT_DIR'] = 'merge/indel_bubbles/k$*'
        args['K'] = '$*'
        
        cluster_dir = "merge/indel_bubbles/cluster"
        
        cmds = [cfg.get_setup_cmd()]
        cmds.append('time %s %s -b' % (script, cfg.create_args(pipe.settings, 'commands', task, args) ))
        
        resources = create_resources(jobname='$('+ library.name + ')-k$*-extend', config_key=task, settings=pipe.settings)
        sgereq = create_sge_req_string(resources=resources) + ' ' + cfg.get_value(pipe.settings, 'commands', 'qsub')
    
        extend_indels = Rule(targets=['merge/indel_bubbles/k%/$(' + library.name + ')-b.fa'], \
                                prerequisites=['assembly/k%/'], \
                                sgereq=sgereq.strip(), \
                                preSgeReqCmds=['mkdir -p %s %s' % (cluster_dir, args['OUTPUT_DIR'])], \
                                postSgeReqCmds=cmds, \
                                logfile= os.path.join(cluster_dir, 'k$*.extend.log'))
        extension_rules.append(extend_indels)
        
        main_makefile_components.extend(extension_rules)
        
        # 2. merge overlapping/redundant extended indels
        task = 'abyss-rmdups-iterative'
        script = os.path.join(PACKAGE_DIR, "wrappers", task)

        args['INPUT_DIR'] = 'merge/indel_bubbles'
        args['OUTPUT_DIR'] = 'merge/indel_bubbles'
        args['LIB'] = '$(%s)' % library.name
        args['K'] = max(pipe.ks.keys())
        
        #print str(pipe.ks)
        
        pe_contigs = 'assembly/k%s/%s-contigs.fa' % (str(max(pipe.ks.keys())), pipe.lib)
        
        if not os.path.isfile(pe_contigs):
            print_message("ERROR", "Cannot find %s" % pe_contigs, True)
            
        merged_indels = '%s/$(%s)-b.fa' % (args['OUTPUT_DIR'], library.name)
        contigs_indels_cat = 'merge/$(%s)-cat.fa' % library.name
                
        cmds = [cfg.get_setup_cmd()]
        #Use the '-s b' option to merge overlapping/redundant only *-b.fa
        cmds.append('time %s -s b %s' % (script, cfg.create_args(pipe.settings, 'commands', task, args)))
        cmds.append('sed -e "s/>/>indel_/g" %s |cat - %s > %s' % (merged_indels, pe_contigs, contigs_indels_cat))
                
        # 3. merge indels and contigs
        args['INPUT_DIR'] = contigs_indels_cat
        args['OUTPUT_DIR'] = 'merge'
        args['LIB'] = '$(%s)' % library.name
        args['K'] = max(pipe.ks.keys())
        
        #Use the '--no-ol' option to NOT merge overlapping contigs/indels; sequences contained in a longer sequences would be removed, however
        cmds.append('time %s --no-ol %s' % (script, cfg.create_args(pipe.settings, 'commands', task, args)))
        cmds.append('rm -f %s' % contigs_indels_cat)

        resources = create_resources(jobname='$(%s)-merge' % library.name, config_key=task, settings=pipe.settings)
        sgereq = create_sge_req_string(resources=resources) + ' ' + cfg.get_value(pipe.settings, 'commands', 'qsub')
        
        merge_contigs = Rule(targets=['%s/$(%s)-contigs.fa' % (args['OUTPUT_DIR'], library.name)], \
                             prerequisites=['$(' + kmers.name + ')'], \
                             preSgeReqCmds=['mkdir -p merge/cluster'], \
                             sgereq=sgereq.strip(), \
                             postSgeReqCmds=cmds, \
                             logfile='merge/cluster/merge.log')
        main_makefile_components.append(merge_contigs)
        makeAllPrereqs.extend(merge_contigs.targets)

    # stats    
    calc_stats = make_calculate_assembly_stats_rule(pipe=pipe, \
                                                    prereqs=merge_contigs.targets, \
                                                    library_macro=library, \
                                                    fastas=['assembly/k*/*-contigs.fa', \
                                                            'merge/*-contigs.fa'], \
                                                    threshold=pipe.readlength)
    main_makefile_components.append(calc_stats)
    makeAllPrereqs.extend(calc_stats.targets)    
        
    # -R    Prepare reads for aligning to contigs
    main_makefile_components.extend( make_prepare_reads_rules(pipe) )
    
    # -r    Align reads to contigs
    prereqs = []
    prereqs.append('reads_to_contigs/%s.in' % pipe.lib)
    prereqs.extend(merge_contigs.targets)    
    align_reads = make_r2c_rule(pipe=pipe, prereqs=prereqs, library_macro=library)
    main_makefile_components.append(align_reads)
    makeAllPrereqs.extend(align_reads.targets)
    
    if use_ref_genome:    
        # -u    Prepare contigs for aligning to reference genome   
        prepare_contigs = make_prepare_contigs_rule(pipe=pipe, prereqs=merge_contigs.targets, library_macro=library)
        main_makefile_components.append(prepare_contigs)
            
        # -c    Align contigs to reference genome
        align_contigs = make_align_contigs_rule(pipe=pipe, prereqs=prepare_contigs.targets, library_macro=library)
        main_makefile_components.append(align_contigs)
        makeAllPrereqs.extend(align_contigs.targets)    
            
        if use_r2g_bam:
            r2gbam = get_r2g_bam(pipe)
            r2gbam = os.path.relpath(r2gbam, pipe.working_dir)   
        
            # -f    Call fusion events    
            prereqs = []
            prereqs.extend(align_reads.targets)
            prereqs.extend(align_contigs.targets)
            prereqs.append(r2gbam)
            fusions = make_fusions_fule(pipe=pipe, prereqs=prereqs, library_macro=library, r2gbam=r2gbam)
            main_makefile_components.append(fusions)
            makeAllPrereqs.extend(fusions.targets)
                
            # -i    Call indel events
            prereqs = []
            prereqs.extend(align_reads.targets)
            prereqs.extend(align_contigs.targets)
            prereqs.append(r2gbam)
            indels = make_indels_rule(pipe=pipe, prereqs=prereqs, library_macro=library, r2gbam=r2gbam)
            main_makefile_components.append(indels)
            makeAllPrereqs.extend(indels.targets)

    # all
    all = Rule(targets=['all'], \
               prerequisites=makeAllPrereqs)
    main_makefile_components.insert(0, all)
    
    main_makefile_components.insert(0, library)
    
    #write everything to the main make
    mainMakeFile = os.path.join(pipe.working_dir, 'qmake', 'main.makefile')
    write_items_to_file(main_makefile_components, mainMakeFile)
    
    return mainMakeFile

    
def runStrandedTranscriptome(pipe, use_ref_genome=True, use_r2g_bam=True):
    """
    Create the makefiles for running the strand-specific transcriptome pipeline.
    
    Args:
        pipe           (trans-abyss.PipeStranded) : the Trans-ABySS strand-specific transcriptome pipe object
        use_ref_genome (bool)                     : the pipeline will use the reference genome
        use_r2g_bam    (bool)                     : the pipeline will use the reads-to-genome BAM file
    Returns:
        (str) : path of the driver makefile
    """

    # Generate the make script(s) for every possible target
    # NOTE: The paths of all input/output files used in the qmake scripts are relative to the library directory!
    
    main_makefile_components = []
    makeAllPrereqs = []
    
    library = Macro(name='LIBRARY', \
                    definition=pipe.meta_lib)
                    
    library_plus = Macro(name='LIBRARY_PLUS', \
                         definition=pipe.lib_plus)

    library_minus = Macro(name='LIBRARY_MINUS', \
                          definition=pipe.lib_minus)

    if not pipe.ks_plus:
        pipe.ks_plus = pipe.get_ks(os.path.join(pipe.working_dir, "assembly", "plus_strand"))
    if not pipe.ks_minus:
        pipe.ks_minus = pipe.get_ks(os.path.join(pipe.working_dir, "assembly", "minus_strand"))

    # get the kmers from the plus assemblies
    kmers_plus = Macro(name='KMERS_PLUS', \
                       definition='$(%s)' % create_patsubst_cmd('assembly/plus_strand/k%/', 'filter/plus_strand/k%/$(' + library_plus.name + ')-contigs.fa'), \
                       immediate=True)
    main_makefile_components.append(kmers_plus)
    
    # filter extend plus assemblies
    task = 'abyss-ta-filter'
    script = os.path.join(PACKAGE_DIR, "wrappers", task)
    args = {}
    args['LIB'] = '$(%s)' % library_plus.name
    args['READLENGTH'] = pipe.readlength
    args['INPUT_DIR'] = '$^'
    args['OUTPUT_DIR'] = 'filter/plus_strand/k$*'
    args['K'] = '$*'
    cmds = [cfg.get_setup_cmd()]
    cmds.append('time %s %s' % (script, cfg.create_args(pipe.settings, 'commands', task, args) ))
    
    resources = create_resources(jobname='$(%s)-k$*-filter' % library_plus.name, config_key=task, settings=pipe.settings)
    sgereq = create_sge_req_string(resources=resources) + ' ' + cfg.get_value(pipe.settings, 'commands', 'qsub')
    
    filter_extend_contigs_plus = Rule(targets=['filter/plus_strand/k%/$(' + library_plus.name + ')-contigs.fa'], \
                                prerequisites=['assembly/plus_strand/k%/'], \
                                sgereq=sgereq.strip(), \
                                preSgeReqCmds=['mkdir -p filter/plus_strand/{k$*,cluster}'], \
                                postSgeReqCmds=cmds, \
                                logfile='filter/plus_strand/cluster/filter.k$*.log')
    main_makefile_components.append(filter_extend_contigs_plus)
    
    # merge plus
    task = 'abyss-rmdups-iterative'
    script = os.path.join(PACKAGE_DIR, "wrappers", task)
    args = {}
    args['LIB'] = '$(%s)' % library_plus.name
    args['READLENGTH'] = pipe.readlength
    args['INPUT_DIR'] = 'filter/plus_strand'
    args['OUTPUT_DIR'] = 'merge/plus_strand'
    args['K'] = max(pipe.ks_plus.keys())
    cmds = [cfg.get_setup_cmd()]
    cmds.append('time %s -p plus_ %s' % (script, cfg.create_args(pipe.settings, 'commands', task, args) ) )
    #cmds.append('sed -e "s/>/>plus_/g" merge/plus_strand/$(%s)-contigs.fa' % library_plus.name)

    resources = create_resources(jobname='$(%s)-merge-plus' % library_plus.name, config_key=task, settings=pipe.settings)
    sgereq = create_sge_req_string(resources=resources) + ' ' + cfg.get_value(pipe.settings, 'commands', 'qsub')
    
    merge_contigs_plus = Rule(targets=['merge/plus_strand/$(' + library_plus.name + ')-contigs.fa'], \
                              prerequisites=['$(%s)' % kmers_plus.name], \
                              sgereq=sgereq.strip(), \
                              preSgeReqCmds=['mkdir -p merge/plus_strand/cluster'], \
                              postSgeReqCmds=cmds, \
                              logfile='merge/plus_strand/cluster/merge.log')
    main_makefile_components.append(merge_contigs_plus)
    #makeAllPrereqs.extend(merge_contigs_plus.targets)
    

    # get the kmers from minus assemblies
    kmers_minus = Macro(name='KMERS_MINUS', \
                  definition='$(%s)' % create_patsubst_cmd('assembly/minus_strand/k%/', 'filter/minus_strand/k%/$(' + library_minus.name + ')-contigs.fa'), \
                  immediate=True)
    main_makefile_components.append(kmers_minus)
    
    # filter extend minus assemblies
    task = 'abyss-ta-filter'
    script = os.path.join(PACKAGE_DIR, "wrappers", task)
    args = {}
    args['LIB'] = '$(%s)' % library_minus.name
    args['READLENGTH'] = pipe.readlength
    args['INPUT_DIR'] = '$^'
    args['OUTPUT_DIR'] = 'filter/minus_strand/k$*'
    args['K'] = '$*'
    cmds = [cfg.get_setup_cmd()]
    cmds.append('time %s %s' % (script, cfg.create_args(pipe.settings, 'commands', task, args) ))
    
    resources = create_resources(jobname='$(%s)-k$*-filter' % library_minus.name, config_key=task, settings=pipe.settings)
    sgereq = create_sge_req_string(resources=resources) + ' ' + cfg.get_value(pipe.settings, 'commands', 'qsub')
    
    filter_extend_contigs_minus = Rule(targets=['filter/minus_strand/k%/$(' + library_minus.name + ')-contigs.fa'], \
                                prerequisites=['assembly/minus_strand/k%/'], \
                                sgereq=sgereq.strip(), \
                                preSgeReqCmds=['mkdir -p filter/minus_strand/{k$*,cluster}'], \
                                postSgeReqCmds=cmds, \
                                logfile='filter/minus_strand/cluster/filter.k$*.log')
    main_makefile_components.append(filter_extend_contigs_minus)
    
    # merge minus
    task = 'abyss-rmdups-iterative'
    script = os.path.join(PACKAGE_DIR, "wrappers", task)
    args = {}
    args['LIB'] = '$(%s)' % library_minus.name
    args['READLENGTH'] = pipe.readlength
    args['INPUT_DIR'] = 'filter/minus_strand'
    args['OUTPUT_DIR'] = 'merge/minus_strand'
    args['K'] = max(pipe.ks_minus.keys())
    cmds = [cfg.get_setup_cmd()]
    cmds.append('time %s -p minus_ %s' % (script, cfg.create_args(pipe.settings, 'commands', task, args) ) )
    #cmds.append('sed -e "s/>/>minus_/g" merge/minus_strand/$(%s)-contigs.fa' % library_minus.name)

    resources = create_resources(jobname='$(%s)-merge-plus' % library_minus.name, config_key=task, settings=pipe.settings)
    sgereq = create_sge_req_string(resources=resources) + ' ' + cfg.get_value(pipe.settings, 'commands', 'qsub')
    
    merge_contigs_minus = Rule(targets=['merge/minus_strand/$(%s)-contigs.fa' % library_minus.name], \
                               prerequisites=['$(%s)' % kmers_minus.name], \
                               sgereq=sgereq.strip(), \
                               preSgeReqCmds=['mkdir -p merge/minus_strand/cluster'], \
                               postSgeReqCmds=cmds, \
                               logfile='merge/minus_strand/cluster/merge.log')
    main_makefile_components.append(merge_contigs_minus)
    #makeAllPrereqs.extend(merge_contigs_minus.targets)
    
    #cat merged plus and minus
    resources={}
    resources[JOBNAME] = '$(%s)-cat' % library.name
    sgereq = create_sge_req_string(resources=resources) + ' ' + cfg.get_value(pipe.settings, 'commands', 'qsub')
    
    prerequisites = []
    prerequisites.extend(merge_contigs_plus.targets)
    prerequisites.extend(merge_contigs_minus.targets)
    cmds = ['cat merge/plus_strand/$(%s)-contigs.fa merge/minus_strand/$(%s)-contigs.fa > merge/$(%s)-contigs.fa.incomplete' % (library_plus.name, library_minus.name, library.name)]
    cmds.append('mv merge/$(%s)-contigs.fa.incomplete merge/$(%s)-contigs.fa' % (library.name, library.name))
    cat_merged = Rule(targets=['merge/$(%s)-contigs.fa' % library.name], \
                                prerequisites=prerequisites, \
                                preSgeReqCmds=['mkdir -p merge/cluster'], \
                                sgereq=sgereq.strip(), \
                                postSgeReqCmds=cmds, \
                                logfile='merge/cluster/cat.log')
    main_makefile_components.append(cat_merged)
    makeAllPrereqs.extend(cat_merged.targets)
    

    # stats    
    calc_stats = make_calculate_assembly_stats_rule(pipe=pipe, \
                                                    prereqs=cat_merged.targets, \
                                                    library_macro=library, \
                                                    fastas=['assembly/*_strand/k*/*-contigs.fa', \
                                                            'merge/*_strand/*-contigs.fa', \
                                                            'merge/*-contigs.fa'], \
                                                    threshold=pipe.readlength)
    main_makefile_components.append(calc_stats)
    makeAllPrereqs.extend(calc_stats.targets)

    # -R	prepare reads
    assembly_plus_in = 'assembly/plus_strand/in'
    assembly_minus_in = 'assembly/minus_strand/in'
    
    plus_reads = []
    minus_reads = []
    unknown_reads = []

    fh = open(assembly_plus_in, 'r')
    for line in fh:
        plus_reads.append(line.strip())
    fh.close()
    
    fh = open(assembly_minus_in, 'r')
    for line in fh:
        read = line.strip()
        if read in plus_reads:
            plus_reads.remove(read)
            unknown_reads.append(read)                
        else:
            minus_reads.append(read)
    fh.close()
    
    r2c_plus_in_raw = 'reads_to_contigs/plus_strand/in.raw'
    r2c_plus_in = 'reads_to_contigs/plus_strand/$(%s).in' % library.name
    
    r2c_minus_in_raw = 'reads_to_contigs/minus_strand/in.raw'
    r2c_minus_in = 'reads_to_contigs/minus_strand/$(%s).in' % library.name
    
    r2c_unknown_in_raw = 'reads_to_contigs/unknown/in.raw'
    r2c_unknown_in = 'reads_to_contigs/unknown/$(%s).in' % library.name
    
    for outputfile in [r2c_plus_in_raw, r2c_minus_in_raw, r2c_unknown_in_raw, r2c_plus_in, r2c_minus_in, r2c_unknown_in]:
        if os.path.exists(outputfile) and os.path.islink(outputfile):
            # This is a re-run. The symlink to the file from previous version must be removed.
            os.remove(outputfile)

    fh = open(r2c_plus_in_raw, 'w')
    print "Raw reads files for the plus strand:"
    for read in plus_reads:
        fh.write(read + os.linesep)
        print read
    fh.close()

    fh = open(r2c_minus_in_raw, 'w')
    print "Raw reads files for the minus strand:"
    for read in minus_reads:
        fh.write(read + os.linesep)
        print read
    fh.close()

    fh = open(r2c_unknown_in_raw, 'w')
    print "Raw reads files for unknown strand:"
    for read in unknown_reads:
        fh.write(read + os.linesep)
        print read
    fh.close()
    
    task = 'prepare_reads.py'
    
    for strand in ['plus_strand', 'minus_strand', 'unknown']:
        args = {}
        if strand == 'plus_strand':
            args['INPUT_READSLIST'] = r2c_plus_in_raw
            args['OUTPUT_READSLIST'] = r2c_plus_in
        elif strand == 'minus_strand':
            args['INPUT_READSLIST'] = r2c_minus_in_raw
            args['OUTPUT_READSLIST'] = r2c_minus_in
        elif strand == 'unknown':
            args['INPUT_READSLIST'] = r2c_unknown_in_raw
            args['OUTPUT_READSLIST'] = r2c_unknown_in
        else:
            print_message('ERROR', 'Cannot determine strand.', quit=True)
        args['LIB'] = '$(%s)' % library.name
        args['OUTPUTDIR'] = 'reads_to_contigs/%s/reads' % strand
        args['JOBSCRIPTDIR'] = 'reads_to_contigs/%s/cluster' % strand
        args['PROJECT'] = pipe.project    
        args_str = ' --double-dollar-sign '
        if pipe.debug:
            args_str += ' --debug '    
        if pipe.local:
            args_str += ' --local '
        args_str += cfg.create_args(pipe.settings, 'commands', task, args)
        [prep_reads_jobs, cleanup_reads_job] = prepare_reads.parseArgsFromCommand(args_str)    

        for j in prep_reads_jobs:
            prerequisites = []
            if j.predecessors:
                for p in j.predecessors:
                    if p.outputfiles:
                        prerequisites.append(p.outputfiles[-1])
            
            resources = create_resources(jobname=j.name, config_key=j.config_key, settings=pipe.settings)
            sgereq = create_sge_req_string(resources=resources) + ' ' + cfg.get_value(pipe.settings, 'commands', 'qsub')
                
            cmds = [cfg.get_setup_cmd()]
            cmds.extend(j.content)
            
            main_makefile_components.append( Rule(targets=[j.outputfiles[-1]], \
                                    prerequisites=prerequisites, \
                                    preSgeReqCmds=['mkdir -p reads_to_contigs/%s/reads reads_to_contigs/%s/cluster' % (strand, strand)], \
                                    sgereq=sgereq.strip(), \
                                    postSgeReqCmds=cmds, \
                                    logfile='reads_to_contigs/%s/cluster/%s.log' % (strand, j.name)) )
    
    # -r    Align reads to contigs

    mergebams_prereq = []
    
    for strand in ['plus_strand', 'minus_strand', 'unknown']:
        r2c_makefile = os.path.join('qmake', 'r2c.%s.makefile' % strand)
    
        prereqs = []
        prereqs.append('reads_to_contigs/%s/$(%s).in' % (strand, library.name))
        
        task = 'r2c.py'
        script = os.path.join(PACKAGE_DIR, "utilities", task)

        args = {}
        contigs_fasta = None
        if strand == 'plus_strand':
            contigs_fasta = 'merge/plus_strand/$(%s)-contigs.fa' % library_plus.name
            args['MINK'] = min(pipe.ks_plus.keys())
        elif strand == 'minus_strand':
            contigs_fasta = 'merge/minus_strand/$(%s)-contigs.fa' % library_minus.name
            args['MINK'] = min(pipe.ks_minus.keys())
        elif strand == 'unknown':
            contigs_fasta = 'merge/$(%s)-contigs.fa' % library.name
            args['MINK'] = min( min(pipe.ks_plus.keys()) , min(pipe.ks_minus.keys()) )
        else:
            print_message('ERROR', 'Cannot determine strand.', quit=True)
            
        prereqs.append(contigs_fasta)
        args['CONTIGS'] = contigs_fasta
        args['READSLIST'] = 'reads_to_contigs/%s/$(%s).in' % (strand, library.name)
        args['LIB'] = '$(%s)' % library.name
        args['OUTPUTDIR'] = 'reads_to_contigs/%s' % strand
        args['JOBSCRIPTDIR'] = 'reads_to_contigs/%s/cluster' % strand
        args['PROJECT'] = pipe.project
        args_str = cfg.create_args(pipe.settings, 'commands', task, args)

        create_r2c_makefile_cmd = 'python %s --qmake %s %s' % (script, r2c_makefile, args_str)
        align_reads = Rule(targets=['reads_to_contigs/%s/R2C.COMPLETE' % strand], \
                           prerequisites=prereqs, \
                           preSgeReqCmds=['mkdir -p reads_to_contigs/%s/cluster' % strand, \
                                          cfg.get_setup_cmd(), \
                                          create_r2c_makefile_cmd, \
                                          '$(MAKE) -inherit -- -f ' + r2c_makefile, \
                                          'ls reads_to_contigs/%s/R2C.COMPLETE' % strand])
        main_makefile_components.append(align_reads)
        mergebams_prereq.extend(align_reads.targets)
    
    task = 'samtools_sort'
    cmds = [cfg.get_setup_cmd()]
    cmds.append('{ samtools view -h reads_to_contigs/unknown/$(%s)-contigs.bam && samtools view reads_to_contigs/plus_strand/$(%s)-contigs.bam && samtools view reads_to_contigs/minus_strand/$(%s)-contigs.bam; } |samtools view -hSu - |samtools sort -m 800000000 -o - $$TMPDIR > reads_to_contigs/$(%s)-contigs.bam.incomplete || { rm -f $$TMPDIR*.bam 2> /dev/null; echo \'An error has occurred.\'; exit 1;}' % (library.name, library.name, library.name, library.name))
    cmds.append('samtools index reads_to_contigs/$(%s)-contigs.bam.incomplete' % library.name)
    cmds.append('mv reads_to_contigs/$(%s)-contigs.bam.incomplete.bai reads_to_contigs/$(%s)-contigs.bam.bai' % (library.name, library.name))
    cmds.append('mv reads_to_contigs/$(%s)-contigs.bam.incomplete reads_to_contigs/$(%s)-contigs.bam' % (library.name, library.name))
    cmds.append('touch reads_to_contigs/R2C.COMPLETE')
    
    resources = create_resources(jobname='$(%s)-mergebams' % library.name, config_key=task, settings=pipe.settings)
    sgereq = create_sge_req_string(resources=resources) + ' ' + cfg.get_value(pipe.settings, 'commands', 'qsub')
    
    merge_bams = Rule(targets=['reads_to_contigs/R2C.COMPLETE'], \
                      prerequisites=mergebams_prereq, \
                      preSgeReqCmds=['mkdir -p reads_to_contigs/cluster'], \
                      sgereq=sgereq.strip(), \
                      postSgeReqCmds=cmds, \
                      logfile='reads_to_contigs/cluster/mergebams.log')
    main_makefile_components.append(merge_bams)    
    
    if use_ref_genome:
        # -u    Prepare contigs for aligning to reference genome   
        prepare_contigs = make_prepare_contigs_rule(pipe=pipe, prereqs=cat_merged.targets, library_macro=library)
        main_makefile_components.append(prepare_contigs)
            
        # -c    Align contigs to reference genome
        align_contigs = make_align_contigs_rule(pipe=pipe, prereqs=prepare_contigs.targets, library_macro=library)
        main_makefile_components.append(align_contigs)
        makeAllPrereqs.extend(align_contigs.targets)
            
        # -k    Create UCSC custom track
        tracks = make_tracks_rule(pipe=pipe, prereqs=align_contigs.targets, library_macro=library)
        main_makefile_components.append(tracks)
        makeAllPrereqs.extend(tracks.targets)
        
        # -i    Call indel events
        prereqs = []
        prereqs.extend(merge_bams.targets)
        prereqs.extend(align_contigs.targets)
        indels = make_indels_rule(pipe=pipe, prereqs=prereqs, library_macro=library)
        main_makefile_components.append(indels)
        makeAllPrereqs.extend(indels.targets)    
        
        # -x    Call splicing events
        prereqs = []
        prereqs.extend(merge_bams.targets)
        prereqs.extend(align_contigs.targets)
        splicing = make_splicing_rule(pipe=pipe, prereqs=prereqs, library_macro=library)
        main_makefile_components.append(splicing)
        makeAllPrereqs.extend(splicing.targets)    
        
        if use_r2g_bam:
            # -f    Call fusion events
            r2gbam = get_r2g_bam(pipe)
            r2gbam = os.path.relpath(r2gbam, pipe.working_dir)    
            prereqs = []
            prereqs.extend(merge_bams.targets)
            prereqs.extend(align_contigs.targets)
            prereqs.append(r2gbam)
            fusions = make_fusions_fule(pipe=pipe, prereqs=prereqs, library_macro=library, r2gbam=r2gbam)
            main_makefile_components.append(fusions)
            makeAllPrereqs.extend(fusions.targets)        

    # all
    all = Rule(targets=['all'], \
               prerequisites=makeAllPrereqs)
    main_makefile_components.insert(0, all)

    main_makefile_components.insert(0, library_minus)
    main_makefile_components.insert(0, library_plus)
    main_makefile_components.insert(0, library)
    
    #write everything to the main make
    mainMakeFile = os.path.join(pipe.working_dir, 'qmake', 'main.makefile')
    write_items_to_file(main_makefile_components, mainMakeFile)
    return mainMakeFile        


def runTranscriptome(pipe, use_ref_genome=True, use_r2g_bam=True):
    """
    Create the makefiles for running the transcriptome pipeline.
    
    Args:
        pipe           (trans-abyss.PipeTranscriptome) : the Trans-ABySS transcriptome pipe object
        use_ref_genome (bool)                          : the pipeline will use the reference genome
        use_r2g_bam    (bool)                          : the pipeline will use the reads-to-genome BAM file
    Returns:
        (str) : path of the driver makefile
    """
    
    # Generate the make script(s) for every possible target
    # NOTE: The paths of all input/output files used in the qmake scripts are relative to the library directory!
  
    main_makefile_components = []
    makeAllPrereqs = []
    
    library = Macro(name='LIBRARY', \
                    definition=pipe.lib)
                    
    if not pipe.ks:
        pipe.ks = pipe.get_ks(os.path.join(pipe.working_dir, "assembly"))
    
    # -0    FEM
    
    # get the kmers
    kmers = Macro(name='KMERS', \
                  definition='$(%s)' % create_patsubst_cmd('assembly/k%/', 'filter/k%/$(' + library.name + ')-contigs.fa'), \
                  immediate=True)
    main_makefile_components.append(kmers)
    
    # filter extend
    task = 'abyss-ta-filter'
    script = os.path.join(PACKAGE_DIR, "wrappers", task)
    args = {}
    args['LIB'] = '$(%s)' % library.name
    args['READLENGTH'] = pipe.readlength
    args['INPUT_DIR'] = '$^'
    args['OUTPUT_DIR'] = 'filter/k$*'
    args['K'] = '$*'
    cmds = [cfg.get_setup_cmd()]
    cmds.append('time %s %s' % (script, cfg.create_args(pipe.settings, 'commands', task, args) ))
    
    resources = create_resources(jobname='$(%s)-k$*-filter' % library.name, config_key=task, settings=pipe.settings)
    sgereq = create_sge_req_string(resources=resources) + ' ' + cfg.get_value(pipe.settings, 'commands', 'qsub')
    
    filter_extend_contigs = Rule(targets=['filter/k%/$(' + library.name + ')-contigs.fa'], \
                                prerequisites=['assembly/k%/'], \
                                sgereq=sgereq.strip(), \
                                preSgeReqCmds=['mkdir -p filter/{k$*,cluster}'], \
                                postSgeReqCmds=cmds, \
                                logfile='filter/cluster/filter.k$*.log')
    main_makefile_components.append(filter_extend_contigs)

    # merge
    task = 'abyss-rmdups-iterative'
    script = os.path.join(PACKAGE_DIR, "wrappers", task)
    args = {}
    args['LIB'] = '$(%s)' % library.name
    args['READLENGTH'] = pipe.readlength
    args['INPUT_DIR'] = 'filter'
    args['OUTPUT_DIR'] = 'merge'
    args['K'] = max(pipe.ks.keys())
    cmds = [cfg.get_setup_cmd()]
    cmds.append('time %s %s' % (script, cfg.create_args(pipe.settings, 'commands', task, args) ) )

    resources = create_resources(jobname='$(%s)-merge' % library.name, config_key=task, settings=pipe.settings)
    sgereq = create_sge_req_string(resources=resources) + ' ' + cfg.get_value(pipe.settings, 'commands', 'qsub')
    
    merge_contigs = Rule(targets=['merge/$(%s)-contigs.fa' % library.name], \
                                prerequisites=['$(%s)' % kmers.name], \
                                sgereq=sgereq.strip(), \
                                preSgeReqCmds=['mkdir -p merge/cluster'], \
                                postSgeReqCmds=cmds, \
                                logfile='merge/cluster/merge.log')
    main_makefile_components.append(merge_contigs)
    makeAllPrereqs.extend(merge_contigs.targets)

    # stats    
    calc_stats = make_calculate_assembly_stats_rule(pipe=pipe, \
                                                    prereqs=merge_contigs.targets, \
                                                    library_macro=library, \
                                                    fastas=['assembly/k*/*-contigs.fa', \
                                                            'merge/*-contigs.fa'], \
                                                    threshold=pipe.readlength)
    main_makefile_components.append(calc_stats)
    makeAllPrereqs.extend(calc_stats.targets)
    
    # -R    Prepare reads for aligning to contigs
    main_makefile_components.extend( make_prepare_reads_rules(pipe) )
    
    # -r    Align reads to contigs
    prereqs = []
    prereqs.append('reads_to_contigs/%s.in' % pipe.lib)
    prereqs.extend(merge_contigs.targets)
        
    align_reads = make_r2c_rule(pipe=pipe, prereqs=prereqs, library_macro=library)
    main_makefile_components.append(align_reads)
    makeAllPrereqs.extend(align_reads.targets)
    
    if use_ref_genome:
        # -u    Prepare contigs for aligning to reference genome   
        prepare_contigs = make_prepare_contigs_rule(pipe=pipe, prereqs=merge_contigs.targets, library_macro=library)
        main_makefile_components.append(prepare_contigs)
            
        # -c    Align contigs to reference genome
        align_contigs = make_align_contigs_rule(pipe=pipe, prereqs=prepare_contigs.targets, library_macro=library)
        main_makefile_components.append(align_contigs)
        makeAllPrereqs.extend(align_contigs.targets)
            
        # -k    Create UCSC custom track
        tracks = make_tracks_rule(pipe=pipe, prereqs=align_contigs.targets, library_macro=library)
        main_makefile_components.append(tracks)
        makeAllPrereqs.extend(tracks.targets)
                
        # -i    Call indel events
        prereqs = []
        prereqs.extend(align_reads.targets)
        prereqs.extend(align_contigs.targets)
        indels = make_indels_rule(pipe=pipe, prereqs=prereqs, library_macro=library)
        main_makefile_components.append(indels)
        makeAllPrereqs.extend(indels.targets)    
        
        # -x    Call splicing events
        prereqs = []
        prereqs.extend(align_reads.targets)
        prereqs.extend(align_contigs.targets)
        splicing = make_splicing_rule(pipe=pipe, prereqs=prereqs, library_macro=library)
        main_makefile_components.append(splicing)
        makeAllPrereqs.extend(splicing.targets)

        if use_r2g_bam:
            # -f    Call fusion events
            r2gbam = get_r2g_bam(pipe)
            r2gbam = os.path.relpath(r2gbam, pipe.working_dir)    
            prereqs = []
            prereqs.extend(align_reads.targets)
            prereqs.extend(align_contigs.targets)
            prereqs.append(r2gbam)
            fusions = make_fusions_fule(pipe=pipe, prereqs=prereqs, library_macro=library, r2gbam=r2gbam)
            main_makefile_components.append(fusions)
            makeAllPrereqs.extend(fusions.targets)    

    # all
    all = Rule(targets=['all'], \
               prerequisites=makeAllPrereqs)
    main_makefile_components.insert(0, all)
    
    main_makefile_components.insert(0, library)
    
    #write everything to the main make
    mainMakeFile = os.path.join(pipe.working_dir, 'qmake', 'main.makefile')
    write_items_to_file(main_makefile_components, mainMakeFile)
    return mainMakeFile
