#!/usr/bin/env python

# written by:    Ka Ming Nip
# updated on:    September 3, 2013
# compatibility: Trans-ABySS 1.4.8

import os, sys, string, re, subprocess, argparse
from utilities import cfg, job_runner
from utilities.job_runner import MyJob

try:
    TRANSABYSS_PATH = os.environ['TRANSABYSS_PATH']
except KeyError:
    sys.stderr.write("ERROR: Cannot find TRANSABYSS_PATH in env!\n")
    sys.exit(1)

fileformats={}
fileformats['sam']=[]
fileformats['bam']=[]
fileformats['qseq']=['gz','bz2'] #qseq
fileformats['txt']=['gz','bz2'] #qseq
fileformats['fq']=['gz','bz2'] #fastq
fileformats['fastq']=['gz','bz2'] #fastq

convertreads_config_key = "convert_reads"
samtofastq_config_key = "samtofastq.jar"
qseq2fastq_config_key = "qseq2fastq.py"
qseq2fastq_script = "%s/utilities/qseq2fastq.py" % TRANSABYSS_PATH

ALIGNER_ABYSSMAP = 'abyssmap'
ALIGNER_BWA = 'bwa'
ALIGNER_BWASW = 'bwasw'
ALIGNER_BWAMEM = 'bwamem'
ALIGNER_BOWTIE2 = 'bowtie2'
ALIGNER_FASTQ = 'fastq' # make fastq files, don't care about which aligner it is
ALIGNERS = [ALIGNER_ABYSSMAP, ALIGNER_BWA, ALIGNER_BWASW, ALIGNER_BWAMEM, ALIGNER_BOWTIE2, ALIGNER_FASTQ]

#Given the path, determine the format and the type of compression
def determine_file_format(path):
    formats = fileformats.keys()
    path_lowered = path.lower()

    for format in formats:
        compressions = fileformats[format]
        if path_lowered.endswith(format.lower()):
            return (format, None)
        elif compressions and len(compressions) > 0:
            for compression in compressions:                
                if path_lowered.endswith(format.lower() + '.' + compression.lower()):
                    return (format, compression)
    return (None, None)


#Given the path, format, and compression, give the path to the file's (un)compressed equivalents
def getequivalents(path, format_compression):
    (format, compression) = format_compression
    if format is None:
        return []
    
    compressions = fileformats[format]
    equivalents = []    
    
    #default: assume no compression
    extensionstring = ""
    uncompressed_prefix = path
    
    #there was compression
    if not compression is None:
        extensionstring = '.' + compression
        uncompressed_prefix = path.rstrip(extensionstring)
        equivalents.append(uncompressed_prefix)
    
    for c in compressions:
        if c != compression:
            equivalents.append(uncompressed_prefix + '.' + c)
    
    return equivalents


def generate_commands(aligner, paths, outdir, prefix, outreadslist, sortmem, local, settings, single_end=False, project=None, num_reads_per_fastq=None, double_dollar_sign=False):
    tmpdir_bash_var = '$TMPDIR'
    picard_dir_bash_var = '$PICARD_DIR'
    fastq_bash_var = '$fq'
    qseq_bash_var = '$qseq'
    if double_dollar_sign:
        # Note that bash variables in make scripts use a double-dollar sign '$$', instead of a single-dollar sign '$'
        tmpdir_bash_var = '$' + tmpdir_bash_var
        picard_dir_bash_var = '$' + picard_dir_bash_var
        fastq_bash_var = '$' + fastq_bash_var
        qseq_bash_var = '$' + qseq_bash_var

    # Determine if we want to split reads files    
    split = False
    if num_reads_per_fastq != None:
        try:
            if long(num_reads_per_fastq) > 0:
                split = True
        except ValueError:
            split = False
    
    task_counter = 1
    numtasks = len(paths) * 2
    digits = len(str(numtasks))

    jobs = []
    files_generated = []

    #cur_path = os.getcwd()

    for p in paths:
        (format, compression) = determine_file_format(p)
        #p = os.path.relpath(p, cur_path)
        if format == 'bam' or format == 'sam':
            bam_basename = os.path.basename(p)
            if split or (aligner in [ALIGNER_BWASW, ALIGNER_BWAMEM, ALIGNER_FASTQ]):
                # TASK: Split the reads in the BAM file ...
                samtofastq_job_content = []
                task_id1 = str(task_counter).zfill(digits)
                task_counter += 1
                task_id2 = task_id1
                if not single_end:
                    task_id2 = str(task_counter).zfill(digits)
                    task_counter += 1
                
                # 1. Create the picard working dir for storing the initial FASTQ files
                picard_wd = outdir + os.sep + "picard"
                if os.path.isdir(picard_wd):
                    print "WARNING: The directory '" + picard_wd + "' already exists!"
                else:
                    os.mkdir(picard_wd)
                
                args = {}
                if format == 'bam':
                    # 2. Sort the BAM file by read name
                    read_sorted_bam_prefix = bam_basename + '.readsorted'
                    read_sorted_bam = picard_wd + os.sep + task_id1 + '.' + read_sorted_bam_prefix + ".bam"
                    if local:
                            bam_prefix = picard_wd + os.sep + task_id1 + '.' + read_sorted_bam_prefix
                            samtofastq_job_content.append("samtools sort -n -m %s %s %s" % (str(sortmem), p, bam_prefix))
                    else:
                            samtofastq_job_content.append("samtools sort -n -m %s -o %s %s > %s || { rm -f %s*.bam 2> /dev/null; echo 'An error has occurred.'; exit 1;}" % (str(sortmem), p, tmpdir_bash_var, read_sorted_bam, tmpdir_bash_var))
                    files_generated.append(read_sorted_bam)
                    args['BAM'] = read_sorted_bam
                elif format == 'sam':
                    # SAM file can be used as input to samtofastq.jar
                    args['BAM'] = p
                
                # 3. Use Picard's SamToFastq.jar to get 2 FASTQ files from the BAM file
                fastq1 = picard_wd + os.sep + task_id1 + '.' + bam_basename + '_1.fq'
                fastq2 = ""
                files_generated.append(fastq1)
                if not single_end:
                    fastq2 = picard_wd + os.sep + task_id2 + '.' + bam_basename + '_2.fq'
                    files_generated.append(fastq2)
                
                args['FASTQ1'] = fastq1
                args['FASTQ2'] = fastq2
                
                samtofastq_args = cfg.create_args(settings, 'commands', samtofastq_config_key, args)
                
                # java settings, ie. java -XX:-UseGCOverheadLimit -Xmx7g -jar ...
                javacmd = cfg.get_value(settings, 'java', samtofastq_config_key)
                
                samtofastq_job_content.append("%s -jar %s/SamToFastq.jar %s" % (javacmd, picard_dir_bash_var, samtofastq_args))
                samtofastq_job_name = prefix + "-samtofastq." + task_id1
                if not single_end:
                    samtofastq_job_name += '-' + task_id2
                
                if split: 
                    samtofastq_job = MyJob(content=samtofastq_job_content, id=None, name=samtofastq_job_name, predecessors=None, config_key=samtofastq_config_key)
                    jobs.append(samtofastq_job)
                
                    # 4. Split each FASTQ into 4*num_reads_per_fastq lines per file, then rename each piece to have a "fq" postfix
                    lines_per_file = 4 * long(num_reads_per_fastq)
                    
                    split_prefix1 = outdir + os.sep + task_id1 + '.' + bam_basename + '_1.fq.'                
                    split_fastq_1_job_content = []
                    split_fastq_1_job_content.append("split -d -a 4 -l " + str(lines_per_file) + " " + fastq1 + " " + split_prefix1)                    
                    split_fastq_1_job_content.append("for fq in `ls %s????`;do mv %s %s.fq;done" % (split_prefix1, fastq_bash_var, fastq_bash_var))
                    split_fastq_1_job_name = prefix + "-getfq." + task_id1
                    outputfile = '%s/%s.COMPLETE' % (outdir, split_fastq_1_job_name)
                    split_fastq_1_job_content.append('touch %s' % outputfile)
                    split_fastq_1_job = MyJob(content=split_fastq_1_job_content, id=None, name=split_fastq_1_job_name, predecessors=[samtofastq_job], config_key=convertreads_config_key, outputfiles=[outputfile])
                    jobs.append(split_fastq_1_job)
                    files_generated.append(split_prefix1 + '????.fq')
                    
                    if not single_end:
                        split_prefix2 = outdir + os.sep + task_id2 + '.' + bam_basename + '_2.fq.'
                        split_fastq_2_job_content = []
                        split_fastq_2_job_content.append("split -d -a 4 -l " + str(lines_per_file) + " " + fastq2 + " " + split_prefix2)
                        split_fastq_2_job_content.append("for fq in `ls %s????`;do mv %s %s.fq;done" % (split_prefix2, fastq_bash_var, fastq_bash_var))
                        split_fastq_2_job_name = prefix + "-getfq." + task_id2
                        outputfile = '%s/%s.COMPLETE' % (outdir, split_fastq_2_job_name)
                        split_fastq_2_job_content.append('touch %s' % outputfile)
                        split_fastq_2_job = MyJob(content=split_fastq_2_job_content, id=None, name=split_fastq_2_job_name, predecessors=[samtofastq_job], config_key=convertreads_config_key, outputfiles=[outputfile])
                        jobs.append(split_fastq_2_job)
                        files_generated.append(split_prefix2 + '????.fq')
                else:
                    samtofastq_job_content.append('ln -sf %s %s' % (os.path.relpath(fastq1, outdir), outdir))
                    if not single_end:
                        samtofastq_job_content.append('ln -sf %s %s' % (os.path.relpath(fastq2, outdir), outdir))
                    outputfile = '%s/%s.COMPLETE' % (outdir, samtofastq_job_name)
                    samtofastq_job_content.append('touch %s' % outputfile)
                    
                    samtofastq_job = MyJob(content=samtofastq_job_content, id=None, name=samtofastq_job_name, predecessors=None, config_key=samtofastq_config_key, outputfiles=[outputfile])
                    jobs.append(samtofastq_job)                    
            else:
                # TASK: Not splitting the reads in the BAM file ...
                task_id1 = str(task_counter).zfill(digits)
                task_counter += 1
                
                # 1. Symlink the BAM
                linked_bam = outdir + os.sep + task_id1 + '.' + bam_basename
                linked_bam_index = linked_bam + '.bai'
                if os.path.exists(linked_bam):
                    print "WARNING: The file '" + linked_bam + "' already exists!"
                else:
                    os.symlink(p, linked_bam)
                files_generated.append(linked_bam)
                
        elif format == 'fastq' or format == 'fq':
            task_id = str(task_counter).zfill(digits)
            task_counter += 1
            fq_basename = os.path.basename(p)
            if split:
                # TASK: Split the FASTQ files ...
                
                # Split each FASTQ into 4*num_reads_per_fastq lines per file, then rename each piece to have a "fq" postfix
                lines_per_file = 4 * long(num_reads_per_fastq)
                
                split_fastq_job_content = []
                
                split_prefix = outdir + os.sep + task_id + '.' + fq_basename + '.'
                if compression != None:
                    if compression == 'gz':
                        split_fastq_job_content.append("zcat " + p + "| split -d -a 4 -l " + str(lines_per_file) + " - " + split_prefix)
                    elif compression == 'bz2':
                        split_fastq_job_content.append("bzcat " + p + "| split -d -a 4 -l " + str(lines_per_file) + " - " + split_prefix)
                    else:
                        sys.stderr.write("ERROR: Unknown compression format '" + compression + "'!\n")
                        sys.exit(1)
                else:
                    split_fastq_job_content.append("split -d -a 4 -l " + str(lines_per_file) + " " + p + " " + split_prefix)
                split_fastq_job_content.append("for fq in `ls %s????`;do mv %s %s.fq;done" % (split_prefix, fastq_bash_var, fastq_bash_var))
                
                split_fastq_job_name = prefix + "-getfq." + task_id
                outputfile = '%s/%s.COMPLETE' % (outdir, split_fastq_job_name)
                split_fastq_job_content.append('touch %s' % outputfile)
                
                split_fastq_job = MyJob(content=split_fastq_job_content, id=None, name=split_fastq_job_name, predecessors=None, config_key=convertreads_config_key, outputfiles=[outputfile])
                jobs.append(split_fastq_job)
                files_generated.append(split_prefix + '????.fq')
            elif compression and compression == 'bz2' and aligner in [ALIGNER_BWA, ALIGNER_BWASW, ALIGNER_BWAMEM]:
                # bwa cannot read bzip fastq
                # convert bzip fastq to gzip fastq
                job_content = []
                
                fastq_converted = outdir + os.sep + task_id + '.' + fq_basename + '.fq.gz'
                job_content.append("bzcat " + p + " | gzip -c > " + fastq_converted)
                
                job_name = prefix + "-getfq." + task_id
                outputfile = '%s/%s.COMPLETE' % (outdir, job_name)
                job_content.append('touch %s' % outputfile)
                
                fastq_job = MyJob(content=job_content, id=None, name=job_name, predecessors=None, config_key=convertreads_config_key, outputfiles=[outputfile])
                jobs.append(fastq_job)
                files_generated.append(fastq_converted)
            else:
                # TASK: Not splitting the FASTQ files, no conversion needed ...
                
                # Symlink the FASTQ file
                linked_fq = outdir + os.sep + task_id + '.' + fq_basename
                if os.path.exists(linked_fq):
                    print "WARNING: The file '" + linked_fq + "' already exists!"
                else:
                    os.symlink(p, linked_fq)                
                files_generated.append(linked_fq)
        elif format == 'txt' or format == 'qseq':
            task_id = str(task_counter).zfill(digits)
            task_counter += 1
            qseq_basename = os.path.basename(p)            
            
            if aligner == ALIGNER_BOWTIE2 or aligner == ALIGNER_ABYSSMAP:
                if split:
                    # TASK: Split the QSEQ file...
                
                    split_qseq_job_content = []
                    split_prefix = outdir + os.sep + task_id + '.' + qseq_basename + '.'
                    if compression != None:
                        if compression == 'bz2':
                            split_qseq_job_content.append("bzcat " + p + " | split -d -a 4 -l " + num_reads_per_fastq + " " + split_prefix)
                        elif compression == 'gz':
                            split_qseq_job_content.append("zcat " + p + " | split -d -a 4 -l " + num_reads_per_fastq + " " + split_prefix)
                        else:
                            sys.stderr.write("ERROR: Unknown compression format '" + compression + "'!\n")
                            sys.exit(1)                        
                    else:
                        split_qseq_job_content.append("split -d -a 4 -l " + num_reads_per_fastq + " " + p + " " + split_prefix)
                    
                    split_fastq_job_content.append("for qseq in `ls %s????`;do mv %s %s.txt;done" % (split_prefix, qseq_bash_var, qseq_bash_var))
                    
                    split_qseq_job_name = prefix + "-getqseq." + task_id
                    outputfile = '%s/%s.COMPLETE' % (outdir, split_qseq_job_name)
                    split_fastq_job_content.append('touch %s' % outputfile)
                    
                    split_qseq_job = MyJob(content=split_qseq_job_content, id=None, name=split_qseq_job_name, predecessors=None, config_key=convertreads_config_key, outputfiles=[outputfile])
                    jobs.append(split_qseq_job)
                    files_generated.append(split_prefix + '????.txt')
                    
                else:
                    # TASK: Not splitting the QSEQ file...
                    
                    # Symlink the QSEQ file
                    linked_qseq = outdir + os.sep + task_id + '.' + qseq_basename
                    if os.path.exists(linked_qseq):
                        print "WARNING: The file '" + linked_qseq + "' already exists!"
                    else:
                        os.symlink(p, linked_qseq)                
                    files_generated.append(linked_qseq)
            else:                                                
                args = {}
                args['INPUTFILE'] = p
                args['OUTPUTDIR'] = outdir
                args['PREFIX'] = task_id + "." + qseq_basename
                
                if split:
                    args['READS_PER_FASTQ'] = num_reads_per_fastq
                
                qseq2fastq_args = cfg.create_args(settings, 'commands', qseq2fastq_config_key, args)
                split_fastq_job_content = []
                
                split_fastq_job_content.append("python %s %s" % (qseq2fastq_script, qseq2fastq_args))
                files_generated.append(outdir + os.sep + task_id + "." + qseq_basename + "*")
                
                split_fastq_job_name = prefix + "-getfq." + task_id
                outputfile = '%s/%s.COMPLETE' % (outdir, split_fastq_job_name)
                split_fastq_job_content.append('touch %s' % outputfile)
                
                split_fastq_job = MyJob(content=split_fastq_job_content, id=None, name=split_fastq_job_name, predecessors=None, config_key=qseq2fastq_config_key, outputfiles=[outputfile])
                jobs.append(split_fastq_job)
        else:
            sys.stderr.write("ERROR: Unknown format '" + format + "'!\n")
            sys.exit(1)
    
    # TASK: list (*.fastq/fq(.gz), *.bam) files, creating OUTPUT_READS_LIST
    listreads_job_name = prefix + "-listreads"
    cleanup_job_name = prefix + "-cleanup-reads"
    
    listreads_job_content = []
    cleanup_job_content = []
    # check the existence of output files
    for outputread in files_generated:
        listreads_job_content.append("ls " + outputread)
        cleanup_job_content.append("rm " + outputread)
    
    listreads_job_content.append("python " + TRANSABYSS_PATH + "/utilities/listreads.py " + outdir + " " + outreadslist)
    # every single job submitted earlier would be a predecessor
    listreads_job = MyJob(content=listreads_job_content, id=None, name=listreads_job_name, predecessors=list(jobs), config_key=None, outputfiles=[outreadslist])
    jobs.append(listreads_job)
    
    cleanup_job = MyJob(content=cleanup_job_content, id=None, name=cleanup_job_name, config_key=None)
    
    return [jobs, cleanup_job]

parser = argparse.ArgumentParser(description='Prepare reads files for a specific aligner.')
parser.add_argument('inputreadslist', metavar='INPUT_READS_LIST', type=str, help='path to text file listing the paths to input reads files')
parser.add_argument('outputreadslist', metavar='OUTPUT_READS_LIST', type=str, help='path to text file listing the paths to output reads files')
parser.add_argument('aligner', metavar='FORMAT', type=str, help='Output reads compatibility; one of %s' % str(ALIGNERS))

sample_group = parser.add_mutually_exclusive_group(required=True)
sample_group.add_argument('--genome', action='store_true', default=False, dest='genome', help='use configurations from \'genome.cfg\'')
sample_group.add_argument('--transcriptome', action='store_true', default=False, dest='transcriptome', help='use configurations from \'transcriptome.cfg\'')

parser.add_argument('-d', '--debug', dest='debug', help='run in debug mode; no jobs will be submitted', action='store_true', default=False)
parser.add_argument('-l', '--local', action='store_true', default=False, dest='local', help='run locally')
parser.add_argument('-j', '--jobdir', dest='jobdir', help='job scripts directory', metavar='DIR')
parser.add_argument('-m', '--sort-mem', type=int, default=800000000, dest='sortmem', help='value for `samtools sort -m` [default: %(default)s]', metavar='INT')    
parser.add_argument('-o', '--outputdir', dest='outdir', help='output directory', metavar='DIR')
parser.add_argument('-p', '--prefix', dest='prefix', help='library name', metavar='STR')
parser.add_argument('-P', '--project', dest='project', help='project name', metavar='STR')
parser.add_argument('-r', '--relaxed', dest='lenient', help='look for (un)compressed equivalent of non-existing paths for read files', action='store_true', default=False)
parser.add_argument('-s', '--single-end', action='store_true', default=False, dest='single_end', help='input reads files are single-end')
parser.add_argument('--double-dollar-sign', action='store_true', default=False, dest='double_dollar_sign', help='use \'$$\' for bash variables')
parser.add_argument('--partition', type=int, metavar='INT', default=None, dest='partition', help='split output files into INT number of reads per partition [default: %(default)s]')

def parseArgsFromCommand(command):
    args = parser.parse_args(command.split())
    return createJobsFromArgs(args)

def createJobsFromArgs(args):
    readslist = args.inputreadslist
    outreadslist = args.outputreadslist
    
    if not args.aligner in ALIGNERS:
        sys.stderr.write("ERROR: Unknown aligner specified, %s" % args.aligner + os.linesep)
        sys.exit(1)
            
    
    if os.path.exists(outreadslist):
        print "WARNING: The file '" + outreadslist + "' already exists!"    
    
    
    # Check if the reads list text file (*.in) exists
    if not os.path.isfile(readslist):
        sys.stderr.write("ERROR: Cannot find reads list file, " + readslist  + '\n')
        sys.exit(1)
    
    reads_paths = []
    missing_reads = []
    incorrect_format = []
 
    lenient = args.lenient
    
    # Check if the input reads files ( *.bam, *.fq/fastq(.bz2/.gz), *.txt(.bz2/.gz) ) exist
    fh = open(readslist, 'r')
    for line in fh:
        linestripped = line.strip()
        if len(linestripped) > 0:
            (format, compression) = determine_file_format(linestripped)
            if format is None:
                incorrect_format.append(linestripped)
            else:
                if os.path.isfile(linestripped):
                    reads_paths.append(linestripped)
                elif lenient:
                    equivalents = getequivalents(linestripped, (format, compression))
                    equivalent_found = None
                    for e in equivalents:
                        if os.path.isfile(e):
                            print "WARNING: Cannot find file '" + linestripped + "'"
                            print "... will use its equivalent '" + e + "'"
                            equivalent_found = e
                            break
                    if equivalent_found:
                        reads_paths.append(equivalent_found)
                    else:
                        missing_reads.append(linestripped)
                else:
                    missing_reads.append(linestripped)    
    fh.close()
    
    erronous_reads_list = False
    
    if len(missing_reads) > 0:
        erronous_reads_list = True
        sys.stderr.write("ERROR: Cannot find the following read files:\n")
        for f in missing_reads:
            sys.stderr.write(f + '\n')
    
    if len(incorrect_format) > 0:
        erronous_reads_list = True    
        sys.stderr.write("ERROR: The following read files have incorrect format:\n")
        for f in incorrect_format:
            sys.stderr.write(f + '\n')
    
    if erronous_reads_list:
        sys.exit(1)    

    outdir = args.outdir
    if outdir:
        if not os.path.isdir(outdir):
            os.makedirs(outdir)
            if not os.path.isdir(outdir):            
                sys.stderr.write("ERROR: Cannot create output directory " + outdir  + '\n')
                sys.exit(1)
    else:
        print "No output directory was specified..."
        print "Output files will be generated in the current directory..."
        outdir = os.getcwd()
        print outdir

    jobdir = args.jobdir
    if jobdir:
        if not os.path.isdir(jobdir):
            os.makedirs(jobdir)
            if not os.path.isdir(jobdir):            
                sys.stderr.write("ERROR: Cannot create cluster job working directory " + jobdir  + '\n')
                sys.exit(1)
    else:
        print "No job scripts directory was specified..."
        print "Job scripts and cluster logs will be generated in the current directory..."
        jobdir = os.getcwd()
        print jobdir

    prefix = args.prefix
    if args.prefix == None:
        prefix = os.path.basename(readslist).rstrip(".in")

    # ALL input reads files are confirmed to be found and are in the right format at this point
    # Run the commands locally or submit them as job scripts to the cluster

    try:
        sortmem = int(args.sortmem)
    except TypeError:
        sys.stderr.write("ERROR: `samtools sort` max memory specified (" + args.sortmem + ") is not an integer.")
        sys.exit(1)
        
    if sortmem < 1:
        sys.stderr.write("ERROR: `samtools sort` max memory specified (" + args.sortmem + ") is less than one.")
        sys.exit(1)

    if args.genome and args.transcriptome:
        sys.stderr.write("ERROR: Both genome and transcriptome were selected.")
        sys.exit(1)
    elif not args.genome and not args.transcriptome:
        sys.stderr.write("ERROR: Please specify either '--transcriptome' or '--genome'")
        sys.exit(1)

    config_type = None    
    if args.genome:
        config_type = "genome"
    elif args.transcriptome:
        config_type = "transcriptome"
        
    if args.debug:
        print "WARNING: Running in debug mode!"
        print "*** Only job scripts and/or symlinks will be created. ***"
        print "*** No resource-intensive commands will be run.       ***"
        print "*** No jobs will be submitted to the cluster.         ***"

    cfg_file = '%s/configs/%s.cfg' % (TRANSABYSS_PATH, config_type)
    settings = cfg.initialize_settings(cfg_file, args.project)

    # Generate the job commands
    return generate_commands(aligner=args.aligner, paths=reads_paths, project=args.project, outdir=outdir, prefix=prefix, single_end=args.single_end, outreadslist=outreadslist, sortmem=sortmem, local=args.local, settings=settings, num_reads_per_fastq=args.partition, double_dollar_sign=args.double_dollar_sign)

def __main__():
    sys.stdout.flush() #flush print outputs to screen
    version = "1.4.8 (Build 201305141141)"
    
    args = parser.parse_args()
    [jobs, cleanup_job] = createJobsFromArgs(args)

    # Write the cleanup script, but it will NOT be run.
    job_script_configs = job_runner.getJobScriptConfigs(TRANSABYSS_PATH + "/configs/job_script.cfg")
    template = TRANSABYSS_PATH + "/configs/templates/" + job_script_configs["local"]
    cleanup_script = job_runner.makeJobScript(template=template, content=cleanup_job.content, jobname=cleanup_job.name, logdir=args.jobdir, setup=cfg.get_setup_cmd())
    fh = open(args.jobdir + os.sep + cleanup_job.name + ".sh", 'w')
    fh.write(cleanup_script)
    fh.close()

    if args.local:
        job_runner.run_locally(jobs=jobs, prefix='local', logdir=args.jobdir, debug=args.debug)
    else:
        config_type = None    
        if args.genome:
            config_type = "genome"
        elif args.transcriptome:
            config_type = "transcriptome"
            
        if args.debug:
            print "WARNING: Running in debug mode!"
            print "*** Only job scripts and/or symlinks will be created. ***"
            print "*** No resource-intensive commands will be run.       ***"
            print "*** No jobs will be submitted to the cluster.         ***"
        cfg_file = '%s/configs/%s.cfg' % (TRANSABYSS_PATH, config_type)
        settings = cfg.initialize_settings(cfg_file, args.project)
        job_runner.submit_jobs(jobs=jobs, logdir=args.jobdir, debug=args.debug, settings=settings, head_node=None)

    sys.exit(0)

if __name__ == "__main__":
    __main__()
