"""
qseq2fastq.py

Original written by Richard Corbett
Modified to work with Trans-ABySS by Ka Ming Nip
Last modified on March 13, 2013
"""
import glob, sys, os, re, gzip, bz2, argparse

shift31={'@':'!','A':'"','B':'#','C':'$','D':'%','E':'&','F':'\'','G':'(','H':')','I':'*','J':'+','K':',','L':'-','M':'.','N':'/','O':'0','P':'1','Q':'2','R':'3','S':'4','T':'5','U':'6','V':'7','W':'8','X':'9','Y':':','Z':';','[':'<','\\':'=',']':'>','^':'?','_':'@','`':'A','a':'B','b':'C','c':'D','d':'E','e':'F','f':'G','g':'H','h':'I','i':'J','j':'K','k':'L','l':'M','m':'N','n':'O','o':'P','p':'Q','q':'R','r':'S','s':'T','t':'U','u':'V','v':'W','w':'X','x':'Y','y':'Z','z':'\[','\{':'\\','|':']','}':'^','~':'_'}

"""
The QSEQ format
FROM: ftp://ftp.era.ebi.ac.uk/meta/doc/sra_1_1/SRA_File_Formats_Guide.pdf

Each record is one line with tab separator in the following format:

0 - Machine name: unique identifier of the sequencer.
1 - Run number: unique number to identify the run on the sequencer.
2 - Lane number: positive integer (currently 1-8).
3 - Tile number: positive integer.
4 - X: x coordinate of the spot. Integer (can be negative).
5 - Y: y coordinate of the spot. Integer (can be negative).
6 - Index: positive integer. No indexing should have a value of 1.
7 - Read Number: 1 for single reads; 1 or 2 for paired ends.
8 - Sequence (BASES)
9 - Quality: the calibrated quality string. (QUALITIES)
10- Filter: Did the read pass filtering? 0 - No, 1 - Yes.
"""

def asciiConvert(quality,strip,striplast):

    asciList=[0]*(len(quality)-strip-striplast)
    
    for position in range(len(quality)):
        if position < strip:
            continue
        if position > len(quality) - striplast - 1:
            continue
        """
        FASTQ 33 is standard Sanger quality encoding, which means that
        you have to take the ASCII value of each character \
        in the quality string obtained from Illumina Sequencers (ascii
        64) and subtract 31 from it to get ascii 33 fastq files.
        The character is'C' == 67 and the '%' == 37. These would
        translate to Q scores of 34 and 4 which is an expected
        range of Phred scores.
        
        The quickest way to distinguish Sanger Q-score encoding
        (ASCII-33) from Illumina (Solexa) Q-score encoding (ASCII-64) \
        is to look for numerals [0-9] in the quality string. The
        numerals have ASCII values from 48-57 so it would be non-sensical
        to subtract 64 from them. If there are numerals in your quality
        string then the Q-score encoding is Sanger.
        """
        # - convert ascii 64 to ascii 33
        asciList[position]=shift31.get(quality[position], '#')
        #asciList[position] = str(unichr((ord(quality[position])-31))) # slower
        
    return asciList    


def writeFastqToDisk(cfilter, strip, striplast, filename, directory, reads_per_partition, prefix, gzip_out, name_format):
    if filename:
        if filename.endswith('.gz'):
            fh_in = gzip.open(filename, 'r')
        elif filename.endswith('.bz2'):
            fh_in = bz2.BZ2File(filename, 'r')
        else:        
            fh_in = open(filename, 'r')
        sys.stderr.write("Reading from file " + filename + " ...\n")
    else:
        fh_in = sys.stdin
        sys.stderr.write("Reading from stdin ...\n")

    outdir = directory
    if not directory:
        outdir = os.getcwd()
    
    num_partition = 1
    if reads_per_partition and reads_per_partition > 0:
        fout = outdir + '/' + prefix + '.' + str(num_partition).zfill(4) + '.fq'
    else:
        fout = outdir + '/' + prefix + '.fq'
    
    if gzip_out:
        fout = fout + '.gz'
        fh_out = gzip.open(fout, 'w', 6)
    else:    
        fh_out = open(fout, 'w')
    sys.stderr.write("Writing to file " + fout + " ...\n")
    
    lc = 0
    total_lines = 0
    num_chsty_fail = 0
        
    if cfilter:
        #chastity flag filtering
        for line in fh_in:                    
            line = line.split("\t")
            
            if line[10].rstrip() == "1":
                # pass
                lc += 1

                name = name_format.format(*line)
                converted_ascii = "".join(asciiConvert(line[9],strip,striplast))
                end_pos = len(line[8]) - striplast
                
                fh_out.write("@%s\n%s\n+\n%s\n" % (name, line[8][strip:end_pos], converted_ascii))
                
                if reads_per_partition and reads_per_partition > 0 and lc >= reads_per_partition:
                    total_lines += lc
                    lc = 0
                    reads_read = 0
                    num_partition += 1
                    fh_out.close()
                    fout = outdir + '/' + prefix + '.' + str(num_partition).zfill(4) + '.fq'
                    if gzip_out:
                        fout = fout + '.gz'
                        fh_out = gzip.open(fout, 'w', 6)
                    else:    
                        fh_out = open(fout, 'w')
                    sys.stderr.write("Writing to file " + fout + " ...\n")
            else:
                # fail
                num_chsty_fail += 1
                continue
        total_lines += lc          
    else:
        for line in fh_in:
            lc += 1
            
            line = line.split("\t")
            name = name_format.format(*line)
            converted_ascii = "".join(asciiConvert(line[9],strip,striplast))
            end_pos = len(line[8]) - striplast

            fh_out.write("@%s\n%s\n+\n%s\n" % (name, line[8][strip:end_pos], converted_ascii))
                
            if reads_per_partition > 0 and lc >= reads_per_partition:
                total_lines += lc
                lc = 0
                num_partition += 1
                fh_out.close()
                fout = outdir + '/' + prefix + '.' + str(num_partition).zfill(4) + '.fq'
                if gzip_out:
                    fout = fout + '.gz'
                    fh_out = gzip.open(fout, 'w', 6)
                else:    
                    fh_out = open(fout, 'w')
                sys.stderr.write("Writing to file " + fout + " ...\n")            
        total_lines += lc
    fh_out.close()
    sys.stderr.write("Number of reads included: %s\n" %total_lines)
    sys.stderr.write("Number of reads excluded: %s\n" %num_chsty_fail)


def writeFastqToStdout(cfilter, strip, striplast, filename, name_format):
  
    if filename:
        if filename.endswith('.gz'):
            fh_in = gzip.open(filename, 'r')
        elif filename.endswith('.bz2'):
            fh_in = bz2.BZ2File(filename, 'r')
        else:        
            fh_in = open(filename, 'r')
        sys.stderr.write("Reading from " + filename + " ...\n")
    else:
        fh_in = sys.stdin
        sys.stderr.write("Reading from stdin ...\n")
            
    total_lines = 0
    num_chsty_fail = 0
    
    sys.stderr.write("Writing to stdout ...\n")
    
    if cfilter:
        #chastity flag filtering
        for line in fh_in:                    
            line = line.split("\t")
            
            if line[10].rstrip() == "1":
                # pass
                total_lines += 1
                name = name_format.format(*line)
                converted_ascii = "".join(asciiConvert(line[9],strip,striplast))
                end_pos = len(line[8]) - striplast
                sys.stdout.write("@%s\n%s\n+\n%s\n" % (name,line[8][strip:end_pos],converted_ascii))
            else:
                # fail
                num_chsty_fail += 1
                continue            
    else:
        for line in fh_in:
            total_lines += 1
            line = line.split("\t")
            name = name_format.format(*line)
            converted_ascii = "".join(asciiConvert(line[9],strip,striplast))
            end_pos = len(line[8]) - striplast
            sys.stdout.write("@%s\n%s\n+\n%s\n" % (name,line[8][strip:end_pos],converted_ascii))
    
    sys.stderr.write("reads included: %s\n" %total_lines)
    sys.stderr.write("reads excluded: %s\n" %num_chsty_fail)


def __main__():    
    parser = argparse.ArgumentParser(description='Convert QSEQ file to FASTQ')
    
    parser.add_argument("-f", "--filter", help="filter out chastity-failed reads", dest="cfilter", default=False, action="store_true")
    parser.add_argument("-s", "--strip", metavar="INT", type=int, help="strip off first INT bases [default: %(default)s]", dest="strip", default=0, action="store")
    parser.add_argument("-l", "--striplast", metavar="INT", type=int, help="strip off last INT bases [default: %(default)s]", dest="striplast", default=0, action="store")
    parser.add_argument("-i", "--inputfile", dest="filename", type=str, metavar="FILE", help="input qseq file (can be compressed with gzip or bzip2)")
    parser.add_argument("-o", "--outdir", metavar="DIR", type=str, help="output directory", dest="directory")
    parser.add_argument("--partition", metavar="INT", type=int, dest="reads_per_partition", default=None, action="store", help="split output files into INT number of reads per partition [default: %(default)s]")
    parser.add_argument("-n", "--name", metavar="FORMAT", type=str, dest="name_format", help="format for read names in FASTQ [default: %(default)s]", default="{0}_{1}:{2}:{3}:{4}:{5}/{7}")
    parser.add_argument("-p", "--prefix", dest = "prefix", type=str, metavar="STR", help="prefix of output file(s)")
    parser.add_argument("-g", "--gzip", dest="gzip_out", action="store_true", help="compress output file(s) with gzip")
    
    args = parser.parse_args()
    
    cfilter = args.cfilter
    strip = args.strip
    striplast = args.striplast
    filename = args.filename
    directory = args.directory    
    reads_per_partition = args.reads_per_partition
    prefix = args.prefix
    gzip_out = args.gzip_out 
    name_format = args.name_format
    
    if filename:
        if os.path.isfile(filename):
            if os.path.getsize(filename) <= 0:
                sys.stderr.write("ERROR: Empty input file, " + filename)
                sys.exit(1)
        else:
            sys.stderr.write("ERROR: Cannot find input file, " + filename)
            sys.exit(1)
    
    if directory:
        if not os.path.exists(directory):
            sys.stderr.write("ERROR: Cannot find output directory, " + directory)
            sys.exit(1)        
    
    write_to_disk = directory or reads_per_partition > 0 or prefix or gzip_out      
        
    if write_to_disk:
        # we are writing output to file(s), so we need a prefix for output files
        if not prefix:
            if filename:
                # get the prefix from input file for our output file(s)
                prefix = os.path.basename(filename)                
            else:
                sys.stderr.write("ERROR: No prefix was specified for output file(s).")
                sys.exit(1)
    
    if len(name_format) <= 0:
        sys.stderr.write("ERROR: No format specified for read names in FASTQ.")
        sys.exit(1)
    sys.stderr.write("Format for read names in FASTQ: %s\n" %name_format);    
        
    sys.stderr.write("Remove chastity-failed reads: %s\n" %cfilter)
    
    if write_to_disk:
        writeFastqToDisk(cfilter, strip, striplast, filename, directory, reads_per_partition, prefix, gzip_out, name_format)
    else:
        writeFastqToStdout(cfilter, strip, striplast, filename, name_format)
    
    sys.stderr.write("Conversion completed.")

if __name__ == "__main__":
    __main__()
