#!/usr/bin/env python

# written by:    Ka Ming Nip
# updated on:    July 23, 2013
# compatibility: Trans-ABySS 1.4.8

import os, sys, string, re, subprocess, argparse
from utilities import cfg, job_runner
from utilities.job_runner import MyJob
from gsc import qmake #@TODO: REMOVE THIS FOR EXTERNAL RELEASE
from gsc.qmake import Rule #@TODO: REMOVE THIS FOR EXTERNAL RELEASE

try:
    TRANSABYSS_PATH = os.environ['TRANSABYSS_PATH']
except KeyError:
    sys.stderr.write("ERROR: Cannot find TRANSABYSS_PATH in env!" + os.linesep)
    sys.exit(1)

ALIGNER_ABYSSMAP = 'abyssmap'
ALIGNER_BWA = 'bwa'
ALIGNER_BWASW = 'bwasw'
ALIGNER_BWAMEM = 'bwamem'
ALIGNER_BOWTIE2 = 'bowtie2'
ALIGNERS = [ALIGNER_ABYSSMAP, ALIGNER_BWA, ALIGNER_BWASW, ALIGNER_BWAMEM, ALIGNER_BOWTIE2]


class R2C_Factory:
    """
    Class to generate and/or submit jobs of aligning reads to contigs.
    """

    def __init__(self, contigs, reads_paths, aligner, prefix=None, outdir=None, jobdir=None, debug=False, cleanup=False, transcriptome=False, genome=False, threads=int(4), single_end=False, local=False, xa2multi=False, strip=False, messy=False, sortmem=int(800000000), k=int(24), project=None, cluster=None, queue=None, indexed=False, samse=False, predecessors=None, makefile=None):
        """
        Args:
            contigs       (str)   : path of the contigs FASTA file
            reads_paths   ([str]) : list of the paths of read files
            aligner       (str)   : name of the aligner
            prefix        (str)   : prefix of output files
            outfir        (str)   : the path of the output directory
            jobdir        (str)   : the path of the directory for job scripts
            debug         (bool)  : flag to run in debug mode
            cleanup       (bool)  : flag to clean up intermediate files
            transcriptome (bool)  : boolean flag to use settings from the configuration file 'transcriptome.cfg'
            genome        (bool)  : flag to use settings from the configuration file 'genome.cfg'
            threads       (int)   : the maximum number of threads to use for multi-threaded programs
            single_end    (bool)  : flag input reads as single-end data
            local         (bool)  : flag to run the jobs locally instead submitting to the cluster
            xa2multi      (bool)  : flag to run xa2multi.pl when using bwa (not bwasw nor bwamem) as aligner
            strip         (bool)  : flag to strip away the quality string in output bam files
            messy         (bool)  : flag to not pipe any commands and keep all intermediate files
            sortmem       (int)   : the argument for `samtools sort -m`
            k             (int)   : the minimum k-mer size
            project       (str)   : the name of the project in the configuration file
            cluster       (str)   : the name of the cluster head node
            queue         (str)   : the name of the cluster queue
            indexed       (bool)  : flag to indicate the contigs file was already indexed
            samse         (bool)  : flag to run `bwa samse` instead of `bwa sampe` for paired-end reads
            predecessors  ([job_runner.MyJob]) : the list of predecessor jobs of all jobs to be created
            makefile      (str)   : the path of the makefile to be created; when set, no jobs would be submitted            
        """
        self.contigs = contigs
        self.reads_paths = reads_paths
        self.prefix = prefix
        self.outdir = outdir
        self.jobdir = jobdir
        self.debug = debug
        self.cleanup = cleanup
        self.transcriptome = transcriptome
        self.genome = genome
        self.threads = threads
        self.single_end = single_end
        self.local = local
        self.xa2multi = xa2multi
        self.strip = strip
        self.messy = messy
        self.sortmem = sortmem
        self.k = k
        self.project = project
        self.cluster = cluster
        self.queue = queue
        self.indexed = indexed
        self.samse = samse
        #self.index_only = index_only
        self.predecessors = predecessors
        self.aligner = aligner
        self.makefile = makefile
        self.setup_params()
        
    def setup_params(self):
        """
        Method to validate the input parameters.
        """
        if not os.path.isfile(self.contigs):
            sys.stderr.write("ERROR: Cannot find contigs FASTA file, " + self.contigs + os.linesep)
            sys.exit(1)
        
        for r in self.reads_paths:
            if not os.path.isfile(r):
                sys.stderr.write("ERROR: Cannot find reads file, " + r + os.linesep)
                sys.exit(1)            

        try:
            self.threads = int(self.threads)
        except TypeError:
            sys.stderr.write("ERROR: Number of threads specified (" + str(self.threads) + ") is not an integer." + os.linesep)
            sys.exit(1)
        
        if self.threads < 1:
            sys.stderr.write("ERROR: Number of threads specified (" + str(self.threads) + ") is less than one." + os.linesep)
            sys.exit(1)
            
        try:
            self.sortmem = int(self.sortmem)
        except TypeError:
            sys.stderr.write("ERROR: `samtools sort` max memory specified (" + str(self.sortmem) + ") is not an integer." + os.linesep)
            sys.exit(1)
            
        if self.sortmem < 1:
            sys.stderr.write("ERROR: `samtools sort` max memory specified (" + str(self.sortmem) + ") is less than one." + os.linesep)
            sys.exit(1)            

        if not self.aligner:
            sys.stderr.write("ERROR: No aligner was selected." + os.linesep)
            sys.exit(1)               
        elif not self.aligner in ALIGNERS:
            sys.stderr.write("ERROR: Unknown aligner specified, %s" % self.aligner + os.linesep)
            sys.exit(1)

        if self.genome and self.transcriptome:
            sys.stderr.write("ERROR: Both genome and transcriptome were selected." + os.linesep)
            sys.exit(1)
        elif not self.genome and not self.transcriptome:
            sys.stderr.write("ERROR: Neither genome or transcriptome were selected. Please specify either '--transcriptome' or '--genome'" + os.linesep)
            sys.exit(1)

    def qmake(self):
        """
        Method to create the makefile for qmake.
        The makefile would not be run.
        """
        #create the jobs
        (jobs, cleanup_job) = self.create_jobs()
        
        #create macros/rules for the jobs
        components = []
                
        for j in jobs:
            prerequisites = []
            if j.predecessors:
                for p in j.predecessors:
                    if p.outputfiles:
                        prerequisites.extend(p.outputfiles)
            
            resources = qmake.create_resources(jobname=j.name, config_key=j.config_key, settings=self.settings)
            
            cmds = [cfg.get_setup_cmd()]
            cmds.extend(j.content)
            
            sgereq = qmake.create_sge_req_string(resources=resources) + ' ' + cfg.get_value(self.settings, 'commands', 'qsub')
            
            components.append( Rule(targets=[j.outputfiles[-1]], \
                                    prerequisites=prerequisites, \
                                    sgereq=sgereq.strip(), \
                                    postSgeReqCmds=cmds, \
                                    logfile='%s/%s.log' % (self.jobdir, j.name)) )
        
        if not self.messy and self.cleanup and cleanup_job:
            prerequisites = []
            if cleanup_job.predecessors:
                for p in cleanup_job.predecessors:
                    if p.outputfiles:
                        prerequisites.extend(p.outputfiles)                    
            cleanup_rule = Rule(targets=['cleanup'], \
                                        prerequisites=prerequisites, \
                                        preSgeReqCmds=cleanup_job.content)
            components.append(cleanup_rule)
        else:
            print "WARNING: Clean-up script will be generated, but it will not be run. Otherwise, use the '--cleanup' option of 'r2c.py' next time."
            script = create_cleanup_script()
            print "clean-up script: " + script
        
        #write the components in reverse order
        components.reverse()
        
        #write the macros and rules to a make file
        qmake.write_items_to_file(components, self.makefile)

    def create_jobs(self):
        """
        Method to create jobs for the alignments.
        No jobs would be submitted.
        
        Returns:
            ([job_runner.MyJob], job_runner.MyJob) : a tuple of MyJob objects.
                                                     The first element of the tuple is a list of jobs to do the alignments
                                                     and generate a final indexed BAM file.
                                                     The second element of the tuple is the job for cleaning up intermediate files.
        """
    
        config_type = None    
        if self.genome:
            config_type = "genome"
        elif self.transcriptome:
            config_type = "transcriptome"
        
        contigs_base_name = os.path.basename(self.contigs)
        if not self.prefix:                  
            if contigs_base_name.endswith(".fa"):            
                if contigs_base_name.endswith("-contigs.fa"):
                    index = contigs_base_name.rindex("-contigs.fa")
                else:
                    index = contigs_base_name.rindex(".fa")
                self.prefix = contigs_base_name[0:index]
            elif contigs_base_name.endswith(".fasta"):
                if contigs_base_name.endswith("-contigs.fasta"):
                    index = contigs_base_name.rindex("-contigs.fasta")
                else:
                    index = contigs_base_name.rindex(".fasta")
                self.prefix = contigs_base_name[0:index]
            else:
                sys.stderr.write("ERROR: Cannot determine prefix. Please specify a prefix." + os.linesep)
                sys.exit(1)
        
        if self.outdir:
            if not os.path.isdir(self.outdir):
                os.makedirs(self.outdir)
                if not os.path.isdir(self.outdir):
                    sys.stderr.write("ERROR: Cannot create output directory " + self.outdir  + os.linesep)
                    sys.exit(1)
        else:
            print "No output directory was specified..."
            print "Output files will be generated in the current directory..."    
            self.outdir = os.getcwd()
            print self.outdir

        if self.jobdir:
            if not os.path.isdir(self.jobdir):
                os.makedirs(self.jobdir)
                if not os.path.isdir(self.jobdir):
                    sys.stderr.write("ERROR: Cannot create cluster job working directory " + self.jobdir  + os.linesep)
                    sys.exit(1)
        else:
            print "No job scripts directory was specified..."
            print "Job scripts and log files will be generated in the current directory..."
            self.jobdir = os.getcwd()
            print self.jobdir

        if self.debug:
            print "WARNING: Running in debug mode!"
            print "*** Only job scripts and/or symlinks will be created. ***"
            print "*** No resource-intensive commands will be run.       ***"
            print "*** No jobs will be submitted to the cluster.         ***"
            
        # Only symlink the reference to the output directory if the reference is not indexed.
        # The index files will then be created in the output directory.
        if not self.indexed:
            link_contigs = self.outdir + os.sep + contigs_base_name
            if not os.path.isfile(link_contigs):
                # make relative symlink
                os.symlink(os.path.relpath(self.contigs, self.outdir), link_contigs)
            else:
                print "WARNING: Contigs file already exists in output directory, " + link_contigs            
            self.contigs = link_contigs

        jobs = None
        cleanupjob = None
        cfg_file = '%s/configs/%s.cfg' % (TRANSABYSS_PATH, config_type)
        self.settings = cfg.initialize_settings(cfg_file, self.project)
    
        double_dollar_sign = False
        if self.makefile:
            #bash variables in make scripts must use double-dollar sign instead of single-dollar sign
            double_dollar_sign = True
    
        if self.aligner == ALIGNER_BWA:
            (jobs, cleanupjob) = generate_job_commands_bwa(contigs=self.contigs, reads_paths=self.reads_paths, prefix=self.prefix, single_end=self.single_end, outdir=self.outdir, threads=self.threads, local=self.local, xa2multi=self.xa2multi, strip=self.strip, messy=self.messy, sortmem=self.sortmem, indexed=self.indexed, samse=self.samse, predecessors=self.predecessors, settings=self.settings, double_dollar_sign=double_dollar_sign)        
        elif self.aligner == ALIGNER_BWASW:
            (jobs, cleanupjob) = generate_job_commands_bwasw(contigs=self.contigs, reads_paths=self.reads_paths, prefix=self.prefix, single_end=self.single_end, outdir=self.outdir, threads=self.threads, local=self.local, strip=self.strip, messy=self.messy, sortmem=self.sortmem, indexed=self.indexed, predecessors=self.predecessors, settings=self.settings, double_dollar_sign=double_dollar_sign)
        elif self.aligner == ALIGNER_BWAMEM:
            (jobs, cleanupjob) = generate_job_commands_bwamem(contigs=self.contigs, reads_paths=self.reads_paths, prefix=self.prefix, single_end=self.single_end, outdir=self.outdir, threads=self.threads, local=self.local, strip=self.strip, messy=self.messy, sortmem=self.sortmem, indexed=self.indexed, predecessors=self.predecessors, settings=self.settings, double_dollar_sign=double_dollar_sign)
        elif self.aligner == ALIGNER_ABYSSMAP:
            (jobs, cleanupjob) = generate_job_commands_abyssmap(contigs=self.contigs, reads_paths=self.reads_paths, prefix=self.prefix, outdir=self.outdir, threads=self.threads, local=self.local, k=self.k, strip=self.strip, single_end=self.single_end, messy=self.messy, sortmem=self.sortmem, indexed=self.indexed, predecessors=self.predecessors, settings=self.settings, double_dollar_sign=double_dollar_sign)
        elif self.aligner == ALIGNER_BOWTIE2:
            (jobs, cleanupjob) = generate_job_commands_bowtie2(contigs=self.contigs, reads_paths=self.reads_paths, prefix=self.prefix, outdir=self.outdir, threads=self.threads, local=self.local, strip=self.strip, single_end=self.single_end, messy=self.messy, sortmem=self.sortmem, indexed=self.indexed, predecessors=self.predecessors, settings=self.settings, double_dollar_sign=double_dollar_sign)
        else:
            print "ERROR: No appropriate aligner was specified."
            sys.exit(1)
    
        return (jobs, cleanupjob)

    def create_cleanup_script(self):
        """
        Method to create the clean up script.
        The script would not be run.
        
        Returns:
            (str) : path of the clean-up script.
        """
        #this clean up script is intended to be run locally
        cleanupscript_name = cleanupjob.name
        if not cleanupscript_name:
            cleanupscript_name = "cleanup-r2c"
        job_script_configs = job_runner.getJobScriptConfigs(TRANSABYSS_PATH + "/configs/job_script.cfg")
        cleanupscript_template = TRANSABYSS_PATH + "/configs/templates/" + job_script_configs["local"]
        
        #template, content, jobname=None, workingdir=None, threads=1, predecessors=None, memory=None, setup=None
        cleanupjobscript_content = job_runner.makeJobScript(template=cleanupscript_template, content=cleanupjob.content, jobname=cleanupscript_name, workingdir=self.outdir, setup=cfg.get_setup_cmd())
        script = self.jobdir + os.sep + cleanupscript_name + ".sh"
        fh = open(script, 'w')
        fh.write(cleanupjobscript_content)
        fh.close()
        return script

    def create_and_submit_jobs(self):
        """
        Method to create and submit the alignment jobs.
        
        Returns:
            (job_runner.MyJob) : the final job to be run. So, any jobs downstream to these alignment jobs can be set up to simply depend on just this job.
        """
        (jobs, cleanupjob) = self.create_jobs()

        if cleanupjob:
            if not self.messy and self.cleanup:
                # cleanup script will be the last task
                jobs.append(cleanupjob)        
            else:
                print "WARNING: Clean-up script will be generated, but it will not be run. Otherwise, use the '--cleanup' option of 'r2c.py' next time."
                script = create_cleanup_script()
                print "clean-up script: " + script
        else:
            print "No clean-up script will need to be run."
            
        if self.local:
            job_runner.run_locally(jobs=jobs, prefix=self.prefix, logdir=self.jobdir, debug=self.debug, workingdir=self.outdir)
        else:
            config_type = None    
            if self.genome:
                config_type = "genome"
            elif self.transcriptome:
                config_type = "transcriptome"
            cfg_file = '%s/configs/%s.cfg' % (TRANSABYSS_PATH, config_type)
            settings = cfg.initialize_settings(cfg_file, self.project)            
            
            job_runner.submit_jobs(jobs=jobs, logdir=self.jobdir, debug=self.debug, settings=settings, head_node=self.cluster, queue=self.queue, workingdir=self.outdir)
    
        return jobs[-1] # last job


def generate_samtobam_command(cmd="", input_sam=None, messy=False, strip=False, local=False, outputprefix=None, double_dollar_sign=False, sortmem=800000000):
    """
    Method to append the commands needed to convert from SAM to BAM.
    
    Args:
        cmd                (str)  : the command that generate SAM formatted output to stdout
        input_sam          (str)  : path of the SAM file to be converted
        messy              (bool) : flag to not use piping and keep all intermediate files
        strip              (bool) : flag to strip away the quality string of the SAM
        local              (bool) : flag to indicate jobs would be run locally
        outputprefix       (str)  : the prefix of output files
        double_dollar_sign (bool) : use '$$' instead '$' for referencing bash variables
        sortmem            (int)  : the argument to `samtools sort -m`
    
    Returns:
        ([str], str, [str]) :
            the first element of the tuple is the list of commands to generate the BAM file
            the second element of the tuple is the path of the final indexed BAM file
            the third element of the tuple is the list of the paths of intermediate file
    """

    intermediates = []
    tmpdir_bash_var = '$TMPDIR'    
    if double_dollar_sign:
        tmpdir_bash_var = '$' + tmpdir_bash_var
    
    cmds = []    
    if messy: # generate intermediate files whenever possible            
        if strip:
            #cmds.append("cat " + input_sam + " | ta-sam_strip_qual.awk > " + outputprefix + ".qual-stripped.sam")
            cmds.append("cat %s | ta-sam_strip_qual.awk > %s.qual-stripped.sam" % (input_sam, outputprefix))
            lastoutputfile = outputprefix + ".qual-stripped.sam"
            intermediates.append(lastoutputfile)
            
        cmds.append("samtools view -hSu " + lastoutputfile + " > " + outputprefix + ".unsorted.bam")
        lastoutputfile = outputprefix + ".unsorted.bam"
        intermediates.append(lastoutputfile)
        
        if local:
            cmds.append("samtools sort -m %s %s %s.incomplete" % (str(sortmem), lastoutputfile, outputprefix))
        else:
            cmds.append("samtools sort -o -m %s %s %s > %s.bam.incomplete || { rm -f %s*.bam 2> /dev/null; echo 'An error has occurred.'; exit 1;}" % (str(sortmem), lastoutputfile, tmpdir_bash_var, outputprefix, tmpdir_bash_var))
    else: # use pipes whenever possible
        if strip:
            cmd += "| ta-sam_strip_qual.awk "
        
        if local:
            cmd += "| samtools view -hSu - | samtools sort -m %s - %s.incomplete" % (str(sortmem), outputprefix)
        else:
            cmd += "| samtools view -hSu - | samtools sort -o -m %s - %s > %s.bam.incomplete || { rm -f %s*.bam 2> /dev/null; echo 'An error has occurred.'; exit 1;}" % (str(sortmem), tmpdir_bash_var, outputprefix, tmpdir_bash_var)
        cmds.append(cmd)

    bam = outputprefix + ".bam"
        
    if local:
        cmds.append('samtools index %s.incomplete.bam' % outputprefix)
        cmds.append('mv %s.incomplete.bam.bai %s.bam.bai' % (outputprefix, outputprefix))
        cmds.append('mv %s.incomplete.bam %s.bam' % (outputprefix, outputprefix))
    else:
        cmds.append('samtools index %s.bam.incomplete' % outputprefix)
        cmds.append('mv %s.bam.incomplete.bai %s.bam.bai' % (outputprefix, outputprefix))
        cmds.append('mv %s.bam.incomplete %s.bam' % (outputprefix, outputprefix))
    
    return (cmds, bam, intermediates)


def generate_bwa_indexing_job(contigs, settings, prefix, predecessors=None):
    """
    Method to create the job to generate bwa index files for the contigs FASTA.
    
    Args:
        contigs      (str)                : path the contigs FASTA file
        settings     ({str:str:str})      : the Trans-ABySS configurations dictionary. {section : key : value}
        prefix       (str)                : the prefix of the job names
        predecessors ([job_runner.MyJob]) : a list of predecessor jobs
        
    Returns:
        ([job_runner.MyJob, [str]]) :
            the first element is the bwa indexing job object
            the second element is a list of the paths of index files to be generated
    """

    config_key = 'r2c_bwa_index'
    cmds = []
    #cmds.append("cd %s" % outdir)
    cmds.append("samtools faidx %s" % contigs)

    #r2c_bwa_index: ${CONTIGS}
    args = {}
    args['CONTIGS'] = contigs
    cmds = []
    cmds.append('bwa index %s' % cfg.create_args(settings, 'commands', config_key, args))
    stamp = contigs + '.INDEX.COMPLETE'
    cmds.append('touch ' + stamp)
    
    name = prefix + "-index"
    my_predecessors = []
    if predecessors:
        my_predecessors.extend(predecessors)

    indexfiles = []
    for index_extension in [".fai", ".amb", ".ann", ".bwt", ".pac", ".sa"]:
        indexfiles.append(contigs + index_extension)

    indexing_job = MyJob(content=cmds, id=None, name=name, predecessors=my_predecessors, config_key=config_key, outputfiles=[stamp])
    
    return [indexing_job, indexfiles] 


def generate_merge_bam_jobs(alignment_jobs=[], output_bam_files=[], outdir=None, prefix=None):
    """
    Method to create the job to merge multiple BAM files or symlink the BAM file with the appropriate name.
    
    Args:
        alignment_jobs    ([job_runner.MyJob]) : a list of MyJob objects for the alignment jobs
        output_bam_files ([str]) : a list of BAM files to be merged.
        prefix           (str) : the prefix of the job name
    
    Returns:
        ([job_runner.MyJob, [str]]):
            the first element is the MyJob object.
            the second element is the list of intermediate files.
    """
    
    #part 1. Symlink or merge the sorted bam files
    intermediates = []
    cmds = []
    num_alignment_jobs = len(alignment_jobs)
    merge_job = None
    stamp = outdir + os.sep + 'R2C.COMPLETE'
    if num_alignment_jobs > 1:
        outputfile = '%s/%s-contigs.bam' % (outdir, prefix)
        cmds.append("samtools merge -f %s.incomplete %s" % (outputfile, " ".join(output_bam_files)))
        cmds.append("samtools index %s.incomplete" % outputfile)
        cmds.append('mv %s.incomplete.bai %s.bai' % (outputfile, outputfile))
        cmds.append('mv %s.incomplete %s' % (outputfile, outputfile))
        cmds.append('touch ' + stamp)
        name = prefix + "-samtools_merge"
        merge_job = MyJob(content=cmds, name=name, predecessors=alignment_jobs, config_key="samtools_merge", outputfiles=[stamp])
        
        # the sorted BAMs are intermediate files because we now have the merged BAM
        intermediates.extend(output_bam_files)
        # we want to remove the BAM indexes, too
        for bam in output_bam_files:
            intermediates.append(bam + ".bai")
    elif num_alignment_jobs == 1:
        # this sorted BAM is not an intermediate file because it is the final BAM
        bam = output_bam_files[0]
        bam_relpath = os.path.relpath(bam, outdir)
        cmds.append("ln -sf %s.bai %s/%s-contigs.bam.bai" % (bam_relpath, outdir, prefix))
        cmds.append("ln -sf %s %s/%s-contigs.bam" % (bam_relpath, outdir, prefix))
        cmds.append('touch ' + stamp)
        alignment_jobs[0].content.extend(cmds)
        alignment_jobs[0].outputfiles = [stamp]
    else:
        sys.stderr.write("ERROR: Nonsensical number of alignment jobs (" + str(num_alignment_jobs) + ")." + os.linesep)
        sys.exit(1)
    
    return [merge_job, intermediates]


def generate_job_commands_bowtie2(contigs=None, reads_paths=None, prefix=None, outdir=None, threads=1, local=False, strip=False, single_end=False, messy=False, sortmem=800000000, indexed=False, predecessors=None, settings=None, double_dollar_sign=False):
    """
    Method to create the job for aligning reads to contigs with Bowtie2.
    
    Args:
        contigs      (str)    : the path of the contigs FASTA file
        reads_paths  ([str])  : the paths of read files
        prefix       (str)    : the prefix of output files
        outdir       (str)    : the path of the output directory
        threads      (int)    : the number of threads to use for Bowtie2
        local        (bool)   : the flag to indicate the job would be run locally
        strip        (bool)   : the flag to strip the quality string from the output BAM file
        single_end   (bool)   : the flag to indicate the reads are single-end data
        messy        (bool)   : the flag to not use piping and keep intermediate files
        sortmem      (int)    : the argument to `samtools sort -m`
        indexed      (bool)   : the flag to indicate the contigs FASTA file was already indexed
        predecessors ([job_runner.MyJob]) : the list of predecessor job objects
        settings     ({str:str:str})      : the Trans-ABySS configurations dictionary. {section : key : value}
        double_dollar_sign (bool)         : use '$$' instead '$' for referencing bash variables
                
    Returns:
        ([job_runner.MyJob], job_runner.MyJob) :
            The first element is a list of jobs to do the alignments and generate a final indexed BAM file.
            The second element is the job for cleaning up intermediate files.
    """

    finaloutputprefix = outdir + os.sep + prefix + "-contigs"
    intermediates = []
    
    indexing_job = None    
    
    #1. bowtie2-build -f FASTA PREFIX
    if not indexed:    
        cmds = []
        #cmds.append("cd %s" % outdir)
        
        config_key = 'r2c_bowtie2-build'
        #r2c_bowtie2-build: -f ${CONTIGS} ${BT2_BASE}
        args = {}
        args['CONTIGS'] = contigs
        args['BT2_BASE'] = finaloutputprefix
        cmds.append('bowtie2-build %s' % cfg.create_args(settings, 'commands', config_key, args))        
        stamp = contigs + '.INDEX.COMPLETE'
        cmds.apend('touch ' + stamp)
        
        name = prefix + "-bowtie2-build"
        my_predecessors = []
        if predecessors:
            my_predecessors.extend(predecessors)
        for indexes_extension in [".1.bt2", ".2.bt2", ".3.bt2", ".4.bt2", ".rev.1.bt2", ".rev.2.bt2"]:
            intermediates.append(finaloutputprefix + indexes_extension)
        indexing_job = MyJob(content=cmds, id=None, name=name, predecessors=my_predecessors, config_key=config_key, outputfiles=[stamp])

        
    #if index_only:
    #    return ([indexing_job], None)
        
    #2. bowtie2
    task_id = 1
    alignment_jobs = []
    output_bam_files = []
        
    if single_end:
        for read in reads_paths:
            outputprefix = outdir + os.sep + os.path.basename(read)
            lastoutputfile = None
        
            cmds = []
            #cmds.append("cd %s" % outdir)
            
            config_key = 'r2c_bowtie2'
            #r2c_bowtie2: -k 20 --end-to-end --very-fast -p ${THREADS} -x ${BT2_BASE} ${READS_FILE_FORMAT} ${READFILES_MATE1} ${READFILES_MATE2} ${READFILES_UNPAIRED}
            args = {}
            args['THREADS'] = str(threads)
            args['BT2_BASE'] = finaloutputprefix
            
            if is_fastq(read):
                args['READS_FILE_FORMAT'] = "-q"
            elif is_txt(read):
                args['READS_FILE_FORMAT'] = "--qseq"
            elif is_fasta(read):
                args['READS_FILE_FORMAT'] = "-f"
            else:
                args['READS_FILE_FORMAT'] = "-r"
            
            args['READFILES_MATE1'] = ''
            args['READFILES_MATE2'] = ''
            args['READFILES_UNPAIRED'] = "-U %s" % read
            
            temp_cmd = 'bowtie2 %s' % cfg.create_args(settings, 'commands', config_key, args)
            
            lastoutputfile = None
            
            if messy: # generate intermediate files whenever possible
                temp_cmd += " > " + outputprefix + ".sam"
                cmds.append(temp_cmd)
                lastoutputfile = outputprefix + ".sam"
                intermediates.append(lastoutputfile)
                
            (samtobam_cmd, outputfile, samtobam_tmpfiles) = generate_samtobam_command(temp_cmd, input_sam=lastoutputfile, messy=messy, strip=strip, local=local, outputprefix=outputprefix, double_dollar_sign=double_dollar_sign, sortmem=sortmem)
            cmds.extend(samtobam_cmd)
            intermediates.extend(samtobam_tmpfiles)
                
            output_bam_files.append(outputfile)
                
            name = prefix + "-bowtie2-" + str(task_id).zfill(4)
            my_predecessors = []
            if indexing_job:
                my_predecessors.append(indexing_job)
            if predecessors:
                my_predecessors.extend(predecessors)
            alignment_jobs.append(MyJob(content=cmds, id=None, name=name, predecessors=my_predecessors, config_key=config_key, threads=threads, outputfiles=[outputfile]))         
            task_id += 1   
    else:
        for pair in list(chunks(reads_paths, 2)):               
            read1 = pair[0]
            read2 = pair[1]
            outputprefix = outdir + os.sep + os.path.basename(read1)
            lastoutputfile = None            
            
            cmds = []
            #cmds.append("cd %s" % outdir)
            
            config_key = 'r2c_bowtie2'
            #r2c_bowtie2: -k 20 --end-to-end --very-fast -p ${THREADS} -x ${BT2_BASE} ${READS_FILE_FORMAT} ${READFILES_MATE1} ${READFILES_MATE2} ${READFILES_UNPAIRED}
            args = {}
            args['THREADS'] = str(threads)
            args['BT2_BASE'] = finaloutputprefix
            
            if is_fastq(read1) and is_fastq(read2):
                args['READS_FILE_FORMAT'] = "-q"
            elif is_txt(read1) and is_txt(read2):
                args['READS_FILE_FORMAT'] = "--qseq"
            elif is_fasta(read1) and is_fasta(read2):
                args['READS_FILE_FORMAT'] = "-f"
            else:
                args['READS_FILE_FORMAT'] = "-r"
            
            args['READFILES_MATE1'] = "-1 %s" % read1
            args['READFILES_MATE2'] = "-2 %s" % read2
            args['READFILES_UNPAIRED'] = ''
            
            temp_cmd = 'bowtie2 %s' % cfg.create_args(settings, 'commands', config_key, args)
            
            lastoutputfile = None
            
            if messy: # generate intermediate files whenever possible
                temp_cmd += " > " + outputprefix + ".sam"
                cmds.append(temp_cmd)
                lastoutputfile = outputprefix + ".sam"
                intermediates.append(lastoutputfile)
                
            (samtobam_cmd, outputfile, samtobam_tmpfiles) = generate_samtobam_command(temp_cmd, input_sam=lastoutputfile, messy=messy, strip=strip, local=local, outputprefix=outputprefix, double_dollar_sign=double_dollar_sign, sortmem=sortmem)
            cmds.extend(samtobam_cmd)
            intermediates.extend(samtobam_tmpfiles)
                
            output_bam_files.append(outputfile)
                
            name = prefix + "-bowtie2-" + str(task_id).zfill(4)
            my_predecessors = []
            if indexing_job:
                my_predecessors.append(indexing_job)
            if predecessors:
                my_predecessors.extend(predecessors)
            alignment_jobs.append(MyJob(content=cmds, id=None, name=name, predecessors=my_predecessors, config_key=config_key, threads=threads, outputfiles=[outputfile]))         
            task_id += 1 
        
    #3. Symlink or merge the sorted bam files
    [merge_job, tmpfiles] = generate_merge_bam_jobs(alignment_jobs=alignment_jobs, output_bam_files=output_bam_files, outdir=outdir, prefix=prefix)
    intermediates.extend(tmpfiles)
    
    #4. Clean up
    cmds = []
    #cmds.append("cd %s" % outdir)
    ftbd = ' '.join(intermediates)
    #cmds.append("ls -1 %s" % ftbd)
    cmds.append("rm -f %s" % ftbd)
    name = prefix + "-cleanup-r2c"
    
    cleanup_predecessor = merge_job
    if not cleanup_predecessor:
        cleanup_predecessor = alignment_jobs[0]
        
    cleanup_job = MyJob(content=cmds, name=name, predecessors=[cleanup_predecessor])

    jobs = []
    if indexing_job:
        jobs.append(indexing_job)
    jobs.extend(alignment_jobs)
    if merge_job:
        jobs.append(merge_job)
    
    return (jobs, cleanup_job)


def is_fastq(path):
    """
    Method to determine whether the path ends with a (compressed) FASTQ postfix.
    
    Args:
        path (str) : the path to be tested
    
    Returns:
        (bool) : whether the path if FASTQ
    """
    return path.endswith(".fastq") or path.endswith(".fastq.gz") or path.endswith(".fastq.bz2") or path.endswith(".fq") or path.endswith(".fq.gz") or path.endswith(".fq.bz2")
    
    
def is_fasta(path):
    """
    Method to determine whether the path ends with a (compressed) FASTA postfix.
    
    Args:
        path (str) : the path to be tested
    
    Returns:
        (bool) : whether the path if FASTA
    """
    return path.endswith(".fasta") or path.endswith(".fasta.gz") or path.endswith(".fasta.bz2") or path.endswith(".fa") or path.endswith(".fa.gz") or path.endswith(".fa.bz2")
    
    
def is_txt(path):
    """
    Method to determine whether the path ends with a (compressed) QSEQ postfix.
    
    Args:
        path (str) : the path to be tested
    
    Returns:
        (bool) : whether the path if QSEQ
    """
    return path.endswith(".txt") or path.endswith(".txt.gz") or path.endswith(".txt.bz2")


def is_bam(path):
    """
    Method to determine whether the path ends with a BAM postfix.
    
    Args:
        path (str) : the path to be tested
    
    Returns:
        (bool) : whether the path if BAM
    """
    return path.endswith(".bam")


def generate_job_commands_abyssmap(contigs=None, reads_paths=None, prefix=None, outdir=None, threads=1, local=False, k=24, strip=False, single_end=False, messy=False, sortmem=800000000, indexed=False, predecessors=None, settings=None, double_dollar_sign=False):
    """
    Method to create the job for aligning reads to contigs with abyss-map.
    
    Args:
        contigs      (str)    : the path of the contigs FASTA file
        reads_paths  ([str])  : the paths of read files
        prefix       (str)    : the prefix of output files
        outdir       (str)    : the path of the output directory
        threads      (int)    : the number of threads to use for abyss-map
        local        (bool)   : the flag to indicate the job would be run locally
        k            (int)    : the k-mer size for abyss-map
        strip        (bool)   : the flag to strip the quality string from the output BAM file
        single_end   (bool)   : the flag to indicate the reads are single-end data
        messy        (bool)   : the flag to not use piping and keep intermediate files
        sortmem      (int)    : the argument to `samtools sort -m`
        indexed      (bool)   : the flag to indicate the contigs FASTA file was already indexed
        predecessors ([job_runner.MyJob]) : the list of predecessor job objects
        settings     ({str:str:str})      : the Trans-ABySS configurations dictionary. {section : key : value}
        double_dollar_sign (bool)         : use '$$' instead '$' for referencing bash variables
                
    Returns:
        ([job_runner.MyJob], job_runner.MyJob) :
            The first element is a list of jobs to do the alignments and generate a final indexed BAM file.
            The second element is the job for cleaning up intermediate files.
    """

    finaloutputprefix = outdir + "/" + prefix + "-contigs"      
    
    jobs=[]
    alignment_jobs = []
    merge_job = None
    cleanup_job = None
    intermediates = []
    output_bam_files = []
    
    indexing_job = None    
    
    #1. abyss-index FASTA
    if not indexed:
        indexes = []
        cmds = []
        #cmds.append("cd %s" % outdir)
        config_key = 'r2c_abyss-index'
        #r2c_abyss-index: ${CONTIGS}
        args = {}
        args['CONTIGS'] = contigs
        
        #Bypass the "Bad Address" error
        cmds.append('abyss-index --verbose --fai --stdout %s > %s.fai' % (contigs, contigs))
        cmds.append('abyss-index --verbose --fm --stdout %s > %s.fm' % (contigs, contigs))
        
        stamp = contigs + '.INDEX.COMPLETE'
        cmds.append('touch ' + stamp)
        
        name = prefix + "-abyss-index"
        my_predecessors = []
        if predecessors:
            my_predecessors.extend(predecessors)
        for indexes_extension in [".fai", ".fm"]:
            indexes.append(contigs + indexes_extension)
        intermediates.extend(indexes)
        indexing_job = MyJob(content=cmds, id=None, name=name, predecessors=my_predecessors, config_key=config_key, outputfiles=[stamp])
           
    #if index_only:
    #    return ([indexing_job], None)
    
    #2. abyss-map ...
    config_key = 'r2c_abyss-map'    
    if single_end:
        task_id = 1
        # one job for each read file
        for read_file in reads_paths:
            cmds = []
            #cmds.append("cd %s" % outdir)
            outputprefix = outdir + os.sep + str(task_id).zfill(4) + '.' + os.path.basename(read_file)           
            [align_cmd, align_finaloutput, align_intermediates] = abyssmap_helper(contigs=contigs, reads=read_file, outputprefix=outputprefix, threads=threads, local=local, k=k, strip=strip, messy=messy, sortmem=sortmem, settings=settings)
            cmds.extend(align_cmd)
            
            intermediates.extend(align_intermediates)
            name = prefix + "-abyssmap-" + str(task_id).zfill(4)
            my_predecessors = []
            if indexing_job:
                my_predecessors.append(indexing_job)
            elif predecessors:
                my_predecessors.extend(predecessors)
            alignment_jobs.append(MyJob(content=cmds, id=None, name=name, predecessors=my_predecessors, config_key=config_key, threads=threads, outputfiles=[align_finaloutput]))
            output_bam_files.append(align_finaloutput)
            task_id += 1
    else:
        lastread = None
        task_id = 1
        # one job for each mate-pair
        for read_file in reads_paths:
            if is_bam(read_file):
                # BAM files should have both mates
                if lastread:
                    print "ERROR: Read files are not paired properly."
                    print lastread
                    print read_file
                    sys.exit(1)
                else:
                    cmds = []
                    #cmds.append("cd %s" % outdir)
                    outputprefix = outdir + os.sep + str(task_id).zfill(4) + '.' + os.path.basename(read_file)           
                    [align_cmd, align_finaloutput, align_intermediates] = abyssmap_helper(contigs=contigs, reads=read_file, outputprefix=outputprefix, threads=threads, local=local, k=k, strip=strip, messy=messy, sortmem=sortmem, settings=settings, double_dollar_sign=double_dollar_sign)
                    cmd += align_cmd
                    
                    intermediates.extend(align_intermediates)                                        
                    name = prefix + "-abyssmap-" + str(task_id).zfill(4)
                    my_predecessors = []
                    if indexing_job:
                        my_predecessors.append(indexing_job)
                    elif predecessors:
                        my_predecessors.extend(predecessors)
                    alignment_jobs.append(MyJob(content=cmd, id=None, name=name, predecessors=my_predecessors, config_key=config_key, threads=threads, outpufiles=[align_finaloutput]))
                    output_bam_files.append(align_finaloutput)
                    task_id += 1
            else:
                if lastread:
                    cmds = []
                    #cmds.append("cd %s" % outdir)
                    outputprefix = outdir + os.sep + str(task_id).zfill(4) + '.' + os.path.basename(read)           
                    [align_cmd, align_finaloutput, align_intermediates] = abyssmap_helper(contigs=contigs, reads=lastread + " " + read, outputprefix=outputprefix, threads=threads, local=local, k=k, strip=strip, messy=messy, sortmem=sortmem, settings=settings, double_dollar_sign=double_dollar_sign)
                    cmds.extend(align_cmd)
                    
                    intermediates.extend(align_intermediates)                                        
                    name = prefix + "-abyssmap-" + str(task_id).zfill(4)
                    my_predecessors = []
                    if indexing_job:
                        my_predecessors.append(indexing_job)
                    elif predecessors:
                        my_predecessors.extend(predecessors)
                    alignment_jobs.append(MyJob(content=cmds, id=None, name=name, predecessors=my_predecessors, config_key=config_key, threads=threads, outputfiles=[align_finaloutput]))
                    output_bam_files.append(align_finaloutput)
                    task_id += 1
                    lastread = None
                else:
                    lastread = read

    #3. Symlink or merge the sorted bam files
    [merge_job, tmpfiles] = generate_merge_bam_jobs(alignment_jobs=alignment_jobs, output_bam_files=output_bam_files, outdir=outdir, prefix=prefix)
    intermediates.extend(tmpfiles)    

    #4. Clean up
    cmds = []
    #cmds.append("cd %s" % outdir)
    ftbd = ' '.join(intermediates)
    #cmds.append("ls -1 %s" % ftbd)
    cmds.append("rm -f %s" % ftbd)
    name = prefix + "-cleanup-r2c"
    
    cleanup_predecessor = merge_job
    if not cleanup_predecessor:
        cleanup_predecessor = alignment_jobs[0]
        
    cleanup_job = MyJob(content=cmds, name=name, predecessors=[cleanup_predecessor])

    jobs = []
    if indexing_job:
        jobs.append(indexing_job)
    jobs.extend(alignment_jobs)
    if merge_job:
        jobs.append(merge_job)
    
    return (jobs, cleanup_job)


def abyssmap_helper(contigs=None, reads=None, outputprefix=None, threads=1, local=False, k=24, strip=False, messy=False, sortmem=800000000, settings=False, double_dollar_sign=False):
    """
    Helper method to generate the commands to align with abyss-map.
    
    Args:
        contigs      (str)    : the path of the contigs FASTA file
        reads        ([str])  : the paths of read files
        outputprefix (str)    : the prefix of output files
        threads      (int)    : the number of threads to use for abyss-map
        local        (bool)   : the flag to indicate the job would be run locally
        k            (int)    : the k-mer size for abyss-map
        strip        (bool)   : the flag to strip the quality string from the output BAM file
        messy        (bool)   : the flag to not use piping and keep intermediate files
        sortmem      (int)    : the argument to `samtools sort -m`
        settings     ({str:str:str})      : the Trans-ABySS configurations dictionary. {section : key : value}
        double_dollar_sign (bool)         : use '$$' instead '$' for referencing bash variables
        
    Returns:
        ([str], str, [str]) :
            the first element of the tuple is the list of commands to generate the BAM file
            the second element of the tuple is the path of the final indexed BAM file
            the third element of the tuple is the list of the paths of intermediate file
    """
    
    intermediates = []
    lastoutputfile = None
        
    config_key = 'r2c_abyss-map'
    #r2c_abyss-map: ${READFILES} ${CONTIGS} -k MINK -j ${THREADS}     
    args = {}
    args['READFILES'] = reads
    args['CONTIGS'] = contigs
    args['MINK'] = str(k)
    args['THREADS'] = str(threads)
    
    cmds = []
    temp_cmd = 'abyss-map %s' % cfg.create_args(settings, 'commands', config_key, args)
    
    lastoutputfile = None
    
    if messy: # generate intermediate files whenever possible
        temp_cmd += " > " + outputprefix + ".sam"
        cmds.append(temp_cmd)
        lastoutputfile = outputprefix + ".sam"
        intermediates.append(lastoutputfile)
        
        cmds.append("abyss-fixmate --qname " + lastoutputfile + " > " + outputprefix + ".fixmate.sam")
        lastoutputfile = outputprefix + ".fixmate.sam"
        intermediates.append(lastoutputfile)
    else: # use pipes whenever possible
        temp_cmd += " | abyss-fixmate --qname "
    
    (samtobam_cmd, outputfile, samtobam_tmpfiles) = generate_samtobam_command(temp_cmd, input_sam=lastoutputfile, messy=messy, strip=strip, local=local, outputprefix=outputprefix, double_dollar_sign=double_dollar_sign, sortmem=sortmem)
    cmds.extend(samtobam_cmd)
    intermediates.extend(samtobam_tmpfiles)
    
    return [cmds, outputfile, intermediates]


def chunks(l, n):
    """
    Yield successive n-sized chunks from l.
    """
    for i in xrange(0, len(l), n):
        yield l[i:i+n]    


def generate_job_commands_bwa(contigs=None, reads_paths=None, prefix=None, single_end=False, outdir=None, threads=1, local=False, xa2multi=False, strip=False, messy=False, sortmem=800000000, indexed=False, samse=False, predecessors=None, settings=None, double_dollar_sign=False):
    """
    Method to create the job for aligning reads to contigs with bwa.
    
    Args:
        contigs      (str)    : the path of the contigs FASTA file
        reads_paths  ([str])  : the paths of read files
        prefix       (str)    : the prefix of output files
        single_end   (bool)   : the flag to indicate the reads are single-end data
        outdir       (str)    : the path of the output directory
        threads      (int)    : the number of threads to use for `bwa aln`
        local        (bool)   : the flag to indicate the job would be run locally
        xa2multi     (bool)   : the flag to run `xa2multi.pl`
        strip        (bool)   : the flag to strip the quality string from the output BAM file
        messy        (bool)   : the flag to not use piping and keep intermediate files
        sortmem      (int)    : the argument to `samtools sort -m`
        indexed      (bool)   : the flag to indicate the contigs FASTA file was already indexed
        samse        (bool)   : the flag to run `bwa samse` instead of `bwa sampe`
        predecessors ([job_runner.MyJob]) : the list of predecessor job objects
        settings     ({str:str:str})      : the Trans-ABySS configurations dictionary. {section : key : value}
        double_dollar_sign (bool)         : use '$$' instead '$' for referencing bash variables
                
    Returns:
        ([job_runner.MyJob], job_runner.MyJob) :
            The first element is a list of jobs to do the alignments and generate a final indexed BAM file.
            The second element is the job for cleaning up intermediate files.
    """

    intermediates = []

    indexing_job = None

    #1. index the contigs fasta file
    if not indexed:        
        [indexing_job, indexfiles] = generate_bwa_indexing_job(contigs, settings, prefix, predecessors)
        intermediates.extend(indexfiles)
    
    #if index_only:
    #    return ([indexing_job], None)

    #2. align the reads to contigs
    bwa_aln_jobs = []
    bwa_sam_jobs = []
    sai_files = []
    output_bam_files = []
        
    if single_end:
        i = 1
        for readfile in reads_paths:                       
            if messy: # generate intermediate files whenever possible
                #bwa aln ...            
                sai = outdir + os.sep + os.path.basename(readfile) + ".sai"
                sai_files.append(sai)
                
                cmds = []                
                #cmds.append("cd %s" % outdir)
                #r2c_bwa_aln: -q 3 -t ${THREADS} ${BAMFLAG} ${CONTIGS} ${READFILE}
                config_key = 'r2c_bwa_aln'                
                args = {}
                args['THREADS'] = str(threads)
                if is_bam(readfile):
                    args['BAMFLAG'] = '-b'
                else:
                    args['BAMFLAG'] = ''
                args['CONTIGS'] = contigs
                args['READFILE'] = readfile
                cmds.append('bwa aln %s > %s' % (cfg.create_args(settings, 'commands', config_key, args), sai))
                
                name = prefix + "-bwa_aln-" + str(i).zfill(4)
                my_predecessors = []
                if indexing_job:
                    my_predecessors.append(indexing_job)
                elif predecessors:
                    my_predecessors.extend(predecessors)
                aln_job = MyJob(content=cmds, id=None, name=name, predecessors=my_predecessors, config_key=config_key, threads=threads, outputfiles=[sai])
                bwa_aln_jobs.append(aln_job)
                
                #bwa samse ...
                cmds = []                
                #cmds.append("cd %s" % outdir)
                #r2c_bwa_samse: -n50 ${CONTIGS} ${SAI} ${READFILE}
                config_key = 'r2c_bwa_samse'
                args = {}
                args['CONTIGS'] = contigs
                args['SAI'] = sai
                args['READFILE'] = readfile
                cmds.append('bwa samse %s > %s.sam' % (cfg.create_args(settings, 'commands', config_key, args), outputprefix))
                
                lastoutputfile = outputprefix + ".sam"
                intermediates.append(lastoutputfile)
                
                if xa2multi:
                    cmds.append("cat %s | xa2multi.pl > %s.xa2multi.sam" % (lastoutputfile, outputprefix))
                    lastoutputfile = outputprefix + ".xa2multi.sam"
                    intermediates.append(lastoutputfile)
                
                (samtobam_cmd, outputfile, samtobam_tmpfiles) = generate_samtobam_command(input_sam=lastoutputfile, messy=messy, strip=strip, local=local, outputprefix=outputprefix, double_dollar_sign=double_dollar_sign, sortmem=sortmem)
                cmds.extend(samtobam_cmd)
                intermediates.extend(samtobam_tmpfiles)
                
                output_bam_files.append(outputfile)
                
                name = prefix + "-bwa_samse-" + str(i).zfill(4)
                sam_job = MyJob(content=cmd, id=None, name=name, predecessors=[aln_job], config_key=config_key, outputfiles=[outputfile])
                bwa_sam_jobs.append(sam_job)
            else: # use pipes whenever possible
                cmds = []                
                #cmds.append("cd %s" % outdir)
                
                #r2c_bwa_aln: -q 3 -t ${THREADS} ${BAMFLAG} ${CONTIGS} ${READFILE}
                config_key = 'r2c_bwa_aln'
                args = {}
                args['THREADS'] = str(threads)
                if is_bam(readfile):
                    args['BAMFLAG'] = '-b'
                else:
                    args['BAMFLAG'] = ''
                args['CONTIGS'] = contigs
                args['READFILE'] = readfile
                temp_cmd = 'bwa aln %s |' % cfg.create_args(settings, 'commands', config_key, args)
                                             
                #r2c_bwa_samse: -n50 ${CONTIGS} ${SAI} ${READFILE}
                config_key = 'r2c_bwa_samse'
                args = {}
                args['CONTIGS'] = contigs
                args['SAI'] = '-'
                args['READFILE'] = readfile
                temp_cmd += 'bwa samse %s' % cfg.create_args(settings, 'commands', config_key, args)
                
                outputprefix = outdir + '/' + os.path.basename(readfile)   
            
                if xa2multi:
                    cmd += " | xa2multi.pl "
                
                (samtobam_cmd, outputfile, samtobam_tmpfiles) = generate_samtobam_command(temp_cmd, input_sam=None, messy=messy, strip=strip, local=local, outputprefix=outputprefix, double_dollar_sign=double_dollar_sign, sortmem=sortmem)
                cmds.extend(samtobam_cmd)
                intermediates.extend(samtobam_tmpfiles)
                
                output_bam_files.append(outputfile)
                
                name = prefix + "-bwa_aln,samse-" + str(i).zfill(4)
                my_predecessors = []
                if indexing_job:
                    my_predecessors.append(indexing_job)
                elif predecessors:
                    my_predecessors.extend(predecessors)
                sam_job = MyJob(content=cmds, id=None, name=name, predecessors=my_predecessors, config_key="r2c_bwa_aln,sam", outputfiles=[outputfile])
                bwa_sam_jobs.append(sam_job)
                        
            i += 1
    else: # paired-end
        i = 1
        cursor = -1
        for index, readfile in enumerate(reads_paths):
            if index <= cursor:
                continue
            
            mate_readfile = readfile
            
            if messy: # generate intermediate files whenever possible
                sai1 = ""
                sai2 = ""
                bamflag1 = ''
                bamflag2 = ''
                if is_bam(readfile):
                    cursor = index
                    bamflag1 = "-b1"
                    bamflag2 = "-b2"
                    sai1 = outdir + "/" + os.path.basename(readfile) + ".b1.sai"
                    sai2 = outdir + "/" + os.path.basename(mate_readfile) + ".b2.sai"
                else: # fq, fq.gz, fastq, fastq.gz
                    mate_readfile = reads_paths[index+1]
                    cursor = index+1                 
                    sai1 = outdir + "/" + os.path.basename(readfile) + ".sai"
                    sai2 = outdir + "/" + os.path.basename(mate_readfile) + ".sai"
                sai_files.append(sai1)
                sai_files.append(sai2)
     
                #bwa aln for mate 1
                cmds = []
                #cmds.append("cd %s" % outdir)
                #r2c_bwa_aln: -q 3 -t ${THREADS} ${BAMFLAG} ${CONTIGS} ${READFILE}
                config_key = 'r2c_bwa_aln'                
                args = {}
                args['THREADS'] = str(threads)
                args['BAMFLAG'] = bamflag1
                args['CONTIGS'] = contigs
                args['READFILE'] = readfile
                cmds.append('bwa aln %s > %s' % (cfg.create_args(settings, 'commands', config_key, args), sai1))
                name = prefix + "-bwa_aln-" + str(i).zfill(4)
                my_predecessors = []
                if indexing_job:
                    my_predecessors.append(indexing_job)
                elif predecessors:
                    my_predecessors.extend(predecessors)
                aln_job1 = MyJob(content=cmds, id=None, name=name, predecessors=my_predecessors, config_key=config_key, threads=threads, outputfiles=[sai1])
                bwa_aln_jobs.append(aln_job1)
                
                #bwa aln for mate 2
                cmds = []
                #cmds.append("cd %s" % outdir)
                args['BAMFLAG'] = bamflag2
                args['READFILE'] = mate_readfile
                cmds.append('bwa aln %s > %s' % (cfg.create_args(settings, 'commands', config_key, args), sai2))
                name = prefix + "-bwa_aln-" + str(i+1).zfill(4)
                my_predecessors = []
                if indexing_job:
                    my_predecessors.append(indexing_job)
                elif predecessors:
                    my_predecessors.extend(predecessors)
                aln_job2 = MyJob(content=cmds, id=None, name=name, predecessors=my_predecessors, config_key=config_key, threads=threads, outputfiles=[sai2])
                bwa_aln_jobs.append(aln_job2)
                
                if samse:
                    #bwa samse for mate 1
                    cmds = []
                    #cmds.append("cd %s" % outdir)
                    outputprefix = outdir + '/' + os.path.basename(readfile)
                    
                    #r2c_bwa_samse: -n50 ${CONTIGS} ${SAI} ${READFILE}
                    config_key = 'r2c_bwa_samse'
                    args = {}
                    args['CONTIGS'] = contigs
                    args['SAI'] = sai1
                    args['READFILE'] = readfile
                    cmds.append('bwa samse %s > %s.sam' % (cfg.create_args(settings, 'commands', config_key, args), outputprefix))
                
                    lastoutputfile = outputprefix + ".sam"
                    intermediates.append(lastoutputfile)
                    
                    if xa2multi:
                        cmds.append("cat " + lastoutputfile + " | xa2multi.pl > " + outputprefix + ".xa2multi.sam")
                        lastoutputfile = outputprefix + ".xa2multi.sam"
                        intermediates.append(lastoutputfile)
                    
                    temp_cmd = "cat " + lastoutputfile + '| awk -F\\t \'{OFS="\t"; if($1!="@SQ" && $1!="@HD" && $1!="@RG" && $1!="@PG" && $1!="@CO"){'
                    temp_cmd  += '$2=or($2,0x40);'
                    if strip:
                        temp_cmd  += '$11="*"' #remove QUAL
                    temp_cmd  += '}; print}\' > ' + outputprefix + ".segment-flagged.sam"
                    cmds.append(temp_cmd)
                    lastoutputfile = outputprefix + ".segment-flagged.sam"
                    intermediates.append(lastoutputfile)
                    
                    (samtobam_cmd, outputfile, samtobam_tmpfiles) = generate_samtobam_command(input_sam=lastoutputfile, messy=messy, strip=False, local=local, outputprefix=outputprefix, double_dollar_sign=double_dollar_sign, sortmem=sortmem)
                    cmds.extend(samtobam_cmd)
                    intermediates.extend(samtobam_tmpfiles)
                
                    output_bam_files.append(outputfile)
                    
                    name = prefix + "-bwa_samse-" + str(i).zfill(4)
                    sam_job1 = MyJob(content=cmds, id=None, name=name, predecessors=[aln_job1], config_key=config_key, outputfiles=[outputfile])
                    bwa_sam_jobs.append(sam_job1)
                    
                    #bwa samse for mate 2
                    cmds = []
                    #cmds.append("cd %s" % outdir)
                    outputprefix = outdir + '/' + os.path.basename(mate_readfile)

                    config_key = 'r2c_bwa_samse'
                    args = {}
                    args['CONTIGS'] = contigs
                    args['SAI'] = sai2
                    args['READFILE'] = mate_readfile
                    cmds.append('bwa samse %s > %s.sam' % (cfg.create_args(settings, 'commands', config_key, args), outputprefix))
                    lastoutputfile = outputprefix + ".sam"
                    intermediates.append(lastoutputfile)
                    
                    if xa2multi:
                        cmds.append("cat " + lastoutputfile + " | xa2multi.pl > " + outputprefix + ".xa2multi.sam")
                        lastoutputfile = outputprefix + ".xa2multi.sam"
                        intermediates.append(lastoutputfile)
                    
                    temp_cmd = "cat " + lastoutputfile + '| awk -F\\t \'{OFS="\t"; if($1!="@SQ" && $1!="@HD" && $1!="@RG" && $1!="@PG" && $1!="@CO"){'
                    temp_cmd += '$2=or($2,0x80);'
                    if strip:
                        temp_cmd += '$11="*"' #remove QUAL
                    temp_cmd += '}; print}\' > ' + outputprefix + ".segment-flagged.sam"
                    cmds.append(temp_cmd)
                    lastoutputfile = outputprefix + ".segment-flagged.sam"
                    intermediates.append(lastoutputfile)
                    
                    (samtobam_cmd, outputfile, samtobam_tmpfiles) = generate_samtobam_command(input_sam=lastoutputfile, messy=messy, strip=False, local=local, outputprefix=outputprefix, double_dollar_sign=double_dollar_sign, sortmem=sortmem)
                    cmds.extend(samtobam_cmd)
                    intermediates.extend(samtobam_tmpfiles)
                
                    output_bam_files.append(outputfile)
                    
                    name = prefix + "-bwa_samse-" + str(i+1).zfill(4)
                    sam_job2 = MyJob(content=cmds, id=None, name=name, predecessors=[aln_job2], config_key=config_key, outputfiles=[outputfile])
                    bwa_sam_jobs.append(sam_job2)
                else:
                    #bwa sampe ...
                    cmds = []
                    #cmds.append("cd %s" % outdir)
                    outputprefix = outdir + os.sep + os.path.basename(readfile)
                    
                    config_key = 'r2c_bwa_sampe'
                    #r2c_bwa_sampe: -n50 -s -P ${CONTIGS} ${SAI_MATE1} ${SAI_MATE2} ${READFILE_MATE1} ${READFILE_MATE2}
                    args = {}
                    args['CONTIGS'] = contigs
                    args['SAI_MATE1'] = sai1
                    args['SAI_MATE2'] = sai2
                    args['READFILE_MATE1'] = readfile
                    args['READFILE_MATE2'] = mate_readfile
                    cmds.append('bwa sampe %s > %s.sam' % (cfg.create_args(settings, 'commands', config_key, args), outputprefix))
                    
                    lastoutputfile = outputprefix + ".sam"
                    intermediates.append(lastoutputfile)
                    
                    if xa2multi:
                        cmds.append("cat " + lastoutputfile + " | xa2multi.pl > " + outputprefix + ".xa2multi.sam")
                        lastoutputfile = outputprefix + ".xa2multi.sam"
                        intermediates.append(lastoutputfile)
                    
                    (samtobam_cmd, outputfile, samtobam_tmpfiles) = generate_samtobam_command(input_sam=lastoutputfile, messy=messy, strip=strip, local=local, outputprefix=outputprefix, double_dollar_sign=double_dollar_sign, sortmem=sortmem)
                    cmds.extend(samtobam_cmd)
                    intermediates.extend(samtobam_tmpfiles)
                
                    output_bam_files.append(outputfile)
                    
                    name = prefix + "-bwa_sampe-" + str(i).zfill(4) + "-" + str(i+1).zfill(4)
                    sam_job = MyJob(content=cmds, id=None, name=name, predecessors=[aln_job1, aln_job2], config_key=config_key, outputfiles=[outputfile])
                    bwa_sam_jobs.append(sam_job) 
            else: # use pipes whenever possible
                bam_mate_1_string = ''
                bam_mate_2_string = ''
                
                if is_bam(readfile):
                    cursor = index
                    bam_mate_1_string = "-b1"
                    bam_mate_2_string = "-b2"
                else: # fq, fq.gz, fastq, fastq.gz
                    mate_readfile = reads_paths[index+1]
                    cursor = index+1
                
                if samse:
                    # bwa aln,samse for mate 1
                    cmds = []
                    #cmds.append("cd %s" % outdir)
                    
                    #r2c_bwa_aln: -q 3 -t ${THREADS} ${BAMFLAG} ${CONTIGS} ${READFILE}
                    config_key = 'r2c_bwa_aln'                
                    args = {}
                    args['THREADS'] = str(threads)
                    args['BAMFLAG'] = bam_mate_1_string
                    args['CONTIGS'] = contigs
                    args['READFILE'] = readfile
                    temp_cmd = 'bwa aln %s |' % cfg.create_args(settings, 'commands', config_key, args)
                                  
                    #r2c_bwa_samse: -n50 ${CONTIGS} ${SAI} ${READFILE}
                    config_key = 'r2c_bwa_samse'
                    args = {}
                    args['CONTIGS'] = contigs
                    args['SAI'] = '-'
                    args['READFILE'] = readfile
                    temp_cmd += 'bwa samse %s' % cfg.create_args(settings, 'commands', config_key, args)
                
                    outputprefix = outdir + '/' + os.path.basename(readfile)
                
                    if xa2multi:
                        temp_cmd += " | xa2multi.pl "
                                       
                    temp_cmd += ' | awk -F\\t \'{OFS="\t"; if($1!="@SQ" && $1!="@HD" && $1!="@RG" && $1!="@PG" && $1!="@CO"){'
                    temp_cmd += '$2=or($2,0x40);' #mate1
                    if strip:
                        temp_cmd += '$11="*"' #remove QUAL
                    temp_cmd += '}; print}\' '
                    
                    (samtobam_cmd, outputfile, samtobam_tmpfiles) = generate_samtobam_command(temp_cmd, input_sam=None, messy=messy, strip=False, local=local, outputprefix=outputprefix, double_dollar_sign=double_dollar_sign, sortmem=sortmem)
                    cmds.extend(samtobam_cmd)
                    intermediates.extend(samtobam_tmpfiles)
                    
                    output_bam_files.append(outputfile)
                    
                    name = prefix + "-bwa_aln,samse-" + str(i).zfill(4)
                    my_predecessors = []
                    if indexing_job:
                        my_predecessors.append(indexing_job)
                    elif predecessors:
                        my_predecessors.extend(predecessors)
                    sam_job1 = MyJob(content=cmds, id=None, name=name, predecessors=my_predecessors, config_key="r2c_bwa_aln,sam", outputfiles=[outputfile])
                    bwa_sam_jobs.append(sam_job1)
                    
                    # bwa aln,samse for mate 2
                    cmds = []
                    #cmds.append("cd %s" % outdir)
                    
                    #r2c_bwa_aln: -q 3 -t ${THREADS} ${BAMFLAG} ${CONTIGS} ${READFILE}
                    config_key = 'r2c_bwa_aln'                
                    args = {}
                    args['THREADS'] = str(threads)
                    args['BAMFLAG'] = bam_mate_2_string
                    args['CONTIGS'] = contigs
                    args['READFILE'] = mate_readfile
                    temp_cmd = 'bwa aln %s |' % cfg.create_args(settings, 'commands', config_key, args)
                    
                    #r2c_bwa_samse: -n50 ${CONTIGS} ${SAI} ${READFILE}
                    config_key = 'r2c_bwa_samse'
                    args = {}
                    args['CONTIGS'] = contigs
                    args['SAI'] = '-'
                    args['READFILE'] = mate_readfile
                    temp_cmd += 'bwa samse %s' % cfg.create_args(settings, 'commands', config_key, args)
                
                    outputprefix = outdir + os.sep + os.path.basename(mate_readfile)
                
                    if xa2multi:
                        temp_cmd += " | xa2multi.pl "
                    
                    temp_cmd += ' | awk -F\\t \'{OFS="\t"; if($1!="@SQ" && $1!="@HD" && $1!="@RG" && $1!="@PG" && $1!="@CO"){'
                    temp_cmd += '$2=or($2,0x80);' #mate2
                    if strip:
                        temp_cmd += '$11="*"' #remove QUAL
                    temp_cmd += '}; print}\' '
                                         
                    (samtobam_cmd, outputfile, samtobam_tmpfiles) = generate_samtobam_command(temp_cmd, input_sam=None, messy=messy, strip=False, local=local, outputprefix=outputprefix, double_dollar_sign=double_dollar_sign, sortmem=sortmem)
                    cmds.extend(samtobam_cmd)
                    intermediates.extend(samtobam_tmpfiles)
                
                    output_bam_files.append(outputfile)
                    
                    name = prefix + "-bwa_aln,samse-" + str(i+1).zfill(4)
                    my_predecessors = []
                    if indexing_job:
                        my_predecessors.append(indexing_job)
                    elif predecessors:
                        my_predecessors.extend(predecessors)
                    sam_job2 = MyJob(content=cmds, id=None, name=name, predecessors=my_predecessors, config_key="r2c_bwa_aln,sam", outputfiles=[outputfile])
                    bwa_sam_jobs.append(sam_job2)
                else:
                    sai1 = outdir + os.sep + os.path.basename(readfile) + ".b1.sai"
                    sai2 = outdir + os.sep + os.path.basename(mate_readfile) + ".b2.sai"
                    sai_files.append(sai1)
                    sai_files.append(sai2)
                    
                    #bwa aln for mate 1
                    cmds = []
                    #cmds.append("cd %s" % outdir)
                    
                    #r2c_bwa_aln: -q 3 -t ${THREADS} ${BAMFLAG} ${CONTIGS} ${READFILE}
                    config_key = 'r2c_bwa_aln'                
                    args = {}
                    args['THREADS'] = str(threads)
                    args['BAMFLAG'] = bam_mate_1_string
                    args['CONTIGS'] = contigs
                    args['READFILE'] = readfile
                    cmds.append('bwa aln %s > %s' % (cfg.create_args(settings, 'commands', config_key, args), sai1))
                    name = prefix + "-bwa_aln-" + str(i).zfill(4)
                    my_predecessors = []
                    if indexing_job:
                        my_predecessors.append(indexing_job)
                    elif predecessors:
                        my_predecessors.extend(predecessors)
                    aln_job1 = MyJob(content=cmds, id=None, name=name, predecessors=my_predecessors, config_key=config_key, threads=threads, outputfiles=[sai1])
                    bwa_aln_jobs.append(aln_job1)
                    
                    #bwa aln for mate 2
                    cmds = []
                    #cmds.append("cd %s" % outdir)
                    
                    #r2c_bwa_aln: -q 3 -t ${THREADS} ${BAMFLAG} ${CONTIGS} ${READFILE}
                    config_key = 'r2c_bwa_aln'                
                    args = {}
                    args['THREADS'] = str(threads)
                    args['BAMFLAG'] = bam_mate_2_string
                    args['CONTIGS'] = contigs
                    args['READFILE'] = mate_readfile
                    cmds.append('bwa aln %s > %s' % (cfg.create_args(settings, 'commands', config_key, args), sai2))
                    name = prefix + "-bwa_aln-" + str(i+1).zfill(4)
                    my_predecessors = []
                    if indexing_job:
                        my_predecessors.append(indexing_job)
                    elif predecessors:
                        my_predecessors.extend(predecessors)
                    aln_job2 = MyJob(content=cmds, id=None, name=name, predecessors=my_predecessors, config_key=config_key, threads=threads, outputfiles=[sai2])
                    bwa_aln_jobs.append(aln_job2)
                                       
                    #bwa sampe ...
                    cmds = []
                    #cmds.append("cd %s" % outdir)
                    outputprefix = outdir + os.sep + os.path.basename(readfile)
                    
                    config_key = 'r2c_bwa_sampe'
                    #r2c_bwa_sampe: -n50 -s -P ${CONTIGS} ${SAI_MATE1} ${SAI_MATE2} ${READFILE_MATE1} ${READFILE_MATE2}
                    args = {}
                    args['CONTIGS'] = contigs
                    args['SAI_MATE1'] = sai1
                    args['SAI_MATE2'] = sai2
                    args['READFILE_MATE1'] = readfile
                    args['READFILE_MATE2'] = mate_readfile
                    temp_cmd = 'bwa sampe %s ' % cfg.create_args(settings, 'commands', config_key, args)
                                    
                    outputprefix = outdir + os.sep + os.path.basename(readfile)
                
                    if xa2multi:
                        temp_cmd += " | xa2multi.pl "
                    
                    (samtobam_cmd, outputfile, samtobam_tmpfiles) = generate_samtobam_command(temp_cmd, input_sam=None, messy=messy, strip=strip, local=local, outputprefix=outputprefix, double_dollar_sign=double_dollar_sign, sortmem=sortmem)
                    cmds.extend(samtobam_cmd)
                    intermediates.extend(samtobam_tmpfiles)
                    
                    output_bam_files.append(outputfile)
                    
                    name = prefix + "-bwa_sampe-" + str(i).zfill(4) + "-" + str(i+1).zfill(4)
                    sam_job = MyJob(content=cmds, id=None, name=name, predecessors=bwa_aln_jobs, config_key=config_key, outputfiles=[outputfile])
                    bwa_sam_jobs.append(sam_job)
            
            i += 2

    #Sanity check: If you have paired-end reads, you should have even number of SAI files
    num_bwa_aln_jobs = len(bwa_aln_jobs)
    if not single_end and num_bwa_aln_jobs > 0 and num_bwa_aln_jobs%2 > 0:
        sys.stderr.write("ERROR: Odd number of bwa_aln jobs (" + str(num_bwa_aln_jobs) + "). It looks like reads files are not paired properly." + os.linesep)
        sys.exit(1)
    
    intermediates.extend(sai_files)

    #3. Symlink or merge the sorted bam files
    [merge_job, tmpfiles] = generate_merge_bam_jobs(alignment_jobs=bwa_sam_jobs, output_bam_files=output_bam_files, outdir=outdir, prefix=prefix)
    intermediates.extend(tmpfiles)

    #4. Clean up
    cmds = []
    #cmds.append("cd %s" % outdir)
    ftbd = ' '.join(intermediates)
    #cmds.append("ls -1 %s" % ftbd)
    cmds.append("rm -f %s" % ftbd)
    name = prefix + "-cleanup-r2c"
    
    cleanup_predecessor = merge_job
    if not cleanup_predecessor:
        cleanup_predecessor = bwa_sam_jobs[0]
        
    cleanup_job = MyJob(content=cmds, name=name, predecessors=[cleanup_predecessor])

    jobs = []
    if indexing_job:
        jobs.append(indexing_job)
    jobs.extend(bwa_aln_jobs)
    jobs.extend(bwa_sam_jobs)
    if merge_job:
        jobs.append(merge_job)
    
    return (jobs, cleanup_job)


def generate_job_commands_bwasw(contigs=None, reads_paths=None, prefix=None, single_end=False, outdir=None, threads=1, local=False, strip=False, messy=False, sortmem=8000000000, indexed=False, predecessors=None, settings=None, double_dollar_sign=False):
    """
    Method to create the job for aligning reads to contigs with bwasw.
    
    Args:
        contigs      (str)    : the path of the contigs FASTA file
        reads_paths  ([str])  : the paths of read files
        prefix       (str)    : the prefix of output files
        single_end   (bool)   : the flag to indicate the reads are single-end data
        outdir       (str)    : the path of the output directory
        threads      (int)    : the number of threads to use for bwasw
        local        (bool)   : the flag to indicate the job would be run locally
        strip        (bool)   : the flag to strip the quality string from the output BAM file
        messy        (bool)   : the flag to not use piping and keep intermediate files
        sortmem      (int)    : the argument to `samtools sort -m`
        indexed      (bool)   : the flag to indicate the contigs FASTA file was already indexed
        predecessors ([job_runner.MyJob]) : the list of predecessor job objects
        settings     ({str:str:str})      : the Trans-ABySS configurations dictionary. {section : key : value}
        double_dollar_sign (bool)         : use '$$' instead '$' for referencing bash variables
                
    Returns:
        ([job_runner.MyJob], job_runner.MyJob) :
            The first element is a list of jobs to do the alignments and generate a final indexed BAM file.
            The second element is the job for cleaning up intermediate files.
    """

    intermediates = []

    indexing_job = None

    #1. index the contigs fasta file
    if not indexed:
        [indexing_job, indexfiles] = generate_bwa_indexing_job(contigs, settings, prefix, predecessors)
        intermediates.extend(indexfiles)

    #2. align the reads to contigs
    alignment_jobs = []
    output_bam_files = []
    
    task_id = 1
    items = reads_paths
    if not single_end:
        items = list(chunks(reads_paths, 2))
    
    for item in items:
        outputprefix = ''
        cmds = []
        lastoutputfile = None
        
        config_key = 'r2c_bwa_bwasw'
        #r2c_bwa_bwasw: -t ${THREADS} ${CONTIGS} ${READFILE_MATE1} ${READFILE_MATE2}
        args = {}
        args['THREADS'] = str(threads)
        args['CONTIGS'] = contigs        
        
        if single_end:                        
            args['READFILE_MATE1'] = item
            args['READFILE_MATE2'] = ''
            outputprefix = outdir + os.sep + os.path.basename(item)
            
            #cmds.append("cd %s" % outdir)
            temp_cmd = "bwa bwasw %s" % cfg.create_args(settings, 'commands', config_key, args)
        else:
            args['READFILE_MATE1'] = item[0]
            args['READFILE_MATE2'] = item[1]
            outputprefix = outdir + os.sep + os.path.basename(item[0])
            
            #cmds.append("cd %s" % outdir)
            temp_cmd = "bwa bwasw %s" % cfg.create_args(settings, 'commands', config_key, args)
        
        if messy: # generate intermediate files whenever possible
            temp_cmd += " > " + outputprefix + ".sam"            
            lastoutputfile = outputprefix + ".sam"
            intermediates.append(lastoutputfile)
            
        (samtobam_cmd, outputfile, samtobam_tmpfiles) = generate_samtobam_command(temp_cmd, input_sam=lastoutputfile, messy=messy, strip=strip, local=local, outputprefix=outputprefix, double_dollar_sign=double_dollar_sign, sortmem=sortmem)
        cmds.extend(samtobam_cmd)
        intermediates.extend(samtobam_tmpfiles)
        
        output_bam_files.append(outputfile)
            
        name = prefix + "-bwasw-" + str(task_id).zfill(4)
        my_predecessors = []
        if indexing_job:
            my_predecessors.append(indexing_job)
        elif predecessors:
            my_predecessors.extend(predecessors)
        alignment_jobs.append(MyJob(content=cmds, id=None, name=name, predecessors=my_predecessors, config_key=config_key, threads=threads, outputfiles=[outputfile]))
        
        task_id += 1
        
    #3. Symlink or merge the sorted bam files
    [merge_job, tmpfiles] = generate_merge_bam_jobs(alignment_jobs=alignment_jobs, output_bam_files=output_bam_files, outdir=outdir, prefix=prefix)
    intermediates.extend(tmpfiles)

    #4. Clean up
    cmds = []
    #cmds.append("cd %s" % outdir)
    ftbd = ' '.join(intermediates)
    #cmds.append("ls -1 %s" % ftbd)
    cmds.append("rm -f %s" % ftbd)
    name = prefix + "-cleanup-r2c"
    
    cleanup_predecessor = merge_job
    if not cleanup_predecessor:
        cleanup_predecessor = alignment_jobs[0]
        
    cleanup_job = MyJob(content=cmds, name=name, predecessors=[cleanup_predecessor])

    jobs = []
    if indexing_job:
        jobs.append(indexing_job)
    jobs.extend(alignment_jobs)
    if merge_job:
        jobs.append(merge_job)
    
    return (jobs, cleanup_job)


def generate_job_commands_bwamem(contigs=None, reads_paths=None, prefix=None, single_end=False, outdir=None, threads=1, local=False, strip=False, messy=False, sortmem=800000000, indexed=False, predecessors=None, settings=None, double_dollar_sign=False):
    """
    Method to create the job for aligning reads to contigs with bwa mem.
    
    Args:
        contigs      (str)    : the path of the contigs FASTA file
        reads_paths  ([str])  : the paths of read files
        prefix       (str)    : the prefix of output files
        single_end   (bool)   : the flag to indicate the reads are single-end data
        outdir       (str)    : the path of the output directory
        threads      (int)    : the number of threads to use for bwa mem
        local        (bool)   : the flag to indicate the job would be run locally
        strip        (bool)   : the flag to strip the quality string from the output BAM file
        messy        (bool)   : the flag to not use piping and keep intermediate files
        sortmem      (int)    : the argument to `samtools sort -m`
        indexed      (bool)   : the flag to indicate the contigs FASTA file was already indexed
        predecessors ([job_runner.MyJob]) : the list of predecessor job objects
        settings     ({str:str:str})      : the Trans-ABySS configurations dictionary. {section : key : value}
        double_dollar_sign (bool)         : use '$$' instead '$' for referencing bash variables
                
    Returns:
        ([job_runner.MyJob], job_runner.MyJob) :
            The first element is a list of jobs to do the alignments and generate a final indexed BAM file.
            The second element is the job for cleaning up intermediate files.
    """

    intermediates = []

    indexing_job = None

    #1. index the contigs fasta file
    if not indexed:
        [indexing_job, indexfiles] = generate_bwa_indexing_job(contigs, settings, prefix, predecessors)
        intermediates.extend(indexfiles)

    #2. align the reads to contigs
    alignment_jobs = []
    output_bam_files = []
    
    task_id = 1
    items = reads_paths
    if not single_end:
        items = list(chunks(reads_paths, 2))
    
    for item in items:
        outputprefix = ''
        cmds = []
        
        config_key = 'r2c_bwa_mem'
        #r2c_bwa_bwamem: -t ${THREADS} ${CONTIGS} ${READFILE_MATE1} ${READFILE_MATE2}
        args = {}
        args['THREADS'] = str(threads)
        args['CONTIGS'] = contigs        
        
        if single_end:                        
            args['READFILE_MATE1'] = item
            args['READFILE_MATE2'] = ''
            outputprefix = outdir + os.sep + os.path.basename(item)
            
            #cmds.append("cd %s" % outdir)
            temp_cmd = "bwa mem %s" % cfg.create_args(settings, 'commands', config_key, args)
        else:
            args['READFILE_MATE1'] = item[0]
            args['READFILE_MATE2'] = item[1]
            outputprefix = outdir + os.sep + os.path.basename(item[0])
            
            #cmds.append("cd %s" % outdir)
            temp_cmd = "bwa mem %s" % cfg.create_args(settings, 'commands', config_key, args)
        
        lastoutputfile = None
        
        if messy: # generate intermediate files whenever possible
            temp_cmd += " > " + outputprefix + ".sam"
            lastoutputfile = outputprefix + ".sam"
            intermediates.append(lastoutputfile)
                        
        (samtobam_cmd, outputfile, samtobam_tmpfiles) = generate_samtobam_command(temp_cmd, input_sam=lastoutputfile, messy=messy, strip=strip, local=local, outputprefix=outputprefix, double_dollar_sign=double_dollar_sign, sortmem=sortmem)
        cmds.extend(samtobam_cmd)
        intermediates.extend(samtobam_tmpfiles)
        
        output_bam_files.append(outputfile)
            
        name = prefix + "-bwamem-" + str(task_id).zfill(4)
        my_predecessors = []
        if indexing_job:
            my_predecessors.append(indexing_job)
        elif predecessors:
            my_predecessors.extend(predecessors)
        alignment_jobs.append(MyJob(content=cmds, id=None, name=name, predecessors=my_predecessors, config_key=config_key, threads=threads, outputfiles=[outputfile]))
        
        task_id += 1
        
    #3. Symlink or merge the sorted bam files
    [merge_job, tmpfiles] = generate_merge_bam_jobs(alignment_jobs=alignment_jobs, output_bam_files=output_bam_files, outdir=outdir, prefix=prefix)
    intermediates.extend(tmpfiles)  

    #4. Clean up
    cmds = []
    #cmds.append("cd %s" % outdir)
    ftbd = ' '.join(intermediates)
    #cmds.append("ls -1 %s" % ftbd)
    cmds.append("rm -f %s" % ftbd)
    name = prefix + "-cleanup-r2c"
    
    cleanup_predecessor = merge_job
    if not cleanup_predecessor:
        cleanup_predecessor = alignment_jobs[0]
        
    cleanup_job = MyJob(content=cmds, name=name, predecessors=[cleanup_predecessor])

    jobs = []
    if indexing_job:
        jobs.append(indexing_job)
    jobs.extend(alignment_jobs)
    if merge_job:
        jobs.append(merge_job)
    
    return (jobs, cleanup_job)


parser = argparse.ArgumentParser(description='Align reads to contigs.')

parser.add_argument('contigsfasta', metavar='FASTA', type=str, help='path to contigs fasta file')
parser.add_argument('readslist', metavar='READS_LIST', type=str, help='path to text file listing the paths to input reads files')
parser.add_argument('aligner', metavar='ALIGNER', type=str, help='name of aligner; one of %s' % str(ALIGNERS))

sample_group = parser.add_mutually_exclusive_group(required=True)
sample_group.add_argument('--genome', action='store_true', default=False, dest='genome', help='use configurations from \'genome.cfg\'')
sample_group.add_argument('--transcriptome', action='store_true', default=False, dest='transcriptome', help='use configurations from \'transcriptome.cfg\'')

abyssmap_group = parser.add_argument_group("abyss-map options")
abyssmap_group.add_argument('--k', type=int, default=24, dest='k', help='k-mer size [default: %(default)s]', metavar='INT')
    
bwa_group = parser.add_argument_group("bwa options")
bwa_group.add_argument('--samse', action='store_true', default=False, dest='samse', help="use 'samse' instead of 'sampe' for paired-end reads")
bwa_group.add_argument('--xa2multi', action='store_true', default=False, dest='xa2multi', help='run xa2multi.pl')

parser.add_argument('-c', '--cleanup', action='store_true', default=False, dest='cleanup', help='remove intermediate files when all jobs are done')
parser.add_argument('-C', '--cluster', dest='cluster', help='submit jobs to cluster head node', metavar='STR')
parser.add_argument('-d', '--debug', dest='debug', help='run in debug mode; no jobs will be submitted', action='store_true', default=False)
parser.add_argument('-i', '--indexed', action='store_true', default=False, dest='indexed', help='reference is already indexed')
#    parser.add_argument('-I', '--index-only', action='store_true', default=False, dest='index_only', help='only index the reference; no alignments')
parser.add_argument('-j', '--jobdir', dest='jobdir', help='directory for job scripts and logs', metavar='DIR')   
parser.add_argument('-l', '--local', action='store_true', default=False, dest='local', help='run aligner on local machine; no cluster jobs')
parser.add_argument('-m', '--sort-mem', type=int, default=800000000, dest='sortmem', help='value for `samtools sort -m` [default: %(default)s]', metavar='INT')
parser.add_argument('-o', '--outdir', dest='outdir', help='output directory', metavar='DIR')
parser.add_argument('-p', '--prefix', dest='prefix', help='library name; used as prefix of output files', metavar='STR')
parser.add_argument('-P', '--project', dest='project', help='project name in configuration file (\'genome.cfg\', \'transcriptome.cfg\')', metavar='STR')
parser.add_argument('-q', '--queue', dest='queue', help='cluster queue', metavar='STR')
parser.add_argument('-s', '--single_end', action='store_true', default=False, dest='single_end', help='input reads files are single-end')
parser.add_argument('-S', '--strip', action='store_true', default=False, dest='strip', help='strip the QUAL columns in SAM records')
parser.add_argument('-t', '--threads', type=int, default=4, dest='threads', help='max number of threads to use for multi-threaded programs [default: %(default)s]', metavar='INT')
parser.add_argument('-w', '--wasteful', dest='messy', help='will not pipe commands together and will keep all intermediate files; overrides --cleanup', action='store_true', default=False)
parser.add_argument('--qmake', type=str, dest='makefile', help='create a makefile at PATH for the jobs; no jobs would be run', metavar='PATH')
parser.add_argument('--version', action='version', version='%(prog)s 1.4.8 (Build 201306181452)')


def parseArgsFromCommand(command):
    """
    Get the options used by parsing the command line.
    
    Args:
        command (str) : the command line
        
    Returns:
        (R2C_Factory) : the job factory object
    """
    args = parser.parse_args(command.split())
    return createFactoryFromArgs(args)


def createFactoryFromArgs(args):
    """
    Create the jobs based on the options used.
    
    Args:
        args (argparse.Namespace) : the argparse Namespace
        
    Returns:
        (R2C_Factory) : the job factory object
    """

    source_contigs = args.contigsfasta
    if not os.path.isfile(source_contigs):
        sys.stderr.write("ERROR: Cannot find contigs FASTA file, " + source_contigs + os.linesep)
        sys.exit(1)

    if not args.aligner in ALIGNERS:
        sys.stderr.write("ERROR: Unknown aligner specified, %s" % args.aligner + os.linesep)
        sys.exit(1)

    readslist = args.readslist
    if not os.path.isfile(readslist):
        sys.stderr.write("ERROR: Cannot find reads list file, " + readslist + os.linesep)
        sys.exit(1)
    
    fh = open(readslist, 'r')
    reads_paths = []
    for line in fh:
        linestripped = line.strip()
        if len(linestripped) > 0:
            if os.path.isfile(linestripped):
                reads_paths.append(linestripped)
            else:
                sys.stderr.write("ERROR: Cannot find reads file, " + linestripped + os.linesep)
                sys.exit(1)
    fh.close()

    return R2C_Factory(contigs=source_contigs, reads_paths=reads_paths, aligner=args.aligner, prefix=args.prefix, outdir=args.outdir, jobdir=args.jobdir, debug=args.debug, cleanup=args.cleanup, transcriptome=args.transcriptome, genome=args.genome, threads=args.threads, single_end=args.single_end, local=args.local, xa2multi=args.xa2multi, strip=args.strip, messy=args.messy, sortmem=args.sortmem, k=args.k, project=args.project, cluster=args.cluster, queue=args.queue, indexed=args.indexed, samse=args.samse, makefile=args.makefile)

    
def __main__():
    """
    The main method for running this script standalone.
    """

    sys.stdout.flush() #flush print outputs to screen
    
    args = parser.parse_args()

    factory = createFactoryFromArgs(args)
    if args.makefile:
        factory.qmake()
    else:
        factory.create_and_submit_jobs() 

if __name__ == "__main__":
    __main__()
