import os
import sys
import random
import argparse
import glob

def print_message(message_type, message, quit=False):
    sys.stderr.write(str(message_type) + ": " + str(message) + os.linesep)
    if quit:
        sys.exit(1)

class FastaSeq:
    def __init__(self, header=None, seq=None):
        self.header = header
        self.seq = seq

def split_fasta(fasta_file, output_dir, verbose, max_bases=None, max_contigs=None, randomize=False):
    completion_stamp = "%s/SPLIT_FASTA.COMPLETE" % (output_dir)
    if os.path.exists(completion_stamp):
        #This could be a re-run; remove the completion stamp.
        os.remove(completion_stamp)
    
    for fa in glob.glob("%s/seq.*.fa" % (output_dir)):
        #This could be a re-run; remove the existing output file.
        os.remove(fa)

    total_num_contigs = 0
    total_num_bases = 0

    fasta = []

    currentseq = None
    output_num = 1
    num_contigs = 0
    num_bases = 0

    fin = open(fasta_file, 'r')    
    fout = open("%s/seq.%d.fa" % (output_dir, output_num), 'w')
    
    for line in fin:
        if line[0] == '>':
            if currentseq:
                if currentseq.header:
                    if currentseq.seq:
                        # The current fasta sequence is now complete
                        
                        if randomize:
                            # put this fasta seq in the list
                            fasta.append(currentseq)
                        else:
                            # write this fasta seq to file                            
                            current_seq_length = len(currentseq.seq)
                            
                            # check if adding the current sequence to this batch would be exceed ANY thresholds of this batch
                            if (max_contigs and num_contigs + 1 > max_contigs) or (max_bases and num_bases + current_seq_length > max_bases):
                                # threshold(s) exceeded, write this sequence to the next batch
                                if verbose:
                                    print_message('seq.%d.fa' % output_num, "%d bases in %d contigs" % (num_bases, num_contigs), quit=False)
                                fout.close()
                                total_num_contigs += num_contigs
                                total_num_bases += num_bases
                                output_num += 1
                                fout = open("%s/seq.%d.fa" % (output_dir, output_num), 'w')
                                # reset the counters
                                num_contigs = 0
                                num_bases = 0
                            
                            fout.write('%s\n%s\n' % (currentseq.header, currentseq.seq))
                            num_contigs += 1
                            num_bases += current_seq_length
                        currentseq = None
                    else:
                        print_message('ERROR', 'FASTA sequence has no sequence string:\n%s\n%s' % (currentseq.header, line.rstrip()), quit=True)
            currentseq = FastaSeq(header=line.rstrip())
        else:
            if currentseq:            
                if currentseq.seq:
                    # concatenate the sequence lines
                    currentseq.seq += line.rstrip()
                else:
                    currentseq.seq = line.rstrip()
            else:
                print_message('ERROR', 'FASTA sequence has no header:\n%s' % line.rstrip(), quit=True)
    fin.close()
    
    #sanity check
    if not currentseq.seq:
        print_message('ERROR', 'FASTA sequence has no sequence string:\n%s' % currentseq.header, quit=True)
    
    if len(fasta) > 0 and randomize:
        # Add the last fasta seq to the list
        fasta.append(currentseq)
         
        random.shuffle(fasta)
        
        for f in fasta:
            current_seq_length = len(f.seq)
            
            # check if adding the current sequence to this batch would be exceed ANY thresholds of this batch
            if (max_contigs and num_contigs + 1 > max_contigs) or (max_bases and num_bases + current_seq_length > max_bases):
                # threshold(s) exceeded, write this sequence to the next batch
                if verbose:
                    print_message('seq.%d.fa' % output_num, "%d bases in %d contigs" % (num_bases, num_contigs), quit=False)
                fout.close()
                total_num_contigs += num_contigs
                total_num_bases += num_bases
                output_num += 1
                fout = open("%s/seq.%d.fa" % (output_dir, output_num), 'w')
                # reset the counters
                num_contigs = 0
                num_bases = 0
            
            fout.write('%s\n%s\n' % (f.header, f.seq))
            num_contigs += 1
            num_bases += current_seq_length
    else:
        # Write the last fasta seq to file
        current_seq_length = len(currentseq.seq)
        
        # check if adding the last sequence to this batch would be exceed ANY thresholds of this batch
        if (max_contigs and num_contigs + 1 > max_contigs) or (max_bases and num_bases + current_seq_length > max_bases):
            # threshold(s) exceeded, write this sequence to the next batch
            if verbose:
                print_message('seq.%d.fa' % output_num, "%d bases in %d contigs" % (num_bases, num_contigs), quit=False)
            fout.close()
            total_num_contigs += num_contigs
            total_num_bases += num_bases
            output_num += 1
            fout = open("%s/seq.%d.fa" % (output_dir, output_num), 'w')
            # reset the counters
            num_contigs = 0
            num_bases = 0
        
        fout.write('%s\n%s\n' % (currentseq.header, currentseq.seq))
        num_contigs += 1
        num_bases += current_seq_length

    if verbose:
        print_message('seq.%d.fa' % output_num, "%d bases in %d contigs" % (num_bases, num_contigs), quit=False)
    fout.close()
    total_num_contigs += num_contigs
    total_num_bases += num_bases
    
    print_message('COMPLETE', "Read %d bases in %d contigs and wrote %d files" % (total_num_bases, total_num_contigs, output_num), quit=False)
    #STAMP for completion
    fout = open(completion_stamp, 'w')    
    fout.close()

if __name__ == '__main__':   
    parser = argparse.ArgumentParser(description='Split a FASTA file into multiple batches')
    parser.add_argument('fasta', metavar='FASTA', type=str, help='path to FASTA file to split')
    parser.add_argument("-c", "--max_contigs", dest="max_contigs", help="max number of contigs per batch.", type=int, metavar='INT')
    parser.add_argument("-b", "--max_bases", dest="max_bases", help="max number of bases per batch.", type=int, metavar='INT')    
    parser.add_argument("-o", "--output_dir", dest="output_dir", help="output directory [default: cwd]. Output files are 'seq.*.fa'.", metavar='DIR')
    parser.add_argument("-r", "--randomize", dest="randomize", help="randomize the order of sequences", action="store_true", default=False)
    parser.add_argument("-v", "--verbose", dest="verbose", help="print verbose information", action="store_true", default=False)
    
    args = parser.parse_args()

    if not os.path.isfile(args.fasta):
        print_message('ERROR', 'Cannot find file: %s' % args.fasta, quit=True)
    else:
        print_message('input fasta', args.fasta, quit=False)

    if args.max_bases:
        if args.max_bases <= 0:
            print_message('ERROR', 'The max number of bases per batch specified (%s) is not a positive integer.' % args.max_bases, quit=True)
        elif args.verbose:
            print_message('max number of bases per batch', args.max_bases, quit=False)
    
    if args.max_contigs:
        if args.max_contigs <= 0:
            print_message('ERROR', 'The max number of contigs per batch specified (%s) is not a positive integer.' % args.max_contigs, quit=True)
        elif args.verbose:
            print_message('max number of contigs per batch', args.max_contigs, quit=False)
    
    outdir = None
    if args.output_dir:
        if not os.path.isdir(args.output_dir):            
            print_message('WARNING', 'Output directory does not exist. Creating output directory...', quit=False)
            os.makedirs(args.output_dir)
        outdir = args.output_dir
    else:
        outdir = os.getcwd()
        print_message('WARNING', 'No output directory was specified. Will create output files in the current directory...', quit=False)        
    
    if args.verbose:
        print_message('output directory', outdir, quit=False)
        
    split_fasta(args.fasta, outdir, args.verbose, max_bases=args.max_bases, max_contigs=args.max_contigs, randomize=args.randomize)
    
