#!/usr/bin/env python

import os
import sys
import pwd
import getpass
import stat
import socket
import random
import shutil
import glob
import grp
import math
import re
from datetime import datetime
from optparse import OptionParser
from optparse import OptionGroup
from gsc import qmake
from utilities import cfg
from utilities import job_runner
from utilities import r2c
from utilities import prepare_reads
from utilities.job_runner import MyJob

try:
    PACKAGE_DIR = os.environ['TRANSABYSS_PATH']
except KeyError:
    message = "Cannot find TRANSABYSS_PATH in your environment!"
    print_message("ERROR", message, True)

try:
    VERSION = os.environ['TRANSABYSS_VERSION']
    BUILDID = VERSION + ' (public build 201309131152)'
except KeyError:
    message = "Cannot find TRANSABYSS_VERSION in your environment!"
    print_message("ERROR", message, True)
    

def print_message(message_type, message, quit=False):
    sys.stderr.write(str(message_type) + ": " + str(message) + os.linesep)
    if quit:
        sys.exit(1)

def is_pos_int(q):
    integer = 0
    try:
        integer = int(q)
    except TypeError:
        return False
    return integer > 0
    
class Tee(object):
    def __init__(self, name, mode):
        if os.path.isfile(name):
            stat_info = os.stat(name)
            uid = stat_info.st_uid
            user = pwd.getpwuid(uid)[0]
            if not user == getpass.getuser():
                # Back up this file as my file
                shutil.copyfile(name, name + '.tee')
                # Remove the original file
                os.remove(name)
                # Recover the file from the backup
                shutil.move(name + '.tee', name)
                                
        self.file = open(name, mode)
                
        permission = stat.S_IRUSR | stat.S_IWUSR # owner read/write
        permission = permission | stat.S_IRGRP | stat.S_IWGRP # group read/write
        os.chmod(name, permission)
        self.stdout = sys.stdout
        self.stderr = sys.stderr
        sys.stdout = self
        sys.stderr = self        
    def finalize(self):
        if self.stdout:        
            sys.stdout = self.stdout
        if self.stderr:
            sys.stderr = self.stderr
        if self.file and not file.closed:
            self.file.flush()
            self.file.close()
            self.file = None
        self.stdout = None
        self.stderr = None
    def write(self, data):
        self.file.write(data)
        self.stdout.write(data)
        #self.stderr.write(data)

class Pipe:
    def __init__(self, assembly_method, project, lib, readlength, assembly_dir=None, cluster=None, genome=None, queue=None, local=False, debug=False, top_dir=None):
        self.assembly_method = assembly_method
        self.project = project
        self.lib = lib
        self.version = VERSION

        self.top_dir = top_dir
        self.working_dir = None
        self.assembly_dir = assembly_dir
        self.ks = None
        
        #self.analysis = []
        self.genome = genome
        self.cluster = cluster
        self.queue = queue
        self.readlength = readlength
        self.local = local
        self.debug = debug
        
    def set_working_dir(self, cfg_file):
        self.settings = cfg.initialize_settings(cfg_file, project=self.project)
                    
        if not self.top_dir:
            try:
                if not self.project:
                    print_message('ERROR', 'Project name is not specified.', True)
                self.top_dir = self.settings[self.project]['topdir']
            except KeyError:
                print_message('ERROR', 'Cannot find top directory for project "%s" in configuration file "%s"' % (self.project, cfg_file), True)
        
        if not os.path.isdir(self.top_dir):
            print_message('WARNING', 'Top directory does not exist: "%s"' % (self.top_dir), False) 
            print 'Creating top directory...'
            os.makedirs(self.top_dir)
        
        self.working_dir = os.path.join (self.top_dir, self.lib, 'trans-abyss-v%s' % self.version)
    
    def setup(self, source=None):
        if self.working_dir is None:
            print_message('ERROR', 'Cannot find working directory for project.', True)
            #return False
        
        print 'Creating directories...'
        if not os.path.isdir(self.working_dir):
            os.makedirs(self.working_dir)
        
        if source is None:
            folders = ['assembly', 'merge', 'reads_to_contigs', 'contigs_to_genome', 'reads_to_genome']
            for folder in folders:
                destdir = os.path.join(self.working_dir, folder)
                if not os.path.isdir(destdir):
                    os.mkdir(destdir)
            
            print 'Creating symlinks to assembly directories...'        
            self.link_assemblies()
                    
        else:
            # make the top directory for each step, symlink the important files and directories
            important_files = []
            folders = []
            
            if self.assembly_method == 'stranded':
                folders = ['merge',
                           'merge/minus_strand',
                           'merge/plus_strand',
                           'merge/unknown',
                           'reads_to_contigs',
                           'reads_to_contigs/minus_strand',
                           'reads_to_contigs/plus_strand',
                           'reads_to_contigs/unknown',
                           'contigs_to_genome',
                           'reads_to_genome']
                important_files = ['merge/%s-contigs.fa' % self.lib,
                                   'merge/%s_minus-contigs.fa' % self.lib,
                                   'merge/%s_plus-contigs.fa' % self.lib,
                                   'reads_to_contigs/%s.in' % self.lib,
                                   'reads_to_contigs/minus_strand/%s.in' % self.lib,
                                   'reads_to_contigs/plus_strand/%s.in' % self.lib,
                                   'reads_to_contigs/unknown/%s.in' % self.lib,
                                   'reads_to_contigs/minus_strand/in.raw' % self.lib,
                                   'reads_to_contigs/plus_strand/in.raw' % self.lib,
                                   'reads_to_contigs/unknown/in.raw' % self.lib,
                                   'reads_to_contigs/%s-contigs.bam' % self.lib,
                                   'reads_to_contigs/minus_strand/%s-contigs.bam' % self.lib,
                                   'reads_to_contigs/plus_strand/%s-contigs.bam' % self.lib,
                                   'reads_to_contigs/unknown/%s-contigs.bam' % self.lib,
                                   'reads_to_contigs/%s-contigs.bam.bai' % self.lib,
                                   'reads_to_contigs/minus_strand/%s-contigs.bam.bai' % self.lib,
                                   'reads_to_contigs/plus_strand/%s-contigs.bam.bai' % self.lib,
                                   'reads_to_contigs/unknown/%s-contigs.bam.bai' % self.lib,]
            else:
                folders = ['merge',
                           'reads_to_contigs',
                           'contigs_to_genome',
                           'reads_to_genome']
                important_files = ['merge/%s-contigs.fa' % self.lib,
                                   'reads_to_contigs/%s.in' % self.lib,
                                   'reads_to_contigs/%s-contigs.bam' % self.lib,
                                   'reads_to_contigs/%s-contigs.bam.bai' % self.lib]  
            
            for folder in folders:
                destdir = os.path.join(self.working_dir, folder)
                if not os.path.isdir(destdir):
                    os.mkdir(destdir)
            
            source = '%s/%s/trans-abyss-v%s' % (self.top_dir, self.lib, str(options.source))
            
            print 'Creating symlinks to files and directories from the source version (%s)...' % str(options.source)
            #make relative symlink
            os.symlink(os.path.relpath('%s/assembly' % source, self.working_dir), '%s/assembly' % self.working_dir)
                        
            for f in important_files:
                sourcefile = '%s/%s' % (source, f)
                destfile = '%s/%s' % (self.working_dir, f)
                if os.path.isfile(sourcefile):
                    #make relative symlinks
                    (head, tail) = os.path.split(destfile)
                    os.symlink(os.path.relpath(sourcefile, head), destfile)
                # Don't do anything if the source file doesn't exist
                # because the corresponding step might not have been run, yet!
            
            if os.path.isdir('%s/contigs_to_genome/%s-contigs' % (source, self.lib)):
                contigs_to_genome_dir = '%s/contigs_to_genome/%s-contigs' % (self.working_dir, self.lib)
                os.makedirs(contigs_to_genome_dir)
                for folder in ('input', 'output'):
                    #make relative symlinks
                    os.symlink(os.path.relpath('%s/contigs_to_genome/%s-contigs/%s' % (source, self.lib, folder), contigs_to_genome_dir), '%s/%s' % (contigs_to_genome_dir, folder))
            
            if os.path.isdir('%s/reads_to_genome' % source):
                for b in glob.glob('%s/reads_to_genome/*.bam*' % source):
                    os.symlink(b, '%s/reads_to_genome/%s' % (self.working_dir, os.path.basename(b)) )
        
                
    def get_ks(self, path):
        ks = {}
        for assembly in glob.glob(os.path.join(path, 'k*')):            
            if os.path.isdir(assembly):
                # this is a directory with name starting with 'k'
            
                k_dirname = os.path.basename(assembly)
                k = k_dirname[1:]               
                
                if k.isdigit():
                    # the name is kZ, where Z is an integer
                    # user Z as the kmer value for this assembly
                    ks[k] = k_dirname                
                elif ',' in k:                                                        
                    kmers = k.split(',')
                    num_non_kmers = len(kmers)
                    
                    for k in kmers:
                        if k.isdigit():
                            num_non_kmers -= 1
                    
                    if num_non_kmers == 0:
                        # the name is kX,Y,Z where X,Y,Z are numbers
                        # use Z as the kmer value for this assembly
                        ks[kmers[-1]] = k_dirname
        return ks
        
    def link_assemblies(self):
        #print self.assembly_dir, self.working_dir
        
        target_assembly_dir = "%s/assembly" % (self.working_dir)
        if not os.path.isdir(target_assembly_dir):
            os.makedirs(target_assembly_dir)
            
        self.ks = self.get_ks(self.assembly_dir)
        ks = self.ks.keys()
        ks.sort(key=int)
        for k in ks:
            source_k_assembly = "%s/%s" % (self.assembly_dir, self.ks[k])
            target_k_assembly = "%s/k%s" % (target_assembly_dir, k)
            if not os.path.exists(target_k_assembly):
                os.symlink(source_k_assembly, target_k_assembly)
        
        #We look for "$LIB.in", then "in", then *.in
        target_in = "%s/in" % (target_assembly_dir)
        infile = None
        if os.path.isfile(os.path.join(self.assembly_dir, '%s.in' % self.lib)):
            infile = os.path.join(self.assembly_dir, '%s.in' % self.lib)
        elif os.path.isfile(os.path.join(self.assembly_dir, 'in')):
            infile = os.path.join(self.assembly_dir, 'in')
        else:        
            in_files = glob.glob(os.path.join(self.assembly_dir, '*.in'))
            num_ins = len(in_files)
            if num_ins == 1:
                infile = in_files[0]
            elif num_ins > 1:                
                print_message('WARNING', 'More than one "in" files found:\n%s' % str(in_files), False)
            else:
                print_message('WARNING', 'Cannot find "in" file.', False)
        
        if infile:
            if not os.path.exists(target_in):
                shutil.copyfile(infile, target_in)
            else:
                print_message('WARNING', '%s already exists.' % target_in, False)
        else:
            print_message('WARNING', '%s will not be created.' % target_in, False)
    
    def run_jobs(self, jobs, local_prefix, cluster_dir, workingdir=None):
        if not os.path.isdir(cluster_dir):
            os.makedirs(cluster_dir)
        if self.local:
            job_runner.run_locally(jobs, local_prefix, logdir=cluster_dir, debug=self.debug, workingdir=workingdir)
        else:
            job_runner.submit_jobs(jobs, self.settings, self.cluster, logdir=cluster_dir, debug=self.debug, workingdir=workingdir)
                                    
    def merge(self, input_dir, depend=[]):
        if not self.ks:
            self.ks = self.get_ks("%s/assembly" % (self.working_dir))
    
        task = 'abyss-rmdups-iterative'
        script = PACKAGE_DIR + "/wrappers/" + task
                
        args = {}
        args['LIB'] = self.lib
        args['READLENGTH'] = self.readlength
        args['INPUT_DIR'] = input_dir
        args['OUTPUT_DIR'] = '%s/merge' % (self.working_dir)
        args['K'] = max(self.ks.keys())
        
        cluster_dir = '%s/merge/cluster' % (self.working_dir)
        if not os.path.isdir(cluster_dir):
            os.makedirs(cluster_dir)        

        mergedfasta = args['OUTPUT_DIR'] + os.sep + self.lib + '-contigs.fa'
        if os.path.exists(mergedfasta) and os.path.islink(mergedfasta):
            # This is a re-run. The symlink to the file from the previous version must be removed.
            os.remove(mergedfasta)

        job_name = '%s-merge' % self.lib
        job_content = 'time %s %s' % (script, cfg.create_args(self.settings, 'commands', task, args) )
        job = MyJob(content=[job_content], name=job_name, config_key=task, predecessors=depend, queue=self.queue)
        self.run_jobs([job], job_name, cluster_dir, workingdir=self.working_dir)
        return job
        
    def calculate_assembly_stats(self, depend=[]):
        task = 'abyss-fac'
        cluster_dir = '%s/merge/cluster' % self.working_dir
        job_name = '%s-stats' % self.lib        

        if self.readlength:
            job_content = 'time abyss-fac -j -t %s assembly/k*/*-contigs.fa merge/*-contigs.fa > merge/stats.txt' % (self.readlength)
        else:
            job_content = 'time abyss-fac -j assembly/k*/*-contigs.fa merge/*-contigs.fa > merge/stats.txt'
        
        job = MyJob(content=[job_content], name=job_name, config_key=task, queue=self.queue, predecessors=depend)
        self.run_jobs([job], job_name, cluster_dir, workingdir=self.working_dir)

    def prepare_contigs(self, depend=[]):
        c2g_dir = "%s/contigs_to_genome/%s-contigs" % (self.working_dir, self.lib)
        input_dir = "%s/input" % (c2g_dir)
        cluster_dir = "%s/cluster" % (c2g_dir)
        
        if not os.path.isdir(cluster_dir):
            os.makedirs(cluster_dir)
            
        if os.path.exists(input_dir) and os.path.islink(input_dir):
            # This is a re-run. The symlink to the "input" directory from the previous version must be removed.
            os.remove(input_dir)
            
        if os.path.isdir(input_dir):
            fasta_paths = glob.glob(os.path.join(input_dir + '/seq.*.fa'))
            num_fastas = len(fasta_paths)
            if num_fastas > 0:
                print_message('WARNING', 'Removing %d \'seq.*.fa\' found in %s ...' % (num_fastas, input_dir), False)
                for fasta in fasta_paths:
                    os.remove(fasta)
        else:
            os.makedirs(input_dir)
    
        merged_contigs = '%s/merge/%s-contigs.fa' % (self.working_dir, self.lib)
        if not os.path.isfile(merged_contigs):
            print_message('ERROR', 'Cannot find %s ...' % (merged_contigs), True)
    
        task = 'split_fasta.py'
        args = {}
        args['FASTA'] = merged_contigs
        args['OUTPUT_DIR'] = input_dir            
        
        script = PACKAGE_DIR + "/utilities/split_fasta.py"
        
        job_name = '%s-split_fasta' % self.lib
        job_content = 'time python %s %s' % (script, cfg.create_args(self.settings, 'commands', task, args))
        
        job = MyJob(content=[job_content], name=job_name, config_key=task, queue=self.queue, predecessors=depend)            
        self.run_jobs([job], job_name, cluster_dir, workingdir=self.working_dir)
            
    def copy_genome_bam(self):
        # FOR EXTERNAL TRANSABYSS USERS
        
        #############################################################################################
        # @TODO: put your code here to copy your JAGuaR BAM file into the "reads_to_genome" directory
        #############################################################################################
        message = "Please modify 'trans-abyss.py' (def copy_genome_bam) to support copying your JAGuaR/GSNAP *.bam and *.bam.bai files to the 'reads_to_genome' directory."
        print_message("ERROR", message, quit=True)

    def process(self, options):    
        # Action Group 1
        if options.set_up:
            tee = Tee('%s/log.txt' % self.top_dir, 'a')
            print '[%s %s@%s]: %s (%s)' % (str(datetime.now()), getpass.getuser(), socket.gethostname(), self.lib, self.assembly_method)
            print '*** Set up directories ***'
            self.setup(options.source)
            print '-' * 30
            tee.finalize()
        
        tee = None
                        
        if self.working_dir:
            if os.path.isdir(self.working_dir):
                tee = Tee('%s/log.txt' % self.working_dir, 'a')
                print '[%s %s@%s]: %s (%s)' % (str(datetime.now()), getpass.getuser(), socket.gethostname(), self.lib, self.assembly_method)
            else:
                print_message('WARNING', 'Cannot find %s' % self.working_dir, False)
        else:
            print_message('WARNING', 'Cannot find working directory for %s' % self.lib, False)
        
        if options.fem:
            print '*** Filter, extend, and merge contigs ***'
            self.fem()            
            print '*' * 30
        if options.prepare_reads:
            print '*** Prepare reads for alignments ***'
            self.prepare_reads()
            print '*' * 30        
        
        if options.prepare_contigs:
            print '*** Prepare contigs for alignments ***'
            self.prepare_contigs()
            print '*' * 30
            
        if options.symlink_genome_bam:
            print '*** Symlink bam and bam.bai in the read_to_genome directory ***'
            self.copy_r2g_bam()
            print '*' * 30
        
        # Action Group 2    
        if options.align_reads:
            print '*** Align reads to merged assembly ***'
            self.align_reads()
            print '*' * 30
        if options.align_contigs:
            print '*** Align merged assembly to reference genome ***'            
            self.align_contigs()
            print '*' * 30
        
        # Action Group 3
        if options.fusion:
            print '*** Call fusions and large scale rearrangement events ***'
            self.run_fusions()
            print '*' * 30        
        if options.indel:
            print '*** Call indels ***'
            self.run_indels()
            print '*' * 30        
        if options.splicing:
            print '*** Call novel splicing events and calculate coverage ***'
            if self.assembly_method == 'genome' or self.assembly_method == 'targeted_genome':
                print_message('ERROR', 'This step is not supported for genome libraries.', True)
            else:
                self.run_splicing()
            print '*' * 30            
        if options.track:
            print '*** Create UCSC custom track ***'
            if self.assembly_method == 'genome' or self.assembly_method == 'targeted_genome':
                print_message('ERROR', 'This step is not supported for genome libraries.', True)
            else:
                self.run_track()
            print '*' * 30
        
        if tee:
            print '-' * 30
            tee.finalize()        
    
    def prepare_reads(self):        
        input_reads_list = '%s/assembly/in' % (self.working_dir)
        if not os.path.isfile(input_reads_list):
            message = "Cannot find assembly reads list '" + input_reads_list + "'"
            print_message("ERROR", message, True)
        output_reads_list = '%s/reads_to_contigs/%s.in' % (self.working_dir, self.lib)    
        outdir = '%s/reads_to_contigs/reads' % self.working_dir
        jobdir = '%s/reads_to_contigs/cluster' % self.working_dir
        self.prepare_reads_helper(input_reads_list, output_reads_list, outdir, jobdir, self.lib)

    def prepare_reads_helper(self, input_reads_list, output_reads_list, outdir, jobdir, prefix):
        task = 'prepare_reads.py'
        script = PACKAGE_DIR + "/utilities/" + task
        
        args = {}
        args['INPUT_READSLIST'] = input_reads_list
        args['OUTPUT_READSLIST'] = output_reads_list
        args['LIB'] = prefix
        args['OUTPUTDIR'] = outdir
        args['JOBSCRIPTDIR'] = jobdir
        args['PROJECT'] = self.project        

        args_str = cfg.create_args(self.settings, 'commands', task, args)
        [jobs, cleanup_job] = prepare_reads.parseArgsFromCommand(args_str)
    
        job_script_configs = job_runner.getJobScriptConfigs('%s/configs/job_script.cfg' % PACKAGE_DIR)
        template = PACKAGE_DIR + "/configs/templates/" + job_script_configs["local"]    
        cleanup_script = job_runner.makeJobScript(template=template, content=cleanup_job.content, jobname=cleanup_job.name, logdir=jobdir, setup=cfg.get_setup_cmd())
        fh = open(jobdir + os.sep + cleanup_job.name + ".sh", 'w')
        fh.write(cleanup_script)
        fh.close()
        
        self.run_jobs(jobs, 'prepare_reads', jobdir, workingdir=self.working_dir)

    def align_reads(self):        
        fasta_file = '%s/merge/%s-contigs.fa' % (self.working_dir, self.lib)
        if not os.path.isfile(fasta_file):
            message = "Cannot find merged assembly '" + fasta_file + "'"
            print_message("ERROR", message, True)
        
        reads_list = '%s/reads_to_contigs/%s.in' % (self.working_dir, self.lib)
        if not os.path.isfile(reads_list):
            message = "Cannot find reads list '" + reads_list + "'"
            print_message("WARNING", message)
            
            reads_list = '%s/assembly/in' % (self.working_dir)
            print "Attempting to use the assembly reads list '" + reads_list + "'"
            
            if not os.path.isfile(reads_list):
                message = "Cannot find assembly reads list '" + reads_list + "'"
                print_message("ERROR", message, True)
               
        min_k = None
        assembly_topdir = None
        for assembly in glob.glob(os.path.join(self.working_dir + '/assembly', 'k*')):
            if assembly_topdir is None:
                assembly_topdir = os.path.dirname(os.path.realpath(os.path.abspath(assembly)))
            k = os.path.basename(assembly).replace('k', '')
            if min_k is None or int(k) < int(min_k):
                min_k = k           
        if min_k is None:
            message = 'Cannot find minimum k value from %s/assembly' % (self.working_dir)
            print_message("WARNING", message, False)
        
        return self.align_reads_helper(fasta_file, reads_list, self.lib, '%s/reads_to_contigs' % self.working_dir, '%s/reads_to_contigs/cluster' % self.working_dir, min_k)
    
    def align_reads_helper(self, fasta_file, reads_list, prefix, outdir, jobdir, min_k):
        task = 'r2c.py'
        script = PACKAGE_DIR + "/utilities/" + task
        
        args = {}
        args['CONTIGS'] = fasta_file
        args['READSLIST'] = reads_list
        args['LIB'] = prefix
        args['OUTPUTDIR'] = outdir
        args['JOBSCRIPTDIR'] = jobdir
        args['PROJECT'] = self.project
        args['MINK'] = min_k
        
        bam = outdir + os.sep + prefix + '-contigs.bam'
        for outputfile in [bam, bam + '.bai']:
            if os.path.exists(outputfile) and os.path.islink(outputfile):
                # This is a re-run. The symlink to the file from previous version must be removed.
                os.remove(outputfile)
        
        args_str = cfg.create_args(self.settings, 'commands', task, args)
        print '%s %s' % (script, args_str)
        factory = r2c.parseArgsFromCommand(args_str)
        factory.debug = self.debug
        factory.local = self.local
        factory.cluster = self.cluster
        return factory.create_and_submit_jobs()
             
    def run_fusions(self):
        task = 'fusion.py'
        script = PACKAGE_DIR + "/analysis/" + task
        
        args = {}
        args['PATH'] = self.working_dir
        args['LIB'] = self.lib
        if not self.genome:
            try:
                # get the name of reference genome from config file
                self.genome = self.settings[self.project]['reference']
            except KeyError:
                print_message('ERROR', 'Cannot find the name of the reference genome for project "%s" in configuration file "%s"' % (self.project, cfg_file), True)
        
        if self.genome and not self.settings['genomes'].has_key(self.genome):
            print_message('ERROR', 'Reference genome "%s" has not been set up in the [genomes] section of configuration file "%s"' % (self.genome, cfg_file), True)
                   
        if self.genome:
            reference_fasta_path = self.settings['genomes'][self.genome]
            if len(reference_fasta_path) > 0:
                args['GENOME'] = self.genome
            else:
                print_message('ERROR', 'The path to the FASTA file for reference genome "%s" is not specified in the [genomes] section of configuration file "%s"' % (self.genome, cfg_file), True)
        
        r2gdir = self.working_dir + '/reads_to_genome'
        if os.path.isdir(r2gdir):
            genome_bam = self.get_bam_file(r2gdir)        
            if genome_bam is None:
                message = 'Cannot find BAM file (*.bam) in directory: %s' % r2gdir
                print_message("ERROR", message, True)
            if not os.path.isfile(genome_bam + '.bai'):
                message = 'Cannot find BAM index file (*.bam.bai) in directory: %s' % r2gdir
                print_message("ERROR", message, True)            
            args['GENOME_BAM'] = genome_bam
        else:
            message = 'Directory does not exists: %s' % r2gdir
            print_message("ERROR", message, True)
        
        aln_dir = '%s/contigs_to_genome/%s-contigs/output' % (self.working_dir, self.lib)
        alns = []
        if os.path.isdir(aln_dir):
            alns = glob.glob('%s/seq.*.sam' % aln_dir)
        else:
            message = "Cannot find directory, %s" % aln_dir
            print_message("ERROR", message, True)
           
        cluster_dir = '%s/fusions/cluster' % (self.working_dir)
        
        if alns:
            max_task_id = 0
            for aln in alns:
                basename = os.path.basename(aln)
                number = basename.lstrip('seq.').rstrip('.sam')
                if number.isdigit():
                    if int(number) > max_task_id:
                        max_task_id = int(number)                    
                    out_dir_num = cluster_dir + '/' + number
                    if not os.path.isdir(out_dir_num):
                        os.makedirs(out_dir_num)
                    
            args['INPUT'] = '%s/seq.$TA_JOBID.sam' % aln_dir
            args['CTG'] = '%s/contigs_to_genome/%s-contigs/input/seq.$TA_JOBID.fa' % (self.working_dir, self.lib)
            args['OUTPUT_DIR'] = '%s/$TA_JOBID' % cluster_dir
            
            job_name = '%s-fusions' % self.lib
            job_content = 'time python %s %s' % (script, cfg.create_args(self.settings, 'commands', task, args) )
            job = MyJob(content=[job_content], name=job_name, config_key=task, first_task_id='1', last_task_id=str(max_task_id), queue=self.queue)
            
            self.run_jobs([job], job_name, cluster_dir, workingdir=self.working_dir)
                                
            #filtering job
            if job:
                self.run_fusion_filtering(depend=[job])
            #TODO: What to do otherwise?    
        else:
            message = "Cannot find alignment files (seq.*.sam) in directory, %s" % aln_dir
            print_message("ERROR", message, True)
    
            
    def run_fusion_filtering(self, depend=None):
        task = 'fusion.py-filter'
        script = PACKAGE_DIR + "/analysis/" + task.split('-')[0]
        
        args = {}
        args['PATH'] = self.working_dir
        args['LIB'] = self.lib
        
        if not self.genome:
            try:
                # get the name of reference genome from config
                self.genome = self.settings[self.project]['reference']
            except KeyError:
                print_message('ERROR', 'Cannot find the name of the reference genome for project "%s" in configuration file "%s"' % (self.project, cfg_file), True)
        
        if self.genome and not self.settings['genomes'].has_key(self.genome):
            print_message('ERROR', 'Reference genome "%s" has not been set up in the [genomes] section of configuration file "%s"' % (self.genome, cfg_file), True)
                   
        if self.genome:
            reference_fasta_path = self.settings['genomes'][self.genome]
            if len(reference_fasta_path) > 0:
                args['GENOME'] = self.genome
            else:
                print_message('ERROR', 'The path to the FASTA file for reference genome "%s" is not specified in the [genomes] section of configuration file "%s"' % (self.genome, cfg_file), True)

        
        cluster_dir = "%s/fusions/cluster" % (self.working_dir)

        job_name = '%s-fusions-filter' % self.lib
        job_content = 'time python %s %s' % (script, cfg.create_args(self.settings, 'commands', task, args) )
        job = MyJob(content=[job_content], name=job_name, config_key=task, predecessors=depend, queue=self.queue)
        self.run_jobs([job], job_name, cluster_dir, workingdir=self.working_dir)
            
            
    def run_indels(self):
        task = 'snv_caller.py'
        script = PACKAGE_DIR + "/analysis/" + task
        
        args = {}
        args['PATH'] = self.working_dir
        args['LIB'] = self.lib
        if not self.genome:
            try:
                # get the name of reference genome from config
                self.genome = self.settings[self.project]['reference']
            except KeyError:
                print_message('ERROR', 'Cannot find the name of the reference genome for project "%s" in configuration file "%s"' % (self.project, cfg_file), True)
        
        if self.genome and not self.settings['genomes'].has_key(self.genome):
            print_message('ERROR', 'Reference genome "%s" has not been set up in the [genomes] section of configuration file "%s"' % (self.genome, cfg_file), True)
                   
        if self.genome:
            reference_fasta_path = self.settings['genomes'][self.genome]
            if len(reference_fasta_path) > 0:
                args['GENOME'] = self.genome
            else:
                print_message('ERROR', 'The path to the FASTA file for reference genome "%s" is not specified in the [genomes] section of configuration file "%s"' % (self.genome, cfg_file), True)
        
        if self.assembly_method == 'genome' or self.assembly_method == 'targeted_genome':
            r2gdir = self.working_dir + '/reads_to_genome'
            if os.path.isdir(r2gdir):
                genome_bam = self.get_bam_file(r2gdir)
                if genome_bam is None:
                    message = 'Cannot find BAM file (*.bam) in directory: %s' % r2gdir
                    print_message("ERROR", message, True)
                if not os.path.isfile(genome_bam + '.bai'):
                    message = 'Cannot find BAM index file (*.bam.bai) in directory: %s' % r2gdir
                    print_message("ERROR", message, True)
                
                args['GENOME_BAM'] = genome_bam
            else:
                message = 'Directory does not exists: %s' % r2gdir
                print_message("ERROR", message, True)
        
        aln_dir = '%s/contigs_to_genome/%s-contigs/output' % (self.working_dir, self.lib)
        alns = []
        if os.path.isdir(aln_dir):
            alns = glob.glob('%s/seq.*.sam' % aln_dir)
        else:
            message = "Cannot find directory, %s" % aln_dir
            print_message("ERROR", message, True)
        cluster_dir = '%s/indels/cluster' % (self.working_dir)
        
        if alns:
            max_task_id = 0
            for aln in alns:
                basename = os.path.basename(aln)
                number = basename.lstrip('seq.').rstrip('.sam')
                if number.isdigit():
                    if int(number) > max_task_id:
                        max_task_id = int(number)                    
                    out_dir_num = cluster_dir + '/' + number
                    if not os.path.isdir(out_dir_num):
                        os.makedirs(out_dir_num)
                    
            args['ALN'] = '%s/seq.$TA_JOBID.sam' % aln_dir
            args['CTG'] = '%s/contigs_to_genome/%s-contigs/input/seq.$TA_JOBID.fa' % (self.working_dir, self.lib)
            args['OUTPUT_DIR'] = '%s/$TA_JOBID' % cluster_dir
            
            job_name = '%s-indels' % self.lib
            job_content = 'time python %s %s' % (script, cfg.create_args(self.settings, 'commands', task, args) )
            job = MyJob(content=[job_content], name=job_name, config_key=task, first_task_id='1', last_task_id=str(max_task_id), queue=self.queue)
            self.run_jobs([job], job_name, cluster_dir, workingdir=self.working_dir)
                
            #filtering job
            if job:
                self.run_indel_filtering(depend=[job])
            #@TODO: What to do otherwise?    
        else:
            message = "Cannot find alignment files (seq.*.sam) in directory, %s" % aln_dir
            print_message("ERROR", message, True)

            
    def run_indel_filtering(self, depend=None):
        task = 'snv_caller.py-filter'
        script = PACKAGE_DIR + "/analysis/" + task.split('-')[0]
        
        args = {}
        args['PATH'] = self.working_dir
        
        if not self.genome:
            try:
                # get the name of reference genome from config
                self.genome = self.settings[self.project]['reference']
            except KeyError:
                print_message('ERROR', 'Cannot find the name of the reference genome for project "%s" in configuration file "%s"' % (self.project, cfg_file), True)
        
        if self.genome and not self.settings['genomes'].has_key(self.genome):
            print_message('ERROR', 'Reference genome "%s" has not been set up in the [genomes] section of configuration file "%s"' % (self.genome, cfg_file), True)
                   
        if self.genome:
            reference_fasta_path = self.settings['genomes'][self.genome]
            if len(reference_fasta_path) > 0:
                args['GENOME'] = self.genome
            else:
                print_message('ERROR', 'The path to the FASTA file for reference genome "%s" is not specified in the [genomes] section of configuration file "%s"' % (self.genome, cfg_file), True)
        
    
        cluster_dir = "%s/indels/cluster" % (self.working_dir)

        job_name = '%s-indels-filter' % self.lib
        job_content = 'time python %s %s' % (script, cfg.create_args(self.settings, 'commands', task, args) )
        job = MyJob(content=[job_content], name=job_name, config_key=task, predecessors=depend, queue=self.queue)
        self.run_jobs([job], job_name, cluster_dir, workingdir=self.working_dir)
            
            
    def get_bam_file(self, from_dir):
        bam_files = glob.glob(os.path.join(from_dir + '/*.bam'))
        if len(bam_files) == 1:
            return bam_files[0]
        else:
            return None
        
    def get_num_contigs(self, ff):
        num_ctgs = 0
        with open(ff) as f:
            for i, l in enumerate(f):
                if l[0] == '>':
                    num_ctgs += 1
                pass
        return num_ctgs
    
    def qmake(self, options, max_parallel_make_tasks=None):
        if max_parallel_make_tasks < 1:
            print_message('ERROR', 'The maximum number of make tasks cannot be smaller than 1.', quit=True)
    
        # -d    set up the output directory and symlink the ABySS assemblies
        if options.set_up:
            tee = Tee('%s/log.txt' % self.top_dir, 'a')
            print '[%s %s@%s]: %s (%s)' % (str(datetime.now()), getpass.getuser(), socket.gethostname(), self.lib, self.assembly_method)
            print '*** Set up directories ***'
            self.setup(options.source)
            print '-' * 30
            tee.finalize()
        
        tee = None
                        
        if self.working_dir:
            if os.path.isdir(self.working_dir):
                tee = Tee('%s/log.txt' % self.working_dir, 'a')
                print '[%s %s@%s]: %s (%s)' % (str(datetime.now()), getpass.getuser(), socket.gethostname(), self.lib, self.assembly_method)
            else:
                print_message('WARNING', 'Cannot find %s' % self.working_dir, False)
        else:
            print_message('WARNING', 'Cannot find working directory for %s' % self.lib, False)
        
        # -b    symlink the R2G bam
        if options.symlink_genome_bam:
            print '*** Symlink bam and bam.bai in the read_to_genome directory ***'
            self.copy_r2g_bam()
            print '*' * 30
        
        #qmake
        qmake.run_pipe(self, options, max_parallel_make_tasks=max_parallel_make_tasks)


class PipeTranscriptome(Pipe):
    def __init__(self, assembly_method, project, lib, readlength, assembly_dir=None, cluster=None, genome=None, queue=None, local=False, debug=False, top_dir=None):
        Pipe.__init__(self, assembly_method, project, lib, readlength, assembly_dir=assembly_dir, cluster=cluster, genome=genome, queue=queue, local=local, debug=debug, top_dir=top_dir)
        #print self.assembly_method
        
        cfg_file = "%s/configs/transcriptome.cfg" % (PACKAGE_DIR)
        self.set_working_dir(cfg_file)
        #print self.working_dir
    
    def fem(self):
        filter_jobs = self.filter_extend()
        input_dir = '%s/filter' % (self.working_dir)
        if filter_jobs is not None and len(filter_jobs) > 0:
            merge_job = self.merge(input_dir, depend=filter_jobs)
            if merge_job is not None:
                self.calculate_assembly_stats(depend=[merge_job])
            else:
                message = "Cannot submit assembly-merging jobs."
                print_message("ERROR", message, True)            
        else:
            message = "Cannot submit assembly-filtering jobs."
            print_message("ERROR", message, True)

    def filter_extend(self):
        task = 'abyss-ta-filter'
        script = PACKAGE_DIR + "/wrappers/" + task
        
        args = {}
        args['LIB'] = self.lib
        args['READLENGTH'] = self.readlength
                
        if not self.ks:
            assembly_link_dir = "%s/assembly" % self.working_dir
            if not os.path.isdir(assembly_link_dir) and self.assembly_dir:
                self.link_assemblies()
                
            elif os.path.isdir(assembly_link_dir):
                self.ks = self.get_ks(assembly_link_dir)
        
        if self.ks:
            jobs = []
            cluster_dir = '%s/filter/cluster' % self.working_dir
            if not os.path.isdir(cluster_dir):
                os.makedirs(cluster_dir)

            for k in self.ks:
                args['INPUT_DIR'] = '%s/assembly/k%s' % (self.working_dir, k)
                args['OUTPUT_DIR'] = '%s/filter/k%s' % (self.working_dir, k)
                args['K'] = k
                if not os.path.isdir(args['OUTPUT_DIR']):
                    os.makedirs(args['OUTPUT_DIR'])

                job_name = '%s-k%s-filter' % (self.lib, k)
                job_content = 'time %s %s' % (script, cfg.create_args(self.settings, 'commands', task, args) )
                jobs.append(MyJob(content=[job_content], name=job_name, config_key=task, queue=self.queue))
            
            self.run_jobs(jobs, '%s-ks-filter' % self.lib, cluster_dir, workingdir=self.working_dir)
                    
            return jobs
        else:
            message = "Cannot find assembly k-directories, ie. %s/assembly/k*" % self.working_dir
            print_message("ERROR", message, True)
        
    def align_contigs(self):
        top_dir = "%s/contigs_to_genome" % (self.working_dir)            
        job_name = "%s-contigs" % (self.lib)
        aln_dir = "%s/%s" % (top_dir, job_name)
        cluster_dir = "%s/cluster" % (aln_dir)
        output_dir = "%s/output" % (aln_dir)
        input_dir = "%s/input" % (aln_dir)

        #check the "input" directory
        if not os.path.isdir(input_dir):
            print_message("ERROR", "Cannot find directory %s\nDid you forget to run the step \"--prepare_contigs\"?" % input_dir, True)

        #check the number of seq.*.fa
        num_files = len(glob.glob(os.path.join(input_dir + '/seq.*.fa')))
        if num_files <= 0:
            print_message("ERROR", "Cannot find \"seq.*.fa\" in %s\nDid you forget to run the step \"--prepare_contigs\"?" % input_dir, True)
        
        if not os.path.isdir(cluster_dir):
            os.makedirs(cluster_dir)
                        
        if os.path.exists(output_dir) and os.path.islink(output_dir):
            # This is a re-run. The symlink to the "output" directory from the previous version must be removed.
            os.remove(output_dir)
        if not os.path.isdir(output_dir):
            os.makedirs(output_dir)
                    
        #align
        script = 'gmap'
        task = 'gmap'
        jobs = []
        args = {}
        
        if not self.genome:
            try:
                # get the name of reference genome from config file
                self.genome = self.settings[self.project]['reference']
            except KeyError:
                print_message('ERROR', 'Cannot find the name of the reference genome for project "%s" in configuration file "%s"' % (self.project, cfg_file), True)
        
        if self.genome and not self.settings['genomes'].has_key(self.genome):
            print_message('ERROR', 'Reference genome "%s" has not been set up in the [genomes] section of configuration file "%s"' % (self.genome, cfg_file), True)
                   
        if self.genome:
            reference_fasta_path = self.settings['genomes'][self.genome]
            if len(reference_fasta_path) > 0:
                args['GENOME'] = self.genome
                args['GENOMEDIR'] = os.path.split(reference_fasta_path)[0]
            else:
                print_message('ERROR', 'The path to the FASTA file for reference genome "%s" is not specified in the [genomes] section of configuration file "%s"' % (self.genome, cfg_file), True)
        
        args['QUERY'] = '%s/seq.$TA_JOBID.fa' % input_dir
        
        outputfile = '%s/seq.$TA_JOBID.sam' % output_dir
        args['OUTPUT'] = outputfile + '.incomplete'
        
        job_name = '%s-gmap' % self.lib
        job_content = []
        job_content.append('time %s %s' % (script, cfg.create_args(self.settings, 'commands', task, args) ) )
        job_content.append('mv %s.incomplete %s' % (outputfile, outputfile))
        

        jobs.append(MyJob(content=job_content, name=job_name, config_key=task, first_task_id='1', last_task_id=str(num_files), queue=self.queue))
        self.run_jobs(jobs, job_name, cluster_dir, workingdir=self.working_dir)
                
    def run_track(self):
        task = 'align_parser.py'
        script = PACKAGE_DIR + "/utilities/" + task
        
        args = {}
        args['PATH'] = self.working_dir
        args['LIB'] = self.lib
        if not self.genome:
            try:
                # get the name of reference genome from config file
                self.genome = self.settings[self.project]['reference']
            except KeyError:
                print_message('ERROR', 'Cannot find the name of the reference genome for project "%s" in configuration file "%s"' % (self.project, cfg_file), True)
        
        if self.genome and not self.settings['genomes'].has_key(self.genome):
            print_message('ERROR', 'Reference genome "%s" has not been set up in the [genomes] section of configuration file "%s"' % (self.genome, cfg_file), True)
                   
        if self.genome:
            reference_fasta_path = self.settings['genomes'][self.genome]
            if len(reference_fasta_path) > 0:
                args['GENOME'] = self.genome
            else:
                print_message('ERROR', 'The path to the FASTA file for reference genome "%s" is not specified in the [genomes] section of configuration file "%s"' % (self.genome, cfg_file), True)

        cluster_dir = '%s/tracks/cluster' % (self.working_dir)
        if not os.path.isdir(cluster_dir):
            os.makedirs(cluster_dir)
           
        job_name = '%s-track' % self.lib
        job_content = 'time python %s %s' % (script, cfg.create_args(self.settings, 'commands', task, args) )
        job = MyJob(content=[job_content], name=job_name, config_key=task, queue=self.queue)
        self.run_jobs([job], job_name, cluster_dir, workingdir=self.working_dir)
    
    def run_splicing(self):
        task = 'model_matcher.py'
        script = PACKAGE_DIR + "/analysis/" + task
        #args_template = cfg.get_value(self.settings, self.project, task)
        #mem = cfg.get_value(self.settings, 'memory', task)
        
        args = {}
        args['PATH'] = self.working_dir
        args['LIB'] = self.lib
        
        if not self.genome:
            try:
                # get the name of reference genome from config file
                self.genome = self.settings[self.project]['reference']
            except KeyError:
                print_message('ERROR', 'Cannot find the name of the reference genome for project "%s" in configuration file "%s"' % (self.project, cfg_file), True)
        
        if self.genome and not self.settings['genomes'].has_key(self.genome):
            print_message('ERROR', 'Reference genome "%s" has not been set up in the [genomes] section of configuration file "%s"' % (self.genome, cfg_file), True)
                   
        if self.genome:
            reference_fasta_path = self.settings['genomes'][self.genome]
            if len(reference_fasta_path) > 0:
                args['GENOME'] = self.genome
            else:
                print_message('ERROR', 'The path to the FASTA file for reference genome "%s" is not specified in the [genomes] section of configuration file "%s"' % (self.genome, cfg_file), True)
                
        aln_dir = '%s/contigs_to_genome/%s-contigs/output' % (self.working_dir, self.lib)
        alns = []
        if os.path.isdir(aln_dir):
            alns = sorted(glob.glob(os.path.join(aln_dir, '*')))
        cluster_dir = '%s/splicing/cluster' % (self.working_dir)
        
        if alns:
            max_task_id = 0
            for aln in alns:
                basename = os.path.basename(aln)
                number = basename.lstrip('seq.').rstrip('.sam')
                if number.isdigit():
                    if int(number) > max_task_id:
                        max_task_id = int(number)                    
                    out_dir_num = cluster_dir + '/' + number
                    if not os.path.isdir(out_dir_num):
                        os.makedirs(out_dir_num)
                    
            args['ALN'] = '%s/seq.$TA_JOBID.sam' % aln_dir
            args['CTG'] = '%s/contigs_to_genome/%s-contigs/input/seq.$TA_JOBID.fa' % (self.working_dir, self.lib)
            args['OUTPUT_DIR'] = '%s/$TA_JOBID' % cluster_dir
            job_name = '%s-splicing' % self.lib
            job_content = 'time python %s %s' % (script, cfg.create_args(self.settings, 'commands', task, args) )
            job = MyJob(content=[job_content], name=job_name, config_key=task, first_task_id='1', last_task_id=str(max_task_id), queue=self.queue)
            self.run_jobs([job], job_name, cluster_dir, workingdir=self.working_dir)
                
            #filtering job
            if job:
                self.run_splicing_filtering(depend=[job])
                
        else:
            message = "Cannot find alignment files (seq.*.sam) in directory, %s" % aln_dir
            print_message("ERROR", message, True)
            
    
    def run_splicing_filtering(self, depend=None):
        task = 'model_matcher.py-filter'
        script = PACKAGE_DIR + "/analysis/" + task.split('-')[0]
        
        args = {}
        args['PATH'] = self.working_dir
        if not self.genome:
            try:
                # get the name of reference genome from config file
                self.genome = self.settings[self.project]['reference']
            except KeyError:
                print_message('ERROR', 'Cannot find the name of the reference genome for project "%s" in configuration file "%s"' % (self.project, cfg_file), True)
        
        if self.genome and not self.settings['genomes'].has_key(self.genome):
            print_message('ERROR', 'Reference genome "%s" has not been set up in the [genomes] section of configuration file "%s"' % (self.genome, cfg_file), True)
                   
        if self.genome:
            reference_fasta_path = self.settings['genomes'][self.genome]
            if len(reference_fasta_path) > 0:
                args['GENOME'] = self.genome
            else:
                print_message('ERROR', 'The path to the FASTA file for reference genome "%s" is not specified in the [genomes] section of configuration file "%s"' % (self.genome, cfg_file), True)
                   
        cluster_dir = "%s/splicing/cluster" % (self.working_dir)

        job_name = '%s-splicing-filter' % self.lib
        job_content = 'time python %s %s' % (script, cfg.create_args(self.settings, 'commands', task, args) )
        job = MyJob(content=[job_content], name=job_name, config_key=task, predecessors=depend, queue=self.queue)
        self.run_jobs([job], job_name, cluster_dir, workingdir=self.working_dir)
    
class PipeGenome(Pipe):
    def __init__(self, assembly_method, project, lib, readlength, assembly_dir=None, cluster=None, genome=None, queue=None, local=False, debug=False, top_dir=None):
        Pipe.__init__(self, assembly_method, project, lib, readlength, assembly_dir=assembly_dir, cluster=cluster, genome=genome, queue=queue, local=local, debug=debug, top_dir=top_dir)
        #print self.assembly_method
        
        cfg_file = "%s/configs/genome.cfg" % (PACKAGE_DIR)    
        #print cfg_file
        self.set_working_dir(cfg_file)
        #print self.working_dir
    
    # for targetted genome libraries
    def concat_contigs_indels(self):
        assembly_dir = '%s/assembly' % (self.working_dir)
        target_dir = '%s/filter' % (self.working_dir)

        cluster_dir = '%s/cluster' % (target_dir)
        jobs = []
        if self.cluster and not os.path.isdir(cluster_dir):
            os.makedirs(cluster_dir)
        
        self.ks = self.get_ks(assembly_dir)
        for k in self.ks.keys():
            target_dir = '%s/filter/k%s' % (self.working_dir, k)
            if not os.path.isdir(target_dir):
                os.makedirs(target_dir)

            source_contigs = '%s/k%s/%s-contigs.fa' % (assembly_dir, k, self.lib)
            if not os.path.isfile(source_contigs):
                message = "Cannot find %s" % source_contigs
                print_message("ERROR", message, True)
            
            source_indels = '%s/k%s/%s-indel.fa' % (assembly_dir, k, self.lib)
            if not os.path.isfile(source_indels):
                message = "Cannot find %s" % source_indels
                print_message("ERROR", message, True)

            target_contigs = '%s/%s-contigs.fa' % (target_dir, self.lib)
            
            cmd = 'cat %s %s > %s' % (source_contigs, source_indels, target_contigs)
            
            job_name = '%s-combine-k%s' % (self.lib, k)
            jobs.append(MyJob(content=[cmd], name=job_name, config_key=None, predecessors=None, queue=self.queue))
        
        self.run_jobs(jobs, '%s-combine-ks', cluster_dir, workingdir=self.working_dir)
        return jobs   
    
    # for full genome libraries
    def include_indel_bubbles(self):
        task = 'abyss-ta-filter'
        script = PACKAGE_DIR + "/wrappers/" + task
            
        args = {}
        args['LIB'] = self.lib
        args['READLENGTH'] = self.readlength        
            
        #bubble extending jobs
        extend_jobs = []
        cluster_dir = "%s/merge/indel_bubbles/cluster" % (self.working_dir)
        if not os.path.isdir(cluster_dir):
            os.makedirs(cluster_dir)
        
        assembly_dir = self.working_dir + '/assembly'
        ks = sorted(glob.glob(os.path.join(assembly_dir, 'k*')))
        k_sizes = []
        for k in ks:            
            args['K'] = os.path.basename(k).replace('k', '')
            args['INPUT_DIR'] = k
            args['OUTPUT_DIR'] = '%s/merge/indel_bubbles/k%s' % (self.working_dir, args['K'])
            k_sizes.append(int(args['K']))
        
            if not os.path.isdir(args['OUTPUT_DIR']):
                os.mkdir(args['OUTPUT_DIR'])
            
            job_name = '%s-k%s-extend' % (self.lib, args['K'])
            #Use '-b' option for indels only
            job_content = 'time %s %s -b' % (script, cfg.create_args(self.settings, 'commands', task, args) )
            extend_jobs.append(MyJob(content=[job_content], name=job_name, config_key=task, predecessors=None, queue=self.queue))
        self.run_jobs(extend_jobs, '%s-ks-extend' % self.lib, cluster_dir, workingdir=self.working_dir)
               
        #bubble merging jobs - dependent on first jobs
        task = 'abyss-rmdups-iterative'
        script = PACKAGE_DIR + "/wrappers/" + task

        args['INPUT_DIR'] = self.working_dir + '/merge/indel_bubbles'
        args['OUTPUT_DIR'] = self.working_dir + '/merge/indel_bubbles'
        args['LIB'] = self.lib
        args['K'] = max(k_sizes)

        largest_k = self.working_dir + '/assembly/k' + str(max(k_sizes))
        pe_contigs = largest_k + '/' + self.lib + '-contigs.fa'
        
        if not os.path.isfile(pe_contigs):
            message = "Cannot find %s" % pe_contigs
            print_message("ERROR", message, True)
        
        merged_indels = args['OUTPUT_DIR'] + '/' + self.lib + '-b.fa'
        contigs_indels_cat = self.working_dir + '/merge/' + self.lib + '-cat.fa'

        #Use the '-s b' option to merge only *-b.fa
        job_content = []
        job_content.append('time %s %s -s b' % (script, cfg.create_args(self.settings, 'commands', task, args)))
        job_content.append('sed -e "s/>/>indel_/g" %s |cat - %s > %s' % (merged_indels, pe_contigs, contigs_indels_cat))
        
        # Remove duplicates in contigs-indel concatenated set
        args['INPUT_DIR'] = contigs_indels_cat
        args['OUTPUT_DIR'] = self.working_dir + '/merge'
        args['LIB'] = self.lib
        
        job_content.append('time %s %s --no-ol' % (script, cfg.create_args(self.settings, 'commands', task, args)))
        job_content.append('rm -f %s' % contigs_indels_cat)

        final_merged_assembly = args['OUTPUT_DIR'] + os.sep + self.lib + '-contigs.fa'
        if os.path.exists(final_merged_assembly) and os.path.islink(final_merged_assembly):
            # This is a re-run. The symlink to the file from the previous version must be removed.
            os.remove(final_merged_assembly)
        
        job_name = "%s-merge" % (self.lib)
        merge_job = MyJob(content=job_content, name=job_name, config_key=task, predecessors=extend_jobs, queue=self.queue)
        self.run_jobs([merge_job], job_name, cluster_dir, workingdir=self.working_dir)
        return merge_job
    
    def fem(self):
        if self.assembly_method == 'targeted_genome':            
            input_dir = '%s/filter' % (self.working_dir)
            jobs = self.concat_contigs_indels()
            if jobs:
                merge_job = self.merge(input_dir, depend=jobs)
                if merge_job is not None:
                    self.calculate_assembly_stats(depend=[merge_job])
                else:
                    print_message("ERROR", "Cannot submit job for merging assemblies!", True)
            else:
                print_message("ERROR", "Cannot submit jobs for concatenating *-indels.fa *-contigs.fa!", True)
        elif self.assembly_method == 'genome':
            merge_job = self.include_indel_bubbles()
            if merge_job is not None:
                self.calculate_assembly_stats(depend=[merge_job])
            else:
                print_message("ERROR", "Cannot submit job for merging assemblies!", True)
        else:
            message = 'Unknown assembly method: %s' % str(self.assembly_method)
            print_message("ERROR", message, True)
    
    def align_contigs(self):
        top_dir = "%s/contigs_to_genome" % (self.working_dir)            
        job_name = "%s-contigs" % (self.lib)
        aln_dir = "%s/%s" % (top_dir, job_name)
        cluster_dir = "%s/cluster" % (aln_dir)
        output_dir = "%s/output" % (aln_dir)
        input_dir = "%s/input" % (aln_dir)

        #check the "input" directory
        if not os.path.isdir(input_dir):
            print_message("ERROR", "Cannot find directory %s\nDid you forget to run the step \"--prepare_contigs\"?" % input_dir, True)

        #check the number of seq.*.fa
        num_files = len(glob.glob(os.path.join(input_dir + '/seq.*.fa')))
        if num_files <= 0:
            print_message("ERROR", "Cannot find \"seq.*.fa\" in %s\nDid you forget to run the step \"--prepare_contigs\"?" % input_dir, True)
        
        if not os.path.isdir(cluster_dir):
            os.makedirs(cluster_dir)            
        if os.path.exists(output_dir) and os.path.islink(output_dir):
            # This is a re-run. The symlink to the "output" directory from the previous version must be removed.
            os.remove(output_dir)
        if not os.path.isdir(output_dir):
            os.makedirs(output_dir)
              
        #alignment            
        script = 'bwa'
        task = 'c2g_bwa'
        jobs = []
        args = {}
        
        if not self.genome:
            try:
                # get the name of reference genome from config
                self.genome = self.settings[self.project]['reference']
            except KeyError:
                print_message('ERROR', 'Cannot find the name of the reference genome for project "%s" in configuration file "%s"' % (self.project, cfg_file), True)
        
        if self.genome and not self.settings['genomes'].has_key(self.genome):
            print_message('ERROR', 'Reference genome "%s" has not been set up in the [genomes] section of configuration file "%s"' % (self.genome, cfg_file), True)
        
        if self.genome:
            reference_fasta_path = self.settings['genomes'][self.genome]
            if len(reference_fasta_path) > 0:
                args['TARGET'] = reference_fasta_path
            else:
                print_message('ERROR', 'The path to the FASTA file for reference genome "%s" is not specified in the [genomes] section of configuration file "%s"' % (self.genome, cfg_file), True)             
        

        args['QUERY'] = '%s/seq.$TA_JOBID.fa' % input_dir
        outputfile = '%s/seq.$TA_JOBID.sam' % output_dir
        args['OUTPUT'] = outputfile + '.incomplete'
        
        job_name = '%s-bwamem' % self.lib
        job_content = []
        job_content.append('time %s %s' % (script, cfg.create_args(self.settings, 'commands', task, args) ) )
        job_content.append('mv %s.incomplete %s' % (outputfile, outputfile))
        
        jobs.append(MyJob(content=job_content, name=job_name, config_key=task, first_task_id='1', last_task_id=str(num_files), queue=self.queue))
        self.run_jobs(jobs, job_name, cluster_dir, workingdir=self.working_dir)
               

class PipeStranded(PipeTranscriptome):
    def __init__(self, assembly_method, project, meta_lib, lib_plus, lib_minus, readlength, assembly_dir_plus=None, assembly_dir_minus=None, cluster=None, genome=None, queue=None, local=False, debug=False, top_dir=None):
        PipeTranscriptome.__init__(self, assembly_method, project, meta_lib, readlength, assembly_dir=None, cluster=cluster, genome=genome, queue=queue, local=local, debug=debug, top_dir=top_dir)
        self.meta_lib = meta_lib
        self.lib_plus = lib_plus
        self.lib_minus = lib_minus
        self.assembly_dir_plus = assembly_dir_plus
        self.assembly_dir_minus = assembly_dir_minus
        if self.assembly_dir_plus and os.path.isdir(self.assembly_dir_plus):
            self.ks_plus = self.get_ks(self.assembly_dir_plus)
        if self.assembly_dir_minus and os.path.isdir(self.assembly_dir_minus):
            self.ks_minus = self.get_ks(self.assembly_dir_minus)
    
    #setup directories
    def setup(self, source=None):
        if self.working_dir is None:
            return False
                    
        cur_path = os.getcwd()
        if not os.path.isdir(self.working_dir):
            os.makedirs(self.working_dir)
        os.chdir(self.working_dir)
        
        if source is None:
            folders = ['assembly', 'assembly/plus_strand', 'assembly/minus_strand', 'filter', 'filter/plus_strand', 'filter/minus_strand', 'merge', 'merge/plus_strand', 'merge/minus_strand', 'reads_to_contigs', 'reads_to_contigs/plus_strand', 'reads_to_contigs/minus_strand', 'reads_to_contigs/unknown', 'contigs_to_genome', 'reads_to_genome']
            print 'Creating directories...'
            for folder in folders:
                if not os.path.isdir(folder):
                    os.makedirs(folder)
            
            print 'Creating symlinks to assembly directories...'
            self.link_assemblies()                    
        else:
            # make the top directory for each step, symlink the important files and directories
            folders = ['merge', 'reads_to_contigs', 'contigs_to_genome', 'reads_to_genome']
            print 'Creating directories...'
            for folder in folders:
                if not os.path.isdir(folder):
                    os.mkdir(folder)
            
            source = '%s/%s/trans-abyss-v%s' % (self.top_dir, self.lib, options.source.lstrip('v'))
            
            print 'Creating symlinks to previous run...'
            os.symlink('%s/assembly' % source, 'assembly')
            
            important_files = ['merge/%s-contigs.fa' % self.lib, 'reads_to_contigs/%s.in' % self.lib, 'reads_to_contigs/%s-contigs.bam' % self.lib, 'reads_to_contigs/%s-contigs.bam.bai' % self.lib]
            for f in important_files:
                sourcefile = '%s/%s' % (source, f)
                if os.path.isfile(sourcefile):
                    os.symlink(sourcefile, '%s/%s' % (self.working_dir, f) )
                # Don't do anything if the source file doesn't exist
                # because the corresponding step might not have been run, yet!
            
            if os.path.isdir('%s/contigs_to_genome/%s-contigs' % (source, self.lib)):
                contigs_to_genome_dir = '%s/contigs_to_genome/%s-contigs' % (self.working_dir, self.lib)
                os.makedirs(contigs_to_genome_dir)
                for folder in ('input', 'output'):
                    os.symlink('%s/contigs_to_genome/%s-contigs/%s' % (source, self.lib, folder), '%s/%s' % (contigs_to_genome_dir, folder) )
            
            if os.path.isdir('%s/reads_to_genome' % source):
                for b in glob.glob('%s/reads_to_genome/*.bam*' % source):
                    os.symlink(b, '%s/reads_to_genome/%s' % (self.working_dir, os.path.basename(b)) )
            
        os.chdir(cur_path)        
    
    #link the assemblies for both strands
    def link_assemblies(self):
        #print self.lib_plus + " " + self.assembly_dir_plus
        #print self.lib_minus + " " + self.assembly_dir_minus        
        
        target_assembly_dir = "%s/assembly" % (self.working_dir)
        target_assembly_dir_plus = "%s/plus_strand" % (target_assembly_dir)
        target_assembly_dir_minus = "%s/minus_strand" % (target_assembly_dir)
        for d in [target_assembly_dir, target_assembly_dir_plus, target_assembly_dir_minus]:        
            if not os.path.isdir(d):
                os.makedirs(d)        

        #symlinks for plus strand            
        self.ks_plus = self.get_ks(self.assembly_dir_plus)
        ks_plus = self.ks_plus.keys()
        ks_plus.sort(key=int)
        for k in ks_plus:
            source_k_assembly = "%s/%s" % (self.assembly_dir_plus, self.ks_plus[k])
            target_k_assembly = "%s/k%s" % (target_assembly_dir_plus, k)
            if not os.path.exists(target_k_assembly):
                os.symlink(source_k_assembly, target_k_assembly)
                
        #symlinks for minus strand        
        self.ks_minus = self.get_ks(self.assembly_dir_minus)
        ks_minus = self.ks_minus.keys()
        ks_minus.sort(key=int)
        for k in ks_minus:
            source_k_assembly = "%s/%s" % (self.assembly_dir_minus, self.ks_minus[k])
            target_k_assembly = "%s/k%s" % (target_assembly_dir_minus, k)
            if not os.path.exists(target_k_assembly):
                os.symlink(source_k_assembly, target_k_assembly)
        
        #@TODO: We should look for "$LIB.in", then "in", then *.in
        in_files_plus = glob.glob(os.path.join(self.assembly_dir_plus, '*in'))        
        in_files_minus = glob.glob(os.path.join(self.assembly_dir_minus, '*in'))
        if len(in_files_plus) == 1 and len(in_files_minus) == 1:
            target_in = "%s/in" % (target_assembly_dir)
            
            shutil.copyfile(in_files_plus[0], target_assembly_dir_plus + os.sep + 'in')
            shutil.copyfile(in_files_minus[0], target_assembly_dir_minus + os.sep + 'in')
                        
            #cat both in files together
            #target_in_fh = open(target_in,'wb')
            #shutil.copyfileobj(open(in_files_plus[0],'rb'), target_in_fh)
            #shutil.copyfileobj(open(in_files_minus[0],'rb'), target_in_fh)
            #target_in_fh.close()
        else:
            message = "Cannot find the 'in' files from the assembly directories..." + os.linesep + str(in_files_plus) + os.linesep + str(in_files_minus)
            print_message("ERROR", message, True)
    
    #run fem
    def fem(self):
        merge_plus_job = None
        merge_minus_job = None

        # 1.a) filter,extend contigs in PLUS assemblies
        filter_plus_jobs = self.filter_extend('plus')       
        if filter_plus_jobs and len(filter_plus_jobs) > 0:
            # 2.a) merge PLUS assemblies
            merge_plus_job = self.merge(depend=filter_plus_jobs, strand='plus')
            if not merge_plus_job:
                message = "Cannot submit jobs for merging assemblies for 'plus' strand!"
                print_message("ERROR", message, True)
        else:
            message = "Cannot submit jobs for filtering and extension of contigs for 'plus' strand!"
            print_message("ERROR", message, True)
            
        # 1.b) filter,extend contigs in MINUS assemblies
        filter_minus_jobs = self.filter_extend('minus')
        if filter_minus_jobs and len(filter_minus_jobs) > 0:
            # 2.b) merge MINUS assemblies
            merge_minus_job = self.merge(depend=filter_minus_jobs, strand='minus')
            if not merge_minus_job:
                message = "Cannot submit jobs for merging assemblies for 'minus' strand!"
                print_message("ERROR", message, True)
        else:
            message = "Cannot submit jobs for filtering and extension of contigs for 'minus' strand!"
            print_message("ERROR", message, True)
            
        # 3. concatenate the merged PLUS assembly and the merged MINUS assembly together
        concat_job = self.concat(depend=[merge_plus_job, merge_minus_job])
        if concat_job is not None:
            self.calculate_assembly_stats(depend=[concat_job])
        else:
            message = "Cannot submit job for concatenating the assemblies for 'plus' and 'minus' strand!"
            print_message("ERROR", message, True)   
        
    def filter_extend(self, strand=None):
        task = 'abyss-ta-filter'
        script = PACKAGE_DIR + "/wrappers/" + task
        #args_template = cfg.get_value(self.settings, self.project, task)
        #mem = cfg.get_value(self.settings, 'memory', task)
                
        assembly_link_dir = "%s/assembly" % (self.working_dir)
        if not os.path.isdir(assembly_link_dir):
            self.link_assemblies()
        
        lib = None
        if strand == 'plus':
            lib = self.lib_plus
        elif strand == 'minus':
            lib = self.lib_minus
        else:
            message = "Unknown strand (" + str(strand) + ")"
            print_message("ERROR", message, True)
        
        ks = self.get_ks('%s/assembly/%s_strand' % (self.working_dir, strand))                
        if ks and len(ks) > 0:
            jobs = []
            cluster_dir = '%s/filter/%s_strand/cluster' % (self.working_dir, strand)
            if not os.path.isdir(cluster_dir):
                os.makedirs(cluster_dir)
                
            args = {}
            args['READLENGTH'] = self.readlength
            args['LIB'] = lib 

            for k in ks:
                args['INPUT_DIR'] = '%s/assembly/%s_strand/k%s' % (self.working_dir, strand, k)
                args['OUTPUT_DIR'] = '%s/filter/%s_strand/k%s' % (self.working_dir, strand, k)
                args['K'] = k
                if not os.path.isdir(args['OUTPUT_DIR']):
                    os.makedirs(args['OUTPUT_DIR'])

                job_name = '%s-k%s-filter' % (self.lib, k)
                job_content = 'time %s %s' % (script, cfg.create_args(self.settings, 'commands', task, args) )
                jobs.append(MyJob(content=[job_content], name=job_name, config_key=task, queue=self.queue))
            self.run_jobs(jobs, '%s-ks-filter' % self.lib, cluster_dir, workingdir=self.working_dir)
            
            if strand == 'plus':
                self.ks_plus = ks
            elif strand == 'minus':
                self.ks_minus = ks
                    
            return jobs
        else:
            message = "Cannot find assembly k-directories, ie. %s/assembly/k*" % self.working_dir
            print_message("ERROR", message, True)
        
            
    def merge(self, depend=[], strand=None):
        task = 'abyss-rmdups-iterative'
        script = PACKAGE_DIR + "/wrappers/" + task
                
        lib = None
        mink = None
        if strand == 'plus':
            lib = self.lib_plus
            mink = max(self.ks_plus.keys())
        elif strand == 'minus':
            lib = self.lib_minus
            mink = max(self.ks_minus.keys())
        else:
            message = "Unknown strand (" + str(strand) + ")"
            print_message("ERROR", message, True)
                
        args = {}
        args['LIB'] = lib
        args['INPUT_DIR'] = '%s/filter/%s_strand' % (self.working_dir, strand)
        args['OUTPUT_DIR'] = '%s/merge/%s_strand' % (self.working_dir, strand)
        args['K'] = mink
        args['READLENGTH'] = self.readlength
        
        cluster_dir = '%s/merge/%s_strand/cluster' % (self.working_dir, strand)
        if not os.path.isdir(cluster_dir):
            os.makedirs(cluster_dir)       

        job_name = '%s-merge' % self.lib
        merged_fasta = '%s/merge/%s_strand/%s-contigs.fa' % (self.working_dir, strand, lib)

        # the strand must be flagged in the contig id for the 3 sets of r2c alignments to be merged        
        job_content = []
        job_content.append('time %s -p %s %s' % (script, cfg.create_args(self.settings, 'commands', task, strand, args) ))
        job = MyJob(content=job_content, name=job_name, config_key=task, predecessors=depend, queue=self.queue)
        self.run_jobs([job], job_name, cluster_dir, workingdir=self.working_dir)
        return job
    
    def concat(self, depend=[]):
        merged_fasta_plus = '%s/merge/plus_strand/%s-contigs.fa' % (self.working_dir, self.lib_plus)
        merged_fasta_minus = '%s/merge/minus_strand/%s-contigs.fa' % (self.working_dir, self.lib_minus)
        concat_fasta = '%s/merge/%s-contigs.fa' % (self.working_dir, self.meta_lib)
        
        cluster_dir = '%s/merge/cluster' % (self.working_dir)
        if not os.path.isdir(cluster_dir):
            os.makedirs(cluster_dir)
        
        for outputfile in [merged_fasta_plus, merged_fasta_minus, concat_fasta]:
            if os.path.exists(outputfile) and os.path.islink(outputfile):
                # This is a re-run. The symlink to the file from the previous version must be removed.
                os.remove(outputfile)
        
        job_content = 'time cat %s %s > %s' % (merged_fasta_plus, merged_fasta_minus, concat_fasta)
        job_name = "%s-concat" % (self.lib)   
        job = MyJob(content=[job_content], name=job_name, config_key=None, predecessors=depend, queue=self.queue)
        self.run_jobs([job], job_name, cluster_dir, workingdir=self.working_dir)
        return job

    def calculate_assembly_stats(self, depend=[]):
        task = 'abyss-fac'
        cluster_dir = '%s/merge/cluster' % self.working_dir
        job_name = '%s-stats' % self.lib
        job_content = 'time abyss-fac -j -t %s assembly/plus_strand/k*/*-contigs.fa assembly/minus_strand/k*/*-contigs.fa merge/plus_strand/*-contigs.fa merge/minus_strand/*-contigs.fa merge/*-contigs.fa > merge/stats.txt' % str(self.readlength)
        job = MyJob(content=[job_content], name=job_name, config_key=task, queue=self.queue, predecessors=depend)
        self.run_jobs([job], job_name, cluster_dir, workingdir=self.working_dir)

    def prepare_reads(self):    
        assembly_plus_in = '%s/assembly/plus_strand/in' % self.working_dir
        assembly_minus_in = '%s/assembly/minus_strand/in' % self.working_dir
        
        plus_reads = []
        minus_reads = []
        unknown_reads = []

        fh = open(assembly_plus_in, 'r')
        for line in fh:
            plus_reads.append(line.strip())
        fh.close()       
        
        fh = open(assembly_minus_in, 'r')
        for line in fh:
            read = line.strip()
            if read in plus_reads:
                plus_reads.remove(read)
                unknown_reads.append(read)                
            else:
                minus_reads.append(read)
        fh.close()
        
        r2c_plus_in_raw = '%s/reads_to_contigs/plus_strand/in.raw' % self.working_dir
        r2c_plus_in = '%s/reads_to_contigs/plus_strand/%s.in' % (self.working_dir, self.meta_lib)
        
        r2c_minus_in_raw = '%s/reads_to_contigs/minus_strand/in.raw' % self.working_dir
        r2c_minus_in = '%s/reads_to_contigs/minus_strand/%s.in' % (self.working_dir, self.meta_lib)
        
        r2c_unknown_in_raw = '%s/reads_to_contigs/unknown/in.raw' % self.working_dir
        r2c_unknown_in = '%s/reads_to_contigs/unknown/%s.in' % (self.working_dir, self.meta_lib)
        
        for outputfile in [r2c_plus_in_raw, r2c_minus_in_raw, r2c_unknown_in_raw, r2c_plus_in, r2c_minus_in, r2c_unknown_in]:
            if os.path.exists(outputfile) and os.path.islink(outputfile):
                # This is a re-run. The symlink to the file from previous version must be removed.
                os.remove(outputfile)

        fh = open(r2c_plus_in_raw, 'w')
        print "Raw reads files for the plus strand:"
        for read in plus_reads:
            fh.write(read + os.linesep)
            print read
        fh.close()

        fh = open(r2c_minus_in_raw, 'w')
        print "Raw reads files for the minus strand:"
        for read in minus_reads:
            fh.write(read + os.linesep)
            print read
        fh.close()

        fh = open(r2c_unknown_in_raw, 'w')
        print "Raw reads files for unknown strand:"
        for read in unknown_reads:
            fh.write(read + os.linesep)
            print read
        fh.close()
        
        self.prepare_reads_helper(r2c_plus_in_raw, r2c_plus_in, '%s/reads_to_contigs/plus_strand/reads' % self.working_dir, '%s/reads_to_contigs/plus_strand/cluster' % self.working_dir, self.meta_lib)
        self.prepare_reads_helper(r2c_minus_in_raw, r2c_minus_in, '%s/reads_to_contigs/minus_strand/reads' % self.working_dir, '%s/reads_to_contigs/minus_strand/cluster' % self.working_dir, self.meta_lib)
        self.prepare_reads_helper(r2c_unknown_in_raw, r2c_unknown_in, '%s/reads_to_contigs/unknown/reads' % self.working_dir, '%s/reads_to_contigs/unknown/cluster' % self.working_dir, self.meta_lib)
                

    def align_reads(self):
        merged_fasta_plus = '%s/merge/plus_strand/%s-contigs.fa' % (self.working_dir, self.lib_plus)
        merged_fasta_minus = '%s/merge/minus_strand/%s-contigs.fa' % (self.working_dir, self.lib_minus)
        concat_fasta = '%s/merge/%s-contigs.fa' % (self.working_dir, self.meta_lib)
                
        outdir_plus = '%s/reads_to_contigs/plus_strand' % self.working_dir
        outdir_minus = '%s/reads_to_contigs/minus_strand' % self.working_dir
        outdir_unknown = '%s/reads_to_contigs/unknown' % self.working_dir
        
        min_k = None
        assembly_topdir = None
        kdirs = glob.glob(self.working_dir + '/assembly/plus_strand/k*')
        kdirs.extend(glob.glob(self.working_dir + '/assembly/minus_strand/k*'))
        for assembly in kdirs:
            if assembly_topdir is None:
                assembly_topdir = os.path.dirname(os.path.realpath(os.path.abspath(assembly)))
            k = os.path.basename(assembly).replace('k', '')
            if min_k is None or int(k) < int(min_k):
                min_k = k           
        if min_k is None:
            message = 'Cannot find minimum k value from %s/assembly' % (self.working_dir)
            print_message("WARNING", message, False)
        
        plus_job = self.align_reads_helper(merged_fasta_plus, outdir_plus + "/%s.in" % self.meta_lib, self.meta_lib, outdir_plus, outdir_plus + "/cluster", min_k)
        if not plus_job:
            message = "Cannot submit reads-to-contigs jobs for reads from plus strand."
            print_message("ERROR", message, True)

        minus_job = self.align_reads_helper(merged_fasta_minus, outdir_minus + "/%s.in" % self.meta_lib, self.meta_lib, outdir_minus, outdir_minus + "/cluster", min_k)
        if not minus_job:
            message = "Cannot submit reads-to-contigs jobs for reads from minus strand."
            print_message("ERROR", message, True)

        unknown_job = self.align_reads_helper(concat_fasta, outdir_unknown + "/%s.in" % self.meta_lib, self.meta_lib, outdir_unknown, outdir_unknown + "/cluster", min_k)
        if not unknown_job:
            message = "Cannot submit reads-to-contigs jobs for reads from unknown strand."
            print_message("ERROR", message, True)
       
        plus_bam = '%s/%s-contigs.bam' % (outdir_plus, self.meta_lib)
        minus_bam = '%s/%s-contigs.bam' % (outdir_minus, self.meta_lib)
        unknown_bam = '%s/%s-contigs.bam' % (outdir_unknown, self.meta_lib)
        merged_bam = '%s/reads_to_contigs/%s-contigs.bam' % (self.working_dir, self.meta_lib)

        for outputfile in [merged_bam, merged_bam + '.bai']:
            if os.path.exists(outputfile) and os.path.islink(outputfile):
                # This is a re-run. The symlink to the file from previous version must be removed.
                os.remove(outputfile)

        task = 'samtools_sort'
        
        job_content = []
        
        # Since the bam files have different headers, we cannot merge them using the usual way.
        # We must combine the sam records from the files like so:
        # - the header from unknown bam
        # - alignments from unknown bam, plus bam, minus bam (the order doesn't matter)
        job_content.append('{ samtools view -h %s && samtools view %s && samtools view %s; } |samtools view -hSu - |samtools sort -m 800000000 -o - $TMPDIR > %s.incomplete || { rm -f $TMPDIR*.bam 2> /dev/null; echo \'An error has occurred.\'; exit 1;}' % (unknown_bam, plus_bam, minus_bam, merged_bam) )
        job_content.append('samtools index %s.incomplete' % merged_bam)
        job_content.append('mv %s.incomplete.bai %s.bai' % (merged_bam, merged_bam))
        job_content.append('mv %s.incomplete %s' % (merged_bam, merged_bam))
        job_content.append('touch R2C.COMPLETE')
        
        cluster_dir = '%s/reads_to_contigs/cluster' % self.working_dir
        if not os.path.isdir(cluster_dir):
            os.makedirs(cluster_dir)
            
        job_name = '%s-mergebams' % self.meta_lib
        merge_job = MyJob(content=job_content, name=job_name, config_key=task, predecessors=[plus_job, minus_job, unknown_job], queue=self.queue)
        self.run_jobs([merge_job], job_name, cluster_dir, workingdir=self.working_dir)

def make_pipes_from_input_file(options):
    pipes = []
    startinglibrary = options.library
    numlibraries = options.numlibraries
    if not is_pos_int(numlibraries):
        message = 'The number of libraries specified (%s) is not a positive integer.' % str(numlibraries)
        print_message("ERROR", message, True)        
    seenstart = False        
    
    fh = open(options.input, 'r')
    for line in fh:
        cols = line.strip().split()
        num_cols = len(cols)
        if num_cols > 0: #otherwise, it's just a line of empty spaces                        
            if num_cols >= 6:            
                # LIBRARY ASSEMBLY PROJECT READLENGTH TYPE METALIBRARY
                # example:
                # A12345A /assembly/dir/A12345A Project_Name 75 plus_strand A12345
                # A12345B /assembly/dir/A12345B Project_Name 75 minus_strand A12345                    
                
                library, assembly_dir, project, readlength, library_type, meta_lib = cols
                if startinglibrary:
                    if startinglibrary == meta_lib:
                        seenstart = True
                    elif not seenstart:
                        continue
                    elif numlibraries and int(numlibraries)<=len(pipes):
                        break                
                
                if not is_pos_int(readlength):
                    message = 'The read length specified (%s) is not a positive integer.' % readlength
                    print_message("ERROR", message, True)
                if not os.path.isdir(assembly_dir):
                    message = 'Cannot find the assembly directory specified: %s' % assembly_dir
                    # Only a warning because users can potentially run Trans-ABySS on any FASTA file.
                    print_message("WARNING", message, False)
                
                library_type = library_type.lower()
                if library_type == 'plus_strand' or library_type == 'minus_strand':
                    line2 = fh.next()
                    cols2 = line2.strip().split()
                    if len(cols2) < 6:
                        message = "Less than 6 columns detected for stranded library..." + os.linesep + line + os.linesep + line2
                        print_message("ERROR", message, True)
                    library2, assembly_dir2, project2, readlength2, library_type2, meta_lib2 = cols2
                    if project != project2:
                        message = "The projects do not agree for stranded library..." + os.linesep + line + os.linesep + line2
                        print_message("ERROR", message, True)
                    if library_type == library_type2:
                        message = "Both strands cannot be the same..." + os.linesep + line + os.linesep + line2
                        print_message("ERROR", message, True)
                    if meta_lib != meta_lib2:
                        message = "Meta-library is not the same..." + os.linesep + line + os.linesep + line2
                        print_message("ERROR", message, True)
                    
                    lib1 = None
                    lib2 = None
                    adir1 = None
                    adir2 = None
                    if library_type == 'plus_strand':
                        lib1 = library
                        adir1 = assembly_dir
                        lib2 = library2
                        adir2 = assembly_dir2
                    else:
                        lib2 = library
                        adir2 = assembly_dir
                        lib1 = library2
                        adir1 = assembly_dir2                        

                    pipes.append(PipeStranded('stranded', project, meta_lib, lib1, lib2, str(min(readlength,readlength2)), adir1, adir2, cluster=options.cluster, queue=options.queue, local=options.local, debug=options.debug))
                elif 'transcriptome' == library_type:
                    pipes.append(PipeTranscriptome('transcriptome', project, library, readlength, assembly_dir=assembly_dir, cluster=options.cluster, queue=options.queue, local=options.local, debug=options.debug))
                elif 'genome' == library_type:
                    pipes.append(PipeGenome('genome', project, library, readlength, assembly_dir=assembly_dir, cluster=options.cluster, queue=options.queue, local=options.local, debug=options.debug))
                elif 'targeted_genome' == library_type or 'targetted_genome' == library_type:
                    pipes.append(PipeGenome('targeted_genome', project, library, readlength, assembly_dir=assembly_dir, cluster=options.cluster, queue=options.queue, local=options.local, debug=options.debug))
                else:
                    message = "Unknown library type..." + os.linesep + line
                    print_message("ERROR", message, True)                       
            elif num_cols >= 4:
                # LIB ASSEMBLY PROJECT READLENGTH TYPE
                # OR
                # LIB ASSEMBLY PROJECT READLENGTH
                
                library, assembly_dir, project, readlength = cols[0:4]
                if startinglibrary:
                    if startinglibrary == library:
                        seenstart = True
                    elif not seenstart:
                        continue
                    elif numlibraries and int(numlibraries)<=len(pipes):
                        break
                
                library_type = None
                if num_cols == 5:
                    library_type = cols[4].lower()
                else:
                    if not options.run_genome and not options.run_transcriptome:
                        message = "You must use either '--run_genome' or '--run_transcriptome'." + os.linesep + "You may alternatively specify the library type in the 5th column of input file."
                        print_message("ERROR", message, True)
                    elif options.run_genome and options.run_transcriptome:
                        message = "You must use only one of '--run_genome' or '--run_transcriptome'"
                        print_message("ERROR", message, True)                            
                
                if not is_pos_int(readlength):
                    message = 'The read length specified (%s) is not a positive integer.' % readlength
                    print_message("ERROR", message, True)
                if not os.path.isdir(assembly_dir):
                    # Only a warning because users can potentially run Trans-ABySS on any FASTA file in the right format.
                    message = 'Cannot find the assembly directory specified: %s' % assembly_dir
                    print_message("WARNING", message, False)                  

                if 'transcriptome' == library_type or options.run_transcriptome:
                    pipes.append(PipeTranscriptome('transcriptome', project, library, readlength, assembly_dir=assembly_dir, cluster=options.cluster, queue=options.queue, local=options.local, debug=options.debug))
                elif 'genome' == library_type or (options.run_genome and not options.targeted):
                    pipes.append(PipeGenome('genome', project, library, readlength, assembly_dir=assembly_dir, cluster=options.cluster, queue=options.queue, local=options.local, debug=options.debug))
                elif 'targeted_genome' == library_type or 'targetted_genome' == library_type or (options.run_genome and options.targeted):
                    pipes.append(PipeGenome('targeted_genome', project, library, readlength, assembly_dir=assembly_dir, cluster=options.cluster, queue=options.queue, local=options.local, debug=options.debug))
                elif 'plus_strand' == library_type or 'minus_strand' == library_type:
                    message = "Meta-library name is not specified in the 5th column..." + os.linesep + line
                    print_message("ERROR", message, True)
                else:
                    message = "Unknown library type..." + os.linesep + line
                    print_message("ERROR", message, True)
            elif num_cols == 3:
                # LIB ASSEMBLY PROJECT
                library, assembly_dir, project = cols[0:3]
                if startinglibrary:
                    if startinglibrary == library:
                        seenstart = True
                    elif not seenstart:
                        continue
                    elif numlibraries and int(numlibraries)<=len(pipes):
                        break
                
                readlength = options.readlength
                if options.fem and options.run_transcriptome:
                    if not readlength:
                        message = "Please specify the read length with '--readlength'." + os.linesep + "You may alternatively specify the read length in the 4th column of input file."
                        print_message("ERROR", message, True)
                    if not is_pos_int(readlength):
                        message = 'The read length specified (%s) is not a positive integer.' % readlength
                        print_message("ERROR", message, True)
                    
                if not os.path.isdir(assembly_dir):
                    # Only a warning because users can potentially run Trans-ABySS on any FASTA file in the right format.
                    message = 'Cannot find the assembly directory specified: %s' % assembly_dir
                    print_message("WARNING", message, False)                
                
                if not options.run_genome and not options.run_transcriptome:
                    message = "You must use either '--run_genome' or '--run_transcriptome'." + os.linesep
                    print_message("ERROR", message, True)
                elif options.run_genome and options.run_transcriptome:
                    message = "You must use only one of '--run_genome' or '--run_transcriptome'"
                    print_message("ERROR", message, True)
                    
                if options.run_transcriptome:
                    pipes.append(PipeTranscriptome('transcriptome', project, library, readlength, assembly_dir=assembly_dir, cluster=options.cluster, queue=options.queue, local=options.local, debug=options.debug))
                elif options.run_genome and not options.targeted:
                    pipes.append(PipeGenome('genome', project, library, readlength, assembly_dir=assembly_dir, cluster=options.cluster, queue=options.queue, local=options.local, debug=options.debug))
                elif options.run_genome and options.targeted:
                    pipes.append(PipeGenome('targeted_genome', project, library, readlength, assembly_dir=assembly_dir, cluster=options.cluster, queue=options.queue, local=options.local, debug=options.debug))
            else:
                #this line has incorrect format
                message = "This line of input file has less than 3 columns..." + os.linesep + line
                print_message("ERROR", message, True)
    fh.close()
    return pipes


if __name__ == '__main__':   
    usage = "Usage: %prog [options]"
    parser = OptionParser(usage=usage, version=BUILDID)
    
    sample_types = OptionGroup(parser, "Sample Types")
    sample_types.add_option("-G", "--run_genome", dest="run_genome", help="run genome pipeline", action="store_true", default=False)
    sample_types.add_option("-t", "--targeted", dest="targeted", help="genome is targeted, used in conjunction with --run_genome", action="store_true", default=False)
    sample_types.add_option("-T", "--run_transcriptome", dest="run_transcriptome", help="run transcriptome pipeline", action="store_true", default=False)
    parser.add_option_group(sample_types)
    
    inputs = OptionGroup(parser, "Inputs")
    inputs.add_option("-p", "--project", dest="project", help="project name", metavar='STR')
    inputs.add_option("-P", "--topdir", dest="top_dir", help="project directory", metavar='DIR')
    inputs.add_option("-a", "--assembly_dir", dest="assembly_dir", help="ABySS assembly directory", metavar='DIR')
    inputs.add_option("-l", "--library", dest="library", help="library name", metavar='STR')
    inputs.add_option("-L", "--readlength", dest="readlength", help="read length", type="int", metavar='INT')
    inputs.add_option("-X", "--reference", dest="reference", help="name of reference genome in the configuration file", metavar='STR')    
    inputs.add_option("-n", "--input", dest="input", help="input file (library, project, etc.)", metavar='FILE')
    inputs.add_option("-N", "--numlibraries", dest="numlibraries", help="process INT libraries [%default] in input file starting from the library specified with --library, used in conjunction with --input and --library", type="int", metavar='INT', default=1)
    parser.add_option_group(inputs)

    actions = OptionGroup(parser, "Actions", "For both genome and transcriptome analyses")
    actions.add_option("-d", "--set_up", dest="set_up", help="set up directory", action="store_true", default=False)
    actions.add_option("-0", "--fem", dest="fem", action="store_true", help="filter, extend, and merge ABySS assemblies", default=False)
    actions.add_option("-R", "--prepare_reads", dest="prepare_reads", help="prepare read file(s) for aligning reads to merged assembly", action="store_true", default=False)
    
    actions.add_option("-b", "--symlink_genome_bam", dest="symlink_genome_bam", help="Creates symlink for *.bam and *.bam.bai in the 'reads-to-genome' directory", action="store_true", default=False)
    parser.add_option_group(actions)
    
    actions.add_option("-r", "--align_reads", dest="align_reads", help="align reads to merged assembly", action="store_true", default=False)
    actions.add_option("-u", "--prepare_contigs", dest="prepare_contigs", help="prepare fasta files for aligning merged assembly to reference genome", action="store_true", default=False)
    actions.add_option("-c", "--align_contigs", dest="align_contigs", help="align merged assembly to reference genome", action="store_true", default=False)
    actions.add_option("-f", "--fusion", dest="fusion", help="call fusion events", action="store_true", default=False)
    actions.add_option("-i", "--indel", dest="indel", help="call indel events", action="store_true", default=False)
    
    actionsTranscriptome = OptionGroup(parser, "Actions", "For transcriptome analysis only")
    actionsTranscriptome.add_option("-k", "--track", dest="track", help="generate custom UCSC track", action="store_true", default=False)
    actionsTranscriptome.add_option("-x", "--splicing", dest="splicing", help="call novel splicing events", action="store_true", default=False)
    parser.add_option_group(actionsTranscriptome)    
                           
    rerun = OptionGroup(parser, "Re-run", "Options only for re-running analysis")
    rerun.add_option("-S", "--source", dest="source", help="source version number for re-runing alignment and/or analysis steps, used in conjuction with --set_up. Example: '--set_up --source 1.4.4'", metavar='STR')
    parser.add_option_group(rerun)
    
    cluster = OptionGroup(parser, "Host", "Sets which machines the jobs should run on")
    cluster.add_option("-C", "--cluster", dest="cluster", help="submit to cluster head node", metavar='STR')
    cluster.add_option("-q", "--queue", dest="queue", help="submit to cluster queue", metavar='STR')
    cluster.add_option("--local", dest="local", help="run locally", action="store_true", default=False)
    cluster.add_option("--qmake", dest="qmake", help="submit jobs with qmake instead of qsub, up to maximum of INT parallel make tasks", type="int", metavar='INT')
    parser.add_option_group(cluster)
    
    parser.add_option("-D", "--debug", dest="debug", help="run in debug mode; no jobs will be submitted", action="store_true", default=False)
            
    (options, args) = parser.parse_args()
        
    if options.debug:
        message = """Running in debug mode!
*******************************************************************
* No jobs will be run locally or submitted to the cluster.        *
* Only job scripts, directories, and/or symlinks will be created. *
*******************************************************************"""
        print_message("WARNING", message, False)
     
    if options.input:
        # input file specified        
        if not os.path.isfile(options.input):
            message = "Cannot find file: %s" % options.input
            print_message("ERROR", message, True)        
        pipes = make_pipes_from_input_file(options)
        
        for pipe in pipes:
            if options.qmake:
                pipe.qmake(options, options.qmake)
            else:
                pipe.process(options)
    else:
        # No input file specified
        
        # Check the basic requirements
        if not options.top_dir and not options.project:
            print_message("ERROR", "Either '--topdir DIR' or '--project' must be specified along with '--library STR' when '--input FILE' is not used.", True)
        if not options.library:
            print_message("ERROR", "'--library STR' must be specified along with either '--topdir STR' or '--project' when '--input FILE' is not used.", True)
        if (not options.run_genome and not options.run_transcriptome) or (options.run_genome and options.run_transcriptome):
            print_message("ERROR", "Either '--run_genome' or '--run_transcriptome' must be used.", True)
        
        # Check the the requirements for each step
        if not options.source and options.set_up and not options.assembly_dir:
            print_message("WARNING", "'--assembly_dir DIR' is not used along with '--set_up' when '--input FILE' is not used.", False)
        if options.fem and options.run_transcriptome and not options.readlength:
            print_message("ERROR", "'--readlength INT' must be specified along with '--fem' and '--run_transcriptome' when '--input FILE' is not used.", True)
        if (options.align_contigs or options.fusion or options.indel or options.splicing) and not options.project and not options.reference:
            print_message("ERROR", "'--reference STR' must be specified along with '--align_contigs' or '--fusion' or '--indel' or '--splicing' when both '--project STR' and '--input FILE' are not used.", True)

        if options.run_genome:
            assembly_method = 'genome'
            if options.targeted:
                assembly_method = 'targeted_genome'
            pipe = PipeGenome(assembly_method, options.project, options.library, options.readlength, assembly_dir=options.assembly_dir, cluster=options.cluster, queue=options.queue, local=options.local, debug=options.debug, genome=options.reference, top_dir=options.top_dir)
        else:
            assembly_method = 'transcriptome'
            pipe = PipeTranscriptome(assembly_method, options.project, options.library, options.readlength, assembly_dir=options.assembly_dir, cluster=options.cluster, queue=options.queue, local=options.local, debug=options.debug, genome=options.reference, top_dir=options.top_dir)
            
        if options.qmake:
            pipe.qmake(options, options.qmake)
        else:
            pipe.process(options)
    
